/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <sys/wait.h>

#include "memm.h"
#include "socketio.h"
#include "zlist.h"
#include "acceptor.h"
#include "notecraft.h"

int KidHandler::kidshave = 0;

Acceptor::Acceptor(char *programname, int port, char **env, NoteCraft &ncc, char *bindaddr) : nc(ncc)
  {
    /* 10/31/00 If bind address is null, allow, any, for now we'll just pick off
       one. How many interfaces are we really ever going to bind to. */
    SOCKET s; /* the listen socket */
    int dat;
    
#ifdef __linux__
    envp = NULL; // keep title updates from happening, seems to screw up linux ps
#else
    envp = env;
#endif
    origp0 = NULL;
    progname = programname;
    
    err = 0;
    fin = 0;
    s = socket (AF_INET, SOCK_STREAM, 0);
    if (s < 0)
      {
        err = errno;
        return; 
      } 

    fin = s;
    memset (&addr, 0, sizeof(addr));
    addr.sin_family      = AF_INET;
    if (bindaddr != NULL)
      if (strlen(bindaddr) == 0)
        bindaddr = NULL;
        
    if (bindaddr == NULL) // 10/31/00
      addr.sin_addr.s_addr = INADDR_ANY; 
    else
      {
        if (isipaddress(bindaddr))
          addr.sin_addr.s_addr = inet_addr(bindaddr);
        else
          {
            addr.sin_addr.s_addr = lookuphost(bindaddr);
            if (addr.sin_addr.s_addr == 0)
              {
                ncc(NOTESEV) << "Unable to gethostbyname for: " << bindaddr << 
                     " errno = " << errno << " " << strerror(errno) << endl;
                return;
              }
          }
        ncc(NOTEINF) << "Binding to: " << bindaddr << endl;
      }
    // that should be inet_addr(INADDR_ANY); but linux blows up on it, and it's 0 anyway
    addr.sin_port        = htons(port);
    
    dat = TRUE;
    setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&dat, sizeof(dat));
    
    if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0)
      {
        err = errno;
        return; 
      } 
      
    if (listen(s, 5) != 0)
      err = errno;
  }

Acceptor::Acceptor(NoteCraft &ncc) : nc(ncc)
  {
    /* This is dumb, but we need a placeholder acceptor
       for classes that sometimes use acceptors and 
       sometimes don't. Oh well. */
    fin = 0;
    err = 0;
    namelen = 0;
    envp = NULL;
    origp0 = NULL;
  }

Acceptor::~Acceptor(void)
  {
    if (fin != 0)
      close(fin);
    if (origp0 != NULL)
      {
        free (envp[0]);
        envp[0] = origp0;
      }
  }

char *Acceptor::errstr(void)
  { 
    if (err == 0) 
      return "";
    char *sav = strerror(err);
    err = 0;
    return sav; 
  }
  
SOCKET Acceptor::accept(int timeout)
  {
    if (fin == 0)
      return 0;
      
    SOCKET ret;

    /* 10/8/02 Allow for a timeout period in which we will return -1
       and set errno to etimed out */
    settitle("Waiting for connection...");
    namelen = sizeof(addr); // this will get set by accept() shortly...
    while (1) /* loop for accepting connections we don't allow */
      {
        if (timeout > 0) // if they want to time out
          {
            fd_set fd;
            FD_ZERO(&fd);
            FD_SET(fin, &fd);
            struct timeval tme;
            tme.tv_sec = timeout;  // 1/3/03 for some reason I got it right here.
            tme.tv_usec = 0; 

            ret = select (fin+1, &fd, 0, 0, &tme); // connect = read
            if (ret == 0)
              { // we timed out
                errno = ETIMEDOUT;
                return -1;
              }
            if (FD_ISSET(fin, &fd) == 0)
              continue; // it wasn't out socket, block again
            // if we get here, then either there's an error on select
            // or there's a connection to accept
          }
        ret = ::accept(fin, (struct sockaddr *)&addr, &namelen);
        if (ret < 0)
          err = errno;
        else
          {
            /* See if we should accept this connection */
            if (connectionallowed(ret) == FALSE)
              {
                close(ret); // drop 'em like a wet tomato
                continue; // go back to accepting
              }
           
            /* Get peer info */
            struct sockaddr_in peer;
            size_t size = sizeof(peer);
            Memory ttl;
            Memory adr;
            char *s = "Handling connection from ";
            if (getpeername(ret, (struct sockaddr *)&peer, &size) >= 0)
              {
                char *padr = inet_ntoa(peer.sin_addr);
                if (padr != (char *)-1)
                  {
                    Memory t2(s, padr);
                    adr = padr;
                    ttl = t2;
                  }
                else
                  {
                    Memory t2(s, "unknown");
                    adr = "unknown";
                    ttl = t2;
                  }
              }
            else
              {
                Memory t2(s, "unknown");
                adr = "unknown";
                ttl = t2;
              }
            
            settitle(ttl);
            nc(NOTEINF) << "Accepted connection from: " << adr << endl;
          } // if accepted

        return ret;
      } // while 1
  }

int Acceptor::connectionallowed(int sfd)
  {
    // string must be in partial dottet notation, no hostnames.
    // return true if we allow a connection from this place, false if not
    int lc = hostsallowlist.ListCount();
    if (lc == 0)
      return TRUE; // always if list is empty

    int lp;
    for (lp = 0; lp < lc; lp++)
      {
        stringentry *s = (stringentry *)hostsallowlist.GetItem(lp);
        Memory one = s->getstr();
        
        /* Get address of peer */
        Memory mpeer;
        struct sockaddr_in peer;
        size_t size = sizeof(peer);
        if (getpeername(sfd, (struct sockaddr *)&peer, &size) >= 0)
          mpeer = inet_ntoa(peer.sin_addr);
          
        UINT ps = strlen(one);
        /* Make them the same size if we're comparing class C's or something
           192.168.1.2 becomes 192.168.1 then compare that.
           True, it will allow 192.168.10 also, but not a big deal.
           If you care that much, then put the trailing dot: 192.168.1. */
           
        if (strlen(mpeer) >= ps)
          *(mpeer+ps) = '\0'; // shorten it to match same size
        if (strcmp(mpeer, one) == 0)
          return TRUE;
      }
    return FALSE;
  }
  
void Acceptor::sethostsallowlist(char *s)
  {
    /* read this string and make a zlist of stringentry of ip ranges
       that we will accept from. If string/list is empty, we accept from all. */
    char *pos = s;
    hostsallowlist.WipeList();
    
    while (*pos != '\0')
      {
        while (*pos == ' ')
          pos++;
        Memory one;
        while (((*pos >= '0') && (*pos <= '9')) || (*pos == '.'))
          {
            one.addchar(*pos);
            pos++;
          }
        if (strlen(one) > 0)
          {
            stringentry *str = new stringentry(one);
            hostsallowlist.Add(str);
          }
      }
  }
  
void Acceptor::settitle (char *s) 
  {
    UINT lp;
    
    if (envp == NULL)
      return;
    if (origp0 == NULL)
      {
        /* Set it up for the first time. */
        origp0 = envp[0]; 
        envp[0] = (char *)malloc (TITLESIZE+20); // I'm not sure this is legal or portable
      }
    Memory tt(progname, ": ");
    tt.addstr(s);

    for (lp = 0; lp < strlen(tt); lp++)
      {
        if (lp == TITLESIZE)
          {
            envp[0][lp] = '\0';
            return;
          }
        envp[0][lp] = *(tt+lp);
      }
    envp[0][lp] = '\0';
  }

/**************************************************/

extern "C"
  {
    void kidhandler(int sigNum)
      {
        wait (NULL); /* Kill the wabbit */
        /* ugly but works... */
        KidHandler::KidDied();
        
        return;
      }
  }
            
KidHandler::KidHandler(int num, ConfigCraft &cc, Acceptor &ac, RunFunction run) : ccc(cc)
  { // the socket version
    kidswant = num;
    kidshave = 0;
    runfunc = run;
    nsrunfunc = NULL;
    kidcounter = 0;
    acc = &ac;
    ncc = NULL;
    return;
  }

KidHandler::KidHandler (int num, ConfigCraft &cc, NoteCraft &nc, nosocketRunFunction run) : ccc(cc)
  { // the non socket version
    kidswant = num;
    kidshave = 0;
    runfunc = NULL;
    nsrunfunc = run;
    kidcounter = 0;
    ncc = &nc;
    acc = NULL;
    return;
  }

void KidHandler::KidDied(void)
  {
    kidshave--;
  }
  
int KidHandler::run(void)
  {
    /* Fire off this many kids doing runner */
    struct sigaction acstr;
                       
    acstr.sa_handler = kidhandler;
    sigemptyset(&acstr.sa_mask);
    acstr.sa_flags = SA_NOCLDSTOP; 
    sigaction (SIGCHLD, &acstr, NULL);
   
    if (kidswant == 0)
      {
        int ret;
        if (runfunc)
          {
            acc->getNC()(NOTEINF) << "Starting child " << acc->getNC().gettitle() << endl;
            ret = (*runfunc)(1, ccc, *acc);
          }
        else 
          {
            (*ncc)(NOTEINF) << "Starting child " << ncc->gettitle() << endl;
            ret = (*nsrunfunc)(1, ccc, *ncc);
          }
        return ret; // no fork mode returns value.
        // this is so we can test for mem leaks. exit(0) doesn't
        // call desctructors. It just exits.
      }
    
    while (1)
      {
        pid_t kid;
  
        sleep(1);
        while (kidshave < kidswant)
          {
            kid = ForkKid(++kidcounter, *acc); // do we really want to keep track of ID's?
            if (kid == 0) // the child ended
              return 0;
          }
      }
  }
  
pid_t KidHandler::ForkKid(int lp, Acceptor &ac)
  {
    pid_t kid;
    
    kid = fork();
    if (kid == -1)
      {
        err = errno;
        return 0;
      }
    kidshave++;
    if (kid == 0)
      {
        // we are the child
        int ret;
        if (runfunc)
          {
            acc->getNC()(NOTEINF) << "Starting child " << acc->getNC().gettitle() << endl;
            ret = (*runfunc)(lp, ccc, *acc);
          }
        else 
          {
            (*ncc)(NOTEINF) << "Starting child " << ncc->gettitle() << endl;
            ret = (*nsrunfunc)(lp, ccc, *ncc);
          }
        return 0; // after the kid runs, we return 0 so they can clean up nicely.
      }
    return kid; // the parent returns process id
  }


