/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef ACCEPTOR_H
#define ACCEPTOR_H

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>

#include "memm.h"

typedef int SOCKET;

/* Instructions: 

   There are two classes in here....
   
   Acceptor and KidHandler.
   
   Acceptor constructor:
   Acceptor(char *programname, int port, char **envp, NoteCraft &nc, char *bindaddr);
     programname is for logging, port is the port you want this acceptor's socket
     to bind to. envp is actually the argv from main() and nc is a notecraft
     object. See NoteCraft.h, it's a log stream.
     If you set envp to NULL, the settitle functionality will be turned off. Since
     I'm not 100% this is legal to do, if you run into weird sigsevs, you might want
     to set envp to NULL.
     Don't set programname to NULL.
     10/31/00 Added bind address. Rick dutifully informed me it would be 
     a good idea to bind to a particular interface. Silly me.
   
   SOCKET accept(void);
     call this, to block on an accept waiting for a connection. It will return you
     a socket or -1. If it's -1, you can read errorstate() or errstr().
     You can then pass this socket to a socketio class.
     This is a socket factory. You must close (or have socketio close) this 
     file descriptor when you're done with it.
     
   int errorstate(void);
     returns the errno value if there was an error upon construction or accept
     
   char *errstr(void);
     returns strerr of the last errorstate and resets the errorstate. I.E. You can
     only do this once.

   void settitle (char *s);
     This seems to work in AIX but not linux. It allows you to set the string
     that appears in ps when you do a ps x or ps -ef or whatever. 

   NoteCraft &getNC(void);
     Allows access to the log stream from the Acceptor. The acceptor can be 
     passed to a Kidhandler, so you program can get at the log stream this way.

                                        *** 

   Kidhandler is like servermanager. It has two constructors. One that
   acts like servermanager, one that acts like TCPServermanager.
   You tell it how many kids to run, it will run them. If one dies, it
   will fire off another.
   

   KidHandler constructor:
   KidHandler (int num, ConfigCraft &cc, Acceptor &ar, RunFunction runner);
     Acts like TCPServerManager.
     num is the number of child processes you want. set it to 0 and it'll function
     in non-forking mode. The one session runs and then it ends.
     Configcraft is the logstream for it to use. (You sorta have to buy into
     the whole set of toys at once.) 
     Acceptor is a previously constructed Acceptor object. This allows all
     the kids to block accepting on one socket.
     The runfunction is the callback that run() (see below) calls after it
     has forked its children. The prototype for the run function must be
     To get the logstream in runner, you ask the acceptor for the reference.
     
     int runner (int id, ConfigCraft &ar, Acceptor &ac);
     
     id will be set to an increasing number for every kid that's fired off.

   KidHandler (int num, ConfigCraft &cc, NoteCraft &nc, nosocketRunFunction runner); 
     This is a version of the same thing that doesn't need an acceptor. If you want
     to run a program with kids not socket accept related, use this constructor.
     Since there's no Acceptor to get the NoteCraft from, it's passed as a parameter
     to the runner function.

     int runner(int id, ConfigCraft &ar, NoteCraft &nc);
   
   int run(void);
     Call this from main, after you've set everything else up. See below.
     This function never returns, neither in forked nor non forked mode.
     It forks the appropriate number of kids and then calls the RunFunction
     you set in the constructor.
     Addendum: This function does return if it's in no-fork (zero kids) mode.
     AddAddendum: It also returns as the child process function ends. So 
     you may have to prepare cleanup routines for any crap the children left 
     over. The parent process never exits.
     
   static void KidDied(void);
     Don't touch this.
     
   char *errstr(void);
     returns strerror of the last error condition.

   
   An simple example program of how to use everything...
   
   #define TXT_NAME "test"
   int main(int num, char *opts[])
     {
       ConfigCraft ar(TXT_NAME, num, opts); // will read test.cf unless cmd line overrides.
       NoteCraft   ls(TXT_NAME, TXT_NAME "log", val(ar[TXT_LOGLEVEL])); // min loglevel. default = 0
       Acceptor    ac(TXT_NAME, val(ar[TXT_PORT]), opts, ls); // make a socket and listen
       if (ac.errorstate()) // check for error
          cout << ac.errstr() << endl;
       KidHandler kd(val(ar[TXT_KIDS]), ar, ac, runner); // setup the kidhandler.
      
       kd.run(); // spawn children, parent never comes back, kids need cleanup
       return 0;
     }
       
   int runner(int num, Acceptor &ac)
     {
       NoteCraft &ls = ac.getNC();
       
       SOCKET inout;
       inout = ac.accept(); // this blocks
       if (inout == -1)
         cout << ac.errstr() << endl;
       else
         {
           socketio sco (inout, inout); // takes ownership of the socket/FD
           sco.sendstr("Got socket.");
         }
       return 0; // child will exit, kidhandler will make another
     }
     
     
*/     

/* All the stuff we need to accept sockets and spawn children */

class Acceptor;
class NoteCraft;
class ConfigCraft;

#define TITLESIZE 50 // max size for a title in ps

// handy for configcraft
#define TXT_PORT "port"  // acceptor
#define TXT_KIDS "kids"  // kidhandler
#define TXT_BINDADDR "ip-address" // for binding to an interface
#define TXT_HOSTS_ALLOW "hosts-allow"

typedef int(*RunFunction)(int, ConfigCraft &, Acceptor &);
typedef int(*nosocketRunFunction)(int, ConfigCraft &, NoteCraft &);

class Acceptor 
  {
    public:
      Acceptor(NoteCraft &nc); // minimally placeholderish acceptor. ugh.
      Acceptor(char *programname, int port, char **envp, NoteCraft &nc, char *bindaddr);
      ~Acceptor(void);
      SOCKET accept(int timeout = 0);
      int errorstate(void) { return err; }
      char *errstr(void);
      void settitle (char *s);
      NoteCraft &getNC(void) { return nc; }
      SOCKET getsd(void) { return fin; }
      void sethostsallowlist(char *s);
      void getpeeraddress(struct sockaddr_in *&Sin, int &Namelen)
        { Sin = &addr; Namelen = namelen; }
      int connectionallowed(int sfd); // it's handy for fserver, so I made it public
    private:
      
      SOCKET fin;
      int err;
      size_t namelen; // so the caller can get peer address info
      struct sockaddr_in addr;      
      char **envp;
      char *origp0;
      Memory progname;
      NoteCraft &nc;
      zlist hostsallowlist;
  };

class KidHandler 
  {
    public:
      KidHandler (int num, ConfigCraft &cc, Acceptor &ar, RunFunction run);
      KidHandler (int num, ConfigCraft &cc, NoteCraft &nc, nosocketRunFunction run);
      int run(void);
      static void KidDied(void);
      char *errstr(void) { return strerror(err); }
      
    private:
      pid_t ForkKid(int lp, Acceptor &ac);
      int kidswant;
      static int kidshave;
      int kidcounter;
      int err;
      Acceptor *acc; // mutually exclusive, can't require both to be set so can't use ref
      NoteCraft *ncc;
      ConfigCraft &ccc;
      RunFunction runfunc;
      nosocketRunFunction nsrunfunc;
  };




#endif
