/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef HASHTABLE_H
#define HASHTABLE_H

/* A template class to make a hash key value lookup table. 

   usage:
   
   hashtable<int> ht(10); // 10 items in list
   ht.set("hi", 1);
   int i;
   ht.get("hi" i);

   now there's an iterator too.
   resetiterator() and 
   int ip* = iterator();

   ht.resetiterator();
   char *k; int *v;
   while (ht.iterator(k, v) != FALSE)
     {
       printf ("%s = %d\n", k, *v);
     }


   hashtable will resize from initial size passed to constructor if enough
   items are added to hash table.
   List will never return duplicate values if duplicate keys are added.
*/

/* the multiply factor to make room for possible hash results
   the bigger the number the faster the hash will function, 
   the more memory it will take. The type passed to this template
   must have a functioning copy constructor. */

/* Today I learned that you if you don't want to include template method
   bodies in the template class declaration, you have to do this:
   template <class T> hashtable<class T>::function(params) {}
*/

#define HT_EXTRAROOM 2

template <class T> class hashtable
  {
    public:
      hashtable (int initialsize = 100)
        {
          tablelength = initialsize;
          if (tablelength < 10)
            tablelength = 10;

          tablelength *= HT_EXTRAROOM; // so we have room to thrash the hash about
          
          key = new Memory[tablelength]; // self blanking
          value = new T[tablelength]; // array of T objects
          resetiterator();
        }
        
      ~hashtable(void)
        { 
          delete [] value;
          delete [] key;
        }

      void resetiterator(void)
        {
          iterpos = 0;
        }

      int iterator(char *&ikey, T *&ivalue)
        {
          /* return FALSE if there's no more in the list, otherwise it sets
             the pointers you pass in to the next key and value in the list */

          while (iterpos < tablelength)
            {
              if (strlen(key[iterpos]) > 0)
                {
                  ikey = key[iterpos];
                  ivalue = &value[iterpos];
                  iterpos++;
                  return TRUE;
                }
              iterpos++;
            }
          return FALSE;
        }
        
      int set(char *k, T Value) // return FALSE if table is full (future will resize)
        {
          int hv = hashstring(k);
          int pos = findspace(hv);
          if (pos == -1)
            {
              // we ran out of space, make the hashtable bigger and rehash
              resizehash();
              pos = findspace(hv);
              if (pos == -1) // if we STILL can't find space, there's a problem
                return FALSE;
            }
          key[pos] = k;
          value[pos] = Value; // here's where the copy contructor is important
          return TRUE;
        }
        
      int get(char *k, T &Value) // return TRUE and set val if there, FALSE if not found
        { // return TRUE and set val if there, FALSE if not found
          int pos = lookup(k);
          if (pos == -1) 
            return FALSE;
          Value = value[pos];
          return TRUE;
        }
      
    private:

      int hashstring(char *s)
        {
          /* Come up with a good hash for this string to a number within tablesize.
             for now we'll add the chars. */
          int ret = 0;
          int lp;
          int max = strlen(s);
          for (lp = 0; lp < max; lp++)
            ret += (tablelength/4+1) * (int)s[lp]; // always uppercase
          ret %= tablelength;
          return ret;
        }

      int findspace(int pos)
        {
          /* starting from this position, find an empty spot, return -1 if 
             error, which of course should never happen */
          int lp = pos;
          if (strlen(key[lp]) == 0)
            return lp;
          lp++;
          while (lp != pos) // if we get to where we started, stop
            {
              if (lp == tablelength)
                {
                  lp = 0;
                  if (pos == 0) // would otherwise cause an endless loop if pos == 0
                    pos = 1;
                }
              if (strlen(key[lp]) == 0)
                return lp;
              lp++;
            }
          return -1;
        }

      int lookup(char *text)
        {
          /* Given this piece of text, find the position in the hash table and return it.
             -1 if not found */
          int pos = hashstring(text);

          /* now look for it. */
          int lp = pos;
          if (strcmp(text, key[lp]) == 0) // be opportunistic, see if it's right where we expect
            return lp;
          lp++;
          while (lp != pos) // if we get to where we started, stop
            {
              if (lp == tablelength)
                {
                  lp = 0;
                  if (pos == 0) // would otherwise cause an endless loop if pos == 0
                    pos = 1;
                }
              if (strcmp(text, key[lp]) == 0)
                return lp;
              lp++;
            }
          return -1;
        }

      int resizehash(void)
        {
          Memory *oldkey = key;
          T *oldvalue = value;
          int oldtablelength = tablelength;
          
          tablelength *= 2; // double it
          key = new Memory[tablelength];
          value = new T[tablelength]; 
          if (rehash(oldkey, oldvalue, oldtablelength) == FALSE) // from old to new
            {
              // put it back
              delete [] key;
              delete [] value;
              tablelength = oldtablelength;
              key = oldkey;
              value = oldvalue;
              return FALSE;
            }
          delete [] oldkey;
          delete [] oldvalue;
          return TRUE;
        }

      int rehash(Memory *oldkey, T *oldvalue, int oldtablelength)
        {
          // copy all the live items from the old table to the new
          int lp;
          for (lp = 0; lp < oldtablelength; lp++)
            {
              if (strlen(oldkey[lp]) > 0) // something here
                {
                  if (set(oldkey[lp], oldvalue[lp]) == FALSE)
                    {
                      // printf ("sanity failure in rehash."); 
                      /* we can't possibly run out of room, of course this will
                         never happen because set calls rehash and we'll loop forever
                         and blow out the stack. Have a nice day. */
                      return FALSE;
                    }
                }
            }
          return TRUE;
        }

      Memory *key;
      int iterpos;
      T *value;
      int tablelength;
  };
  
#endif


