/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef INCL_MUTEX_H
#define INCL_MUTEX_H

// simple mutex class, instantiating by name makes and holds the lock
// destruction frees it

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <errno.h>

class MutexLock
  {
    public:
      MutexLock(int id)
        { 
          lock[0].sem_num = 0;
          lock[0].sem_op = 0;
          lock[0].sem_flg = 0;
          lock[1].sem_num = 0;
          lock[1].sem_op = 1;
          lock[1].sem_flg = SEM_UNDO | S_IRWXU | S_IRWXG | S_IRWXO;
          unlock[0].sem_num = 0;
          unlock[0].sem_op = -1; // why nowait? avoid error if gets < 0?
          unlock[0].sem_flg = (IPC_NOWAIT | SEM_UNDO | S_IRWXU | S_IRWXG | S_IRWXO);
          semid = semget(id, 1, 0666 | IPC_CREAT);
          if (semid == -1)
            printf ("semget1 errno = %d - %s\n", errno, strerror(errno));
          int ret = semop(semid, &lock[0], 2);
          if (ret == -1)
            printf ("semget2 errno = %d - %s\n", errno, strerror(errno));
        }
      ~MutexLock(void) 
        {
          int ret = semop(semid, &unlock[0], 1);
          if (ret == -1)
            printf ("~semop1 errno = %d - %s\n", errno, strerror(errno));
        }

    private:
      int semid;
      struct sembuf lock[2];
      struct sembuf unlock[1];
  };

/* The logic goes like this. Multiple parrallel reads, or only one write. If writing, no writes or
   reads can happen 
   
   read
   ----
   while (w > 0);
   r++;
   read;
   r--;

   write
   -----
   while (w > 0);
   w++;
   while (r > 0);
   write
   w--;
   
   The r semaphore is sem #0 the w is sem #1 
   
   You use it like this:
      RWMutexLock x(1234);
      
      To wrap a read, do MutexLockRead y(x);
      when it goes out of scope it frees the read

      To wrap a write, do MutexLockWrite y(x);
      when it goes out of scope it frees the write.
      
      Multiple reads but only one write will occur.
         
   
*/

class RWMutexLock
  {
    public:
      RWMutexLock(int id)
        { // set up all the things we're ever going to have to do here.
          const int r = 0;
          const int w = 1;
          
          startread[0].sem_num = w;  // operation 1: while (w>0)
          startread[0].sem_op = 0;
          startread[0].sem_flg = 0;

          startread[1].sem_num = r;  // operaion 2: r++
          startread[1].sem_op = 1;
          startread[1].sem_flg = SEM_UNDO | S_IRWXU | S_IRWXG | S_IRWXO;

          endread  [0].sem_num = r;  // operaion 1: r--
          endread  [0].sem_op = -1;  // this will not block as long as semval >= 0
          endread  [0].sem_flg = SEM_UNDO | S_IRWXU | S_IRWXG | S_IRWXO;

                                         /* * * * */
          
          startwrite[0].sem_num = w;  // operation 1: while (w>0)
          startwrite[0].sem_op = 0;
          startwrite[0].sem_flg = 0;

          startwrite[1].sem_num = w;  // operation 2: w++
          startwrite[1].sem_op = 1;
          startwrite[1].sem_flg = SEM_UNDO | S_IRWXU | S_IRWXG | S_IRWXO;

          startwrite[2].sem_num = r;  // operation 3: while (r>0)
          startwrite[2].sem_op = 0;
          startwrite[2].sem_flg = 0;

          endwrite  [0].sem_num = w;  // operaion 1: w--
          endwrite  [0].sem_op = -1;  // this will not block as long as semval >= 0
          endwrite  [0].sem_flg = SEM_UNDO | S_IRWXU | S_IRWXG | S_IRWXO;

          semid = semget(id, 2, 0666 | IPC_CREAT); // get one for r and one for w
          if (semid == -1)
            printf ("semget1 errno = %d - %s\n", errno, strerror(errno));
        }
      void start_read(void)
        {
          int ret = semop(semid, &startread[0], 2); // 2 things to do
          if (ret == -1)
            printf ("start_read errno = %d - %s\n", errno, strerror(errno));
        }
      void end_read(void)
        {
          int ret = semop(semid, &endread[0], 1); // 1 thing to do
          if (ret == -1)
            printf ("end_read errno = %d - %s\n", errno, strerror(errno));
        }
      void start_write(void)
        {
          int ret = semop(semid, &startwrite[0], 3); // 3 things to do
          if (ret == -1)
            printf ("start_write errno = %d - %s\n", errno, strerror(errno));
        }
      void end_write(void)
        {
          int ret = semop(semid, &endwrite[0], 1); // 1 thing to do
          if (ret == -1)
            printf ("end_write errno = %d - %s\n", errno, strerror(errno));
        }
      ~RWMutexLock(void) 
        {
        }

    private:
      int semid; // the read and write sems
      struct sembuf startread[2];
      struct sembuf endread[1];
      struct sembuf startwrite[3];
      struct sembuf endwrite[1];
  };

/* Obviously these have to go out of scope before the RWMutexlock they refer to does. */

class MutexLockRead
  {
    public:
      MutexLockRead (RWMutexLock &m) : mut(m)
        {
          mut.start_read();
        }
      ~MutexLockRead(void)
        {
          mut.end_read();
        }

    private:
      RWMutexLock &mut;
  };

class MutexLockWrite
  {
    public:
      MutexLockWrite (RWMutexLock &m) : mut(m)
        {
          mut.start_write();
        }
      ~MutexLockWrite(void)
        {
          mut.end_write();
        }

    private:
      RWMutexLock &mut;
  };


#endif

