/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "fc.h"

/*                              Flat File                                */


FlatFile::FlatFile (void)
  {
    id = -1;
    hasindex = FALSE;
    headersize = 0;
    recordsize = 0;
    keysize = 0;
    ref = new refcounter();
    ref->refcount++;
  }
  
FlatFile::FlatFile (FlatConnection &sfc, char *fname, int sheadersize, int srecordsize, int shasindex, int skeysize, int sid) 
  {
    filename = fname;
    headersize = sheadersize;
    recordsize = srecordsize;
    hasindex = shasindex; // we force the name
    keysize = skeysize;
    fc = sfc; // this connection won't own sco.
    id = sid;
    ref = new refcounter();
    ref->refcount++;
  }
  
FlatFile::FlatFile (const FlatFile &sffr)
  { // copy constructor
    FlatFile &ffr = (FlatFile &)sffr;

    /* we don't have to worry like we do in the operator=
       situation because *this was never used for anything
       before. it's a new object, can't possibly point to anything. */
       
    filename = ffr.filename;
    id = ffr.id;
    fc = ffr.fc;
    headersize = ffr.headersize;
    recordsize = ffr.recordsize;
    hasindex = ffr.hasindex;
    keysize = ffr.keysize;
    ref = ffr.ref;
    ref->refcount++;
  }

FlatFile &FlatFile::operator=(const FlatFile &sffr)
  { // copy constructor
    if (this == &sffr)
      return *this;
    FlatFile &ffr = (FlatFile &)sffr;

    // if we were a real connection, and we're the last one, we have
    // to act like destructor.
    if (ref != NULL)
      {
        ref->refcount--;
        if (ref->refcount == 0)
          { // 'destroy' old object/close file
            finalclose();
          }
      }

    filename = ffr.filename;
    id = ffr.id;
    fc = ffr.fc;
    headersize = ffr.headersize;
    recordsize = ffr.recordsize;
    hasindex = ffr.hasindex;
    keysize = ffr.keysize;
    ref = ffr.ref;
    ref->refcount++;    
    return *this;
  }

void FlatFile::finalclose(void)
  { // shared function for copy constructor and destructor
   
    FlatPacket fp;
    FlatResponse fr;
    
    /* Here's the problem. If we give a copy of this file to the flat
       connection when we open it, we'll never hit zero, because the flat
       connection has one. Oh well. So it'll stay around. Open once, use 
       many. :-) */
       
    delete ref;

    if (id == -1)
      return;
    fp.command = FP_CLOSEFILE;
    fp.fileid = id;
    fc.sendfp(fp);
    fc.getfr(fr); // what, they're going to say no?
  }
  
FlatFile::~FlatFile (void)
  { // this will send a close command of this file id
    
    ref->refcount--;
    if (ref->refcount > 0)
      return;
    finalclose();
  }

/*                               Server calls                               */

fresp FlatFile::request(FlatPacket &fp, ULONG &dout, Memory &data, int length, Memory data2, int length2)
  { 
    /* everybody pretty much has to do this, consolidate it here. 
       Most functions have data, so everybody passes something. 
       If length2 is -1, there is no data2. You must specify a length for data2
       
       returns FR_* 
       0 = fail, 1 = good, 2... = server call worked, but specific error 
       If you pass no length, it will assume a NT string, otherwise it sends
       the length you specifiy from data. */
    fresp ret;
    
    ret = fc.sendfp(fp);
    if (ret == FALSE)
      return FR_FAIL;
    int r;
    if (length != -1)
      {
        fc.getsco()->putint(length); // 1/9/01 send length of 'key'
        r = fc.getsco()->sendstr(data, length);
        if (r == FALSE)
          return FR_FAIL;
      }
    
    if (length2 != -1)
      {
        fc.getsco()->putint(length); // 1/9/01 send length of 'key'
        r = fc.getsco()->sendstr(data2, length2);
        if (r == FALSE)
          return FR_FAIL;
      }
      
    // get response packet.
    FlatResponse fr;
    ret = fc.getfr(fr);
    if (ret != FR_OK)
      return FR_FAIL;
    dout = fr.data.recno;
    return fr.rc;
  }

void FlatFile::keyize (char *keyin, Memory &keyout, int keylen)
  {
    // very important that source string is at least keysize len, or we will read bad memory on memcpy
    Memory k(keysize);
    if (keylen == 0)
      keylen = strlen(keyin); // assume string if not specified
    memset (k, ' ', keysize);
    if (keylen > keysize)
      keylen = keysize;
    memcpy (k, keyin, keylen); // ensure a full length space padded key, doesn't copy NT
    keyout = k; // smart copy, does whole memblock, not string.
  }

void FlatFile::szfix(Memory &me, UINT size)
  {
    /* make this m match the size */
    /* Actually, we only have to make it longer if it ends up being too short. */
    char *m = me;
    UINT sz;

    memcpy (&sz, m-4, 4); /* get long back */
    if (sz == size)
      return;
    if (sz > size)
      {
// this is okay and will happen with fastmem printf ("Warning: header data bigger than header db field.\n");
        return;
      }
//printf ("Warning: header data shorter than header db field. Padding.\n");
    int add = size - sz;
    Memory addstr(add+1); // room for nt
    memset (addstr, ' ', add);
    *(addstr+add) = '\0';
    me.addstr(addstr); // pad with spaces
  }

fresp FlatFile::getheader(Memory &out)
  {
    // returns 0 if failed
    // fixing this to return a correctly sized zeroed buffer on success or failure
    FlatPacket fp;
    Memory x(headersize); // no data
    ULONG d;

    memset(x, 0, headersize);
    out = x; // set default if failure
    fp.command = FP_GETHEADER;
    fp.fileid = id;
    if (request(fp, d, x) != FR_OK)
      return FR_FAIL;

    // retrieve command specific data here.
    int r = fc.getsco()->getstr(out);
    if (r == FALSE)
      return FR_FAIL;
    szfix(out, headersize);
    return FR_OK;
  }
  
fresp FlatFile::setheader(Memory &in)
  {
    // returns FR_*
    // in must be of length headersize
    FlatPacket fp;
    fresp ret;
    ULONG d;
    
    fp.command = FP_SETHEADER;
    fp.fileid = id;
    szfix(in, headersize);
    ret = request(fp, d, in, headersize);
    return ret;
  }

/* Record based */      
fresp FlatFile::readrec(long recnoin, Memory &out)
  {
    // returns FR_* 
    FlatPacket fp;
    Memory x; // no data
    ULONG d;
    
    fp.command = FP_READREC;
    fp.fileid = id;
    fp.recno = recnoin;
    if (request(fp, d, x) != FR_OK)
      return FR_FAIL;

    // retrieve command specific data here.
    int r = fc.getsco()->getstr(out);
    if (r == FALSE)
      return FR_FAIL;
    szfix(out, recordsize);
    return FR_OK;
  }
  
fresp FlatFile::writerec(long recnoin, Memory &in)
  {
    // returns FR_* 
    FlatPacket fp;
    fresp ret;
    ULONG d;
    
    fp.command = FP_WRITEREC;
    fp.fileid = id;
    fp.recno = recnoin;
    szfix(in, recordsize);
    ret = request(fp, d, in, recordsize);
    return ret;
  }
  
fresp FlatFile::deleterec(long recnoin)
  {
    // returns FR_* 
    FlatPacket fp;
    fresp ret;
    Memory x;
    ULONG d;
    
    fp.command = FP_DELETEREC;
    fp.fileid = id;
    fp.recno = recnoin;
    ret = request(fp, d, x);
    return ret;
  }
  
/* index based */
fresp FlatFile::iseek(char *keyin, Memory &out, int keyinlen)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;

    Memory k;
    keyize(keyin, k, keyinlen);
    
    fp.command = FP_ISEEK;
    fp.fileid = id;
    if (keyinlen == 0) 
      keyinlen = keysize;
    ret = request(fp, d, k, keyinlen);
    if (ret == FR_NOTFOUND)
      return ret;
    if (ret != FR_OK)
      return ret;

    // retrieve command specific data here.
    int r = fc.getsco()->getstr(out); // being Memory, it will be raw data not a \0 string
    if (r == FALSE)
      return FR_FAIL;
    szfix(out, recordsize);
    return FR_OK;
  }

fresp FlatFile::iselect(char *keyin, zlist &list, int keyinlen)
  {
    // 1/12/01
    /* This is a neat one. Since returning a range of rows requires
       a lot of expensive hits, we have a select function.
       Given a key (partial key expected) it will return a zlist of
       stringentrys, each one a record from the database matching
       that key.
       Exception is a keylength of zero, matches everything. It returns
       the whole database. */
       
    FlatPacket fp;
    fresp ret;

    /* We don't call keyize because it treats keylen=0 special
       and in this case, that's valid. If you wanna pass a string
       you have to send us the length */
    list.WipeList();

    fp.command = FP_ISELECT;
    fp.fileid = id;

    ret = fc.sendfp(fp);
    if (ret == FALSE)
      return FR_FAIL;
    fc.getsco()->putint(keyinlen); // 0 is valid
    if (keyinlen > 0)
      {
        int r = fc.getsco()->sendstr(keyin, keyinlen); // should validate this
        if (r == FALSE)
          return FR_FAIL;
      }
    
    // get response packet.
    FlatResponse fr;
    ret = fc.getfr(fr);
    if (ret != FR_OK)
      return FR_FAIL;
    ret = fr.rc;
    if (ret == FR_NOTFOUND)
      return ret;
    if (ret != FR_OK)
      return ret;

    /* retrieve a length, and a string. If the length is 0,
       that's the end. */
    int sz = fc.getsco()->getint();
    while (sz > 0)
      {
        Memory dt;
        int r = fc.getsco()->getstr(dt);
        if (r == FALSE)
          return FR_FAIL;
        szfix(dt, recordsize);
        stringentry *st = new stringentry(dt); // being a Memory object will cause a raw memory copy, not a \0 string copy
        list.Add(st);
        sz = fc.getsco()->getint();
      }
    return FR_OK;
  }
  
fresp FlatFile::iupdate(char *keyin, Memory &in)
  {
    // in must be of at least recordsize length
    FlatPacket fp;
    fresp ret;
    ULONG d;

    Memory k;
    keyize(keyin, k, keysize);    
    
    fp.command = FP_IUPDATE;
    fp.fileid = id;
    szfix(in, recordsize);
    ret = request(fp, d, k, keysize, in, recordsize);
    return ret;
  }

fresp FlatFile::iatomicupdate(char *keyin, Memory &in)
  {
    // in must be of at least recordsize length
    FlatPacket fp;
    fresp ret;
    ULONG d;

    Memory k;
    keyize(keyin, k, keysize);    
    
    fp.command = FP_IATOMICUPDATE;
    fp.fileid = id;
    szfix(in, recordsize);
    ret = request(fp, d, k, keysize, in, recordsize);
    return ret;
  }
  
fresp FlatFile::iinsert(Memory &in)
  {
    // in must be of at least recordsize length
    FlatPacket fp;
    fresp ret;
    ULONG d;
    
    fp.command = FP_IINSERT;
    fp.fileid = id;
    szfix(in, recordsize);
    ret = request(fp, d, in, recordsize);
    return ret;
  }

fresp FlatFile::iatomicinsert(Memory &in)
  {
    // this is a combination, seek and insert.
    // if the seek succeeds, we return duplicate and don't insert
    FlatPacket fp;
    fresp ret;
    ULONG d;
    
    fp.command = FP_IATOMICINSERT;
    fp.fileid = id;
    szfix(in, recordsize);
    ret = request(fp, d, in, recordsize);
    return ret;
  }

fresp FlatFile::idelete(char *keyin, int keyinlen)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;

    Memory k;
    keyize(keyin, k, keyinlen);    
    
    fp.command = FP_IDELETE;
    fp.fileid = id;
    if (keyinlen == 0)
      keyinlen = keysize;
    ret = request(fp, d, k, keyinlen);
    return ret;
  }

/* Record based operations based on index */
long FlatFile::firstrec(void)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;
    Memory x;

    fp.command = FP_FIRSTREC;
    fp.fileid = id;
    ret = request(fp, d, x);
    if (ret != FR_OK)
      return -1;
    return d;
  }
  
long FlatFile::lastrec(void)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;
    Memory x;

    fp.command = FP_LASTREC;
    fp.fileid = id;
    ret = request(fp, d, x);
    if (ret != FR_OK)
      return -1;
    return d;
  }

long FlatFile::seekrec(char *keyin, int keyinlen)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;

    Memory k;
    keyize(keyin, k, keyinlen);    
    
    fp.command = FP_SEEKREC;
    fp.fileid = id;
    if (keyinlen == 0)
      keyinlen = keysize; // set to file format, use specified length if it was sent
    ret = request(fp, d, k, keyinlen);
    if (ret != FR_OK) // could be not found or error
      return -1;
    return d;
  }

long FlatFile::nextrec(long recin)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;
    Memory x;

    fp.command = FP_NEXTREC;
    fp.fileid = id;
    fp.recno = recin;
    ret = request(fp, d, x);
    if (ret != FR_OK)
      return -1;
    return d;
  }

long FlatFile::prevrec(long recin)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;
    Memory x;

    fp.command = FP_PREVREC;
    fp.fileid = id;
    fp.recno = recin;
    ret = request(fp, d, x);
    if (ret != FR_OK)
      return -1;
    return d;
  }

/* Blobs */
fresp FlatFile::bseek(char *keyin, Memory &out, long &sizeout)
  {
    // 1/9/01 you have to know what the keylength is  and set keyin to it
    // the size of the blob is return in resp data recno
    FlatPacket fp;
    fresp ret;
    ULONG d;

    Memory k;
    keyize(keyin, k, keysize); // 1/9/01 must match full keylength
    
    fp.command = FP_BSEEK;
    fp.fileid = id;
    ret = request(fp, d, k, keysize); 
    if (ret == FR_NOTFOUND)
      return ret;
    if (ret != FR_OK)
      return ret;

    // retrieve command specific data here.
    // caller should deal with the size it gets back by asking, we should put this in protocol
    int r = fc.getsco()->getstr(out);
    if (r == FALSE)
      return FR_FAIL;
    sizeout = d; // size is returned here
    return FR_OK;
  }

fresp FlatFile::binsert(char *keyin, Memory &in, long sizein)
  {
    // returns FR_* 
    FlatPacket fp;
    fresp ret;
    ULONG d;
    
    Memory k;
    keyize(keyin, k, keysize);    
    fp.command = FP_BINSERT;
    fp.fileid = id;
    fp.recno = sizein; // hack hack
    szfix(in, sizein);
    ret = request(fp, d, k, keysize, in, sizein);
    return ret;
  }

fresp FlatFile::batomicinsert(char *keyin, Memory &in, long sizein)
  {
    // returns FR_*, just like above, but server does atomic duplicate check first
    FlatPacket fp;
    fresp ret;
    ULONG d;
    
    Memory k;
    keyize(keyin, k, keysize);
    fp.command = FP_BATOMICINSERT;
    fp.fileid = id;
    fp.recno = sizein; // hack hack
    szfix(in, sizein);
    ret = request(fp, d, k, keysize, in, sizein);
    return ret;
  }

fresp FlatFile::bdelete(char *keyin)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;

    Memory k;
    keyize(keyin, k, keysize);    
    
    fp.command = FP_BDELETE;
    fp.fileid = id;
    ret = request(fp, d, k, keysize);
    return ret;
  }

/* System */
fresp FlatFile::reindex(void)
  {
    FlatPacket fp;
    fresp ret;
    ULONG d;
    Memory x;
    
    fp.command = FP_REINDEX;
    fp.fileid = id;
    ret = request(fp, d, x);
    return ret;
  }

fresp FlatFile::locate (zlist &listout, char *crit1, int pos1, int len1, int type1)
  {
    return FR_NOTIMPLEMENTED;
  }



/*                              Flat Connection                             */

FlatConnection::FlatConnection (socketioref *ssco)
  {
    // takes logical ownership. When this goes out of 
    // scope, it sends a close command to the server and
    // closes the socket.
    sco = ssco;
    if (sco == NULL)
      return;
    sco->refcount++;
  }

FlatConnection::FlatConnection (const FlatConnection &sref)
  {
    FlatConnection &ref = (FlatConnection &)sref;
    sco = ref.getsco();
    if (sco == NULL)
      return;
    sco->refcount++;
  }

FlatConnection &FlatConnection::operator=(const FlatConnection &sref)
  {
    FlatConnection &ref = (FlatConnection &)sref;
    if (this == &ref)
      return *this;

    if (sco != NULL) // dec old one if there was
      sco->refcount--;
      
    sco = ref.getsco();
    if (sco == NULL)
      return *this;
    sco->refcount++;
    return *this;
  }

FlatConnection::~FlatConnection(void)
  {
    if (sco == NULL)
      return;
    sco->refcount--;
    if ((sco != NULL) && (sco->refcount == 0)) // is this the last pointer to this socket?
      {
        // send a close socket command, tell server to hang up
        FlatPacket fp;
        fp.command = FP_CLOSECONN;
        sendfp(fp);
        delete sco;
      }
  }

void FlatConnection::addfile(FlatFile &nf)
  {
    // add this to our zlist of things.
    openfileentry *e = new openfileentry (nf);
    openfilelist.Add(e);
  }
  
FlatFile *FlatConnection::findfile(char *filename)
  {
    // look for it in our zlist, return NULL if not found.
    int lp;
    int max = openfilelist.ListCount();
    for (lp = 0; lp < max; lp++)
      { 
        openfileentry *e = (openfileentry *)openfilelist.GetItem(lp);
        FlatFile &me = e->getfile();
        if (strcmp(me.getfilename(), filename) == 0)
          return &me;
      }
    return NULL; // not found.
  }

FlatFile FlatConnection::Open(char *filename, int headersize, int recordsize, int index, int keysize)
  {
    // factory a flatfile for this file.

    /* Here's my cheap and cheezy. 3/5/2001. So save LOTS of IO to the fserver
       when we get an open request, we ask our connection if they have 
       that file open already or not. If so, just return a pointer to it
       (ref counts already handled) if not, open the file and tell the
       flat connection about. The final close will have to tell
       the flat connection to get rid of it. */
    FlatFile *exists = findfile(filename);
    if (exists != NULL) 
      return *exists;

    // make a bad object so we can return it upon error
    FlatFile fbad(*this, "bad", headersize, recordsize, index, keysize, -1);

    /* this is a bit more complex than FlatFile::request can handle, so we do
       ourselves here. */
    FlatPacket fp;
    fbad.seterr(FR_COMMERROR);
 
    fp.command = FP_OPENFILE;
    fresp ret = sendfp(fp);
    if (ret != FR_OK)
      return fbad;
    int r;
    r = sco->sendstr(filename);
    r &= sco->putint(headersize);
    r &= sco->putint(recordsize);
    r &= sco->putint(index);
    r &= sco->putint(keysize);
    if (r == FALSE)
      return fbad;
      
    // get response packet.
    FlatResponse fr;
    ret = getfr(fr);
    if (ret != FR_OK)
      return fbad;

    fbad.seterr(fr.rc); // file is in good status or not...
    if (fr.rc != FR_OK)
      return fbad;
    int sid = fr.data.id;

    FlatFile fx(*this, filename, headersize, recordsize, index, keysize, sid);
    fx.seterr(FR_OK); // denote a good file status.
    addfile(fx);
    return fx;
  }
  
fresp FlatConnection::sendfp(FlatPacket &fp)
  {
    // if you want to flip endian, this is the place
    fp.protocolid = PROTOCOL_ID;
    int ret = sco->senddata(&fp, sizeof(FlatPacket));
    if (ret == FALSE)
      return FR_FAIL;
    return FR_OK;
  }

fresp FlatConnection::getfr(FlatResponse &fr)
  {
    // if you want to flip endian, this is the place
    int ret;
    ret = sco->getdata(&fr, sizeof(FlatResponse));
    if (ret == FALSE)
      return FR_FAIL;
    if (fr.protocolid != PROTOCOL_ID)
      return FR_PROTOCOLERROR;
    return FR_OK;
  }

/*                              Flat Database                             */
  
FlatDatabase::FlatDatabase(char *sserver, int sport, int stimeout)
  {
    server = sserver;
    port = sport;
    timeout = stimeout;
  }

FlatDatabase::FlatDatabase(FlatDatabase &fd)
  {
    server = fd.getserver();
    port = fd.getport();
    timeout = fd.gettimeout();
  }

FlatDatabase &FlatDatabase::operator=(const FlatDatabase &sfd)
  {
    FlatDatabase &fd = (FlatDatabase &)sfd;
    if (this == &fd)
      return *this;
    server = fd.getserver();
    port = fd.getport();
    timeout = fd.gettimeout();
    return *this;
  }
  
FlatDatabase::~FlatDatabase(void)
  {
    // the presumption is that everything else goes out
    // of scope before this does.
  }
  
FlatConnection FlatDatabase::Connect(void)
  {
    // Make a connection
    Connector cn(server, port, timeout);
    if (cn.errorstate() != 0)
      {
        FlatConnection fc(NULL);
        return fc;
      }
    socketioref *sco = new socketioref(cn.victim(), cn.victim(), timeout);
    FlatConnection fc(sco);
    return fc;
  }
      
