/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* The flat file system record manager client interface. */

#ifndef FC_H
#define FC_H

#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/connector.h>
#include <Q/socketio.h>

#include <flat/fpack.h>

/* You should ask the Database for Connections, All connections are
   persistent. Then you make file references from that connection.
   Then you do operations on that file and it will go down that
   persistent connection or make one on the fly.
   I'm not sure if I'm going to use FlatHandle to lock a connection.
*/
   
class FlatDatabase;
class FlatConnection;
class FlatFile;

/* For making and extracting string key fields */
inline void stringtokey (char *out, char *in, int keylen)
  {
    // make a perfect length space padded key from in to out
    Memory s(in);
    pad (s, keylen, ' '); // too short? 
    memcpy (out, s, keylen);
  }

inline void stringtokeyCASE (char *out, char *in, int keylen)
  {
    // make a perfect length space padded key from in to out and upcase it
    Memory s(in);
    updo (s);
    pad (s, keylen, ' '); // too short? 
    memcpy (out, s, keylen);
  }


/* Hmmm... I need a argh, smart pointer, ref count to the sco as they pass around
   connections. So I subclass socketio, and use that instead */
   
class socketioref : public socketio
  {
    public:
      int refcount;
      socketioref(int Infd, int Outfd, int timeout = 0, int appropriate = TRUE) : 
                  socketio(Infd, Outfd, timeout, appropriate) { refcount = 0; }
  };

class refcounter // for other classes needing refcounts
  {
    public:
      int refcount;
      refcounter(void) { refcount = 0; }
  };

class FlatDatabase
  {
    public:
      FlatDatabase(void) { port = 0; timeout = 0; }
      FlatDatabase(char *server, int port, int timeout);
      FlatDatabase(FlatDatabase &); // cc
      FlatDatabase &operator=(const FlatDatabase &);
      ~FlatDatabase(void);
      FlatConnection Connect(void);
      Memory &getserver(void) { return server; }
      int getport(void) { return port; }
      int gettimeout(void) { return timeout; }
      
    private:
      Memory server;
      int port;
      int timeout;
  };

class FlatConnection
  { // this get factoried from FlatDatabase
    public:
      FlatConnection(void) { sco = NULL; }
      FlatConnection(socketioref *sco);
      ~FlatConnection(void);
      FlatConnection(const FlatConnection &); // copy constructor, will it do =?
      FlatConnection &operator=(const FlatConnection &);
      FlatFile Open(char *filename, int headersize, int recordsize, int index = FALSE, int keysize = 0);
      
      socketioref *getsco(void) { return sco; }
      void setsco(socketioref *ssco) { sco = ssco; }
      fresp sendfp(FlatPacket &fp);
      fresp getfr(FlatResponse &fr);
      
    private:
      void addfile(FlatFile &nf);
      FlatFile *findfile(char *filename);
      
      socketioref *sco; 
      zlist openfilelist; // warning, no copy constructor...
  };

class FlatFile
  {
    public:
      FlatFile (void);
      FlatFile (FlatConnection &fc, char *fname, int headersize, int recordsize, int hasindex, int keysize, int sid);
      FlatFile (const FlatFile &); // copy constructor
      FlatFile &operator=(const FlatFile &);
      ~FlatFile (void); // this will send a close command of this file id

      /* The operations you can do on a file. */
      // you should know the header length
      fresp getheader(Memory &out);
      fresp setheader(Memory &in);

      /* Record based */      
      fresp readrec(long recnoin, Memory &out);
      fresp writerec(long recnoin, Memory &in);
      fresp deleterec(long recnoin);
      
      /* Table scans. */
      fresp locate (zlist &listout, char *crit1, int pos1, int len1, int type1);
      
      /* index based */
      fresp iseek(char *keyin, Memory &out, int keylen = 0);
      fresp iupdate(char *keyin, Memory &in);
      fresp iinsert(Memory &in);
      fresp iatomicinsert(Memory &in);
      fresp iatomicupdate(char *keyin, Memory &in); // 1/2/03 how late am I on this?
      fresp idelete(char *keyin, int keylen = 0);
      fresp iselect(char *keyin, zlist &list, int keyinlen); // no default 0 is special
      
      /* Record based operations based on index */
      long firstrec(void);
      long lastrec(void);
      long seekrec(char *keyin, int keylen = 0);
      long nextrec(long recin);
      long prevrec(long recin);
      
      /* Blobs */
      /* You must supply full key length to all of these. */
      fresp bseek(char *keyin, Memory &out, long &sizeout);
      fresp binsert(char *keyin, Memory &in, long sizein);
      fresp batomicinsert(char *keyin, Memory &in, long sizein);
      fresp bdelete(char *keyin);

      /* System */
      fresp reindex(void);
      void seterr(int e) { err = e; }
      char *getfilename(void) { return filename; }
      
    private:
      fresp request(FlatPacket &fp, ULONG &dout, Memory &data, int len = -1, Memory data2 = "", int len2 = -1);
      void keyize (char *keyin, Memory &keyout, int keysize);
      void szfix(Memory &me, UINT size); // similar to keysize, but does data field
      void finalclose(void);

      FlatConnection fc; // to get at server. 
      int id; // assigned by server, -1 is invalid flag
      int err;
      refcounter *ref; // always point to one instance of this FFile
      
      Memory filename;
      int headersize;
      int recordsize;
      int hasindex;
      int keysize;
  };

class openfileentry : public entry
  {
    public:
      openfileentry (FlatFile &f) { ff = f; } // keep a REAL copy 
      FlatFile &getfile(void) { return ff; }

    private:
      FlatFile ff; 
  };


#endif

