/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/connector.h>
#include <Q/notecraft.h>
#include <Q/configcraft.h>
#include <Q/socketio.h>
#include <Q/acceptor.h>
#include <Q/mutex.h>

#include <flat/fpack.h>
#include <flat/flat.h>

#include "fserver.h"

/* Here's the idea. We have a zary, filled with FFilePtrs. Each one is a file 
   FLATHAND, points to a connection that it came in on (so we can clean them all 
   at once) When we get a new connection, we just need to keep a high watermark of
   incoming sockets. The client will take care of them being open or close. Hmm..
   room for leaks there... Hopefully we'll get an error on a bad socket and
   clean it up that way. */

/* 11/5/02 And now we have our multi-read-single-write mutex class, 
   we can run multiple of these in parrallel, when opening a handle
   for a file, use the same mutex id for all kids. */
   
#define TXT_NAME "fserver"

int main(int num, char *opts[])
  {
    ConfigCraft ar(TXT_NAME, num, opts); // will read test.cf unless cmd line overrides.
    NoteCraft   ls(TXT_NAME, ar[TXT_LOGFILE], val(ar[TXT_LOGLEVEL]));

    /* If it's important enough to be in the flatserver, it's got 600 perms. */
    umask(0066); // octal for rw-------

    ar.adddefault(TXT_BINDADDR, "127.0.0.1"); // be safe
    
    Memory bindaddr = ar[TXT_BINDADDR];
    ls(NOTEINF) << "Kids wanted " << ar[TXT_KIDS] << endl;
    
    Acceptor ac(TXT_NAME, val(ar[TXT_PORT]), opts, ls, bindaddr); // make a socket and listen, bind all
    if (ac.errorstate()) // check for error
      {
        ls (NOTESEV) << "Unable to make acceptor socket: " << ac.errstr() << endl;
        return (1);
      }
    ls(NOTEINF) << "Listening on port: " << ar[TXT_PORT] << endl;
    ls(NOTEINF) << "Build " __DATE__ << " " << __TIME__ << endl;

    ac.sethostsallowlist(ar[TXT_HOSTS_ALLOW]);
    KidHandler kd(val(ar[TXT_KIDS]), ar, ac, runner); // 11/5/02
    kd.run();
    return 0;
  }

int runner (int num, ConfigCraft &ar, Acceptor &ac)
  { /* One of these gets run per process. */
    NoteCraft &ls = ac.getNC();
    
    FlatServer go(ar, ac, ls, num);
    return go.run();
  }

FlatServer::FlatServer(ConfigCraft &Ar, Acceptor &Ac, NoteCraft &Ls, int num) : ar(Ar), ls(Ls), ac(Ac)
  {
    tmout = val(ar[TXT_SOCKETTIMEOUT]);
  }
  
int FlatServer::run(void)
  {
    /* The problem with the old way is that since the listener
       socket is always accepting, the OS never hands off connections
       to other processes and you end up single lining anyway. 
       Making this more like all my other servers */

    int failures = 0;
    while (1) // loop for each session
      {
        SOCKET inout;
        inout = ac.accept(); /* block forever until we get a connection */
        if (inout == -1)
          {
            ls(NOTESEV) << "Error on accept: " << ac.errstr() << endl;
            sleep(5);
            failures++;
            if (failures >= 5)
              return 1; // restart
            continue; // got any other ideas?
          }
          
        // we never want to time out, we'll wait forever.
        // socketio can't do that, so we wait for a really really long time
        socketio Sco (inout, inout, -1); // takes ownership of the socket/FD
        sco = &Sco;
        ls (NOTEINF) << "Session starting." << endl;
        startsession();
        ls (NOTEINF) << "Session ending." << endl;
        ls (NOTEINF) << "Memleak: " << memleak << endl;
      } // closes socket, start a new session.
    
    return 0;
  }

void FlatServer::startsession(void)
  {
    // read and process until exit
    int ret = TRUE;
    while (ret)
      {
        ret = process_request();
      }
  }

int FlatServer::process_request(void)
  {
    /* This is the command handler for a flat client request.
    
       We may need to close files, open files, close connections, 
       all are handled here. 
       This function will return FALSE if we're closing
       the connection, and TRUE otherwise
    */
        
    /* First part of protocol is always a flatpacket */
    FlatPacket fp;
    fresp ret = getfp(*sco, fp);
    if (ret == FR_FAIL)
      if (check_connection() == FALSE)
        return FALSE; // end this session.
 
    FlatResponse fr;
    ls(NOTEDIA) << "Command: " << fp.command << endl;
    
    switch (fp.command)
      {
        case FP_CLOSECONN:
          closeconn(); // close all files on this connection
          return FALSE;
          
        case FP_OPENFILE:
          fr.rc = openfile(*sco, fr);
          break;
          
        case FP_CLOSEFILE:
          fr.rc = closefile(*sco, fp);
          break;
          
        case FP_GETHEADER:
          FSgetheader(*sco, fp); // does its own responding
          return TRUE;
          
        case FP_SETHEADER:
          FSsetheader(*sco, fp); // does its own responding
          return TRUE;
          
        case FP_READREC:
          FSreadrec(*sco, fp);
          return TRUE;
          
        case FP_WRITEREC:
          FSwriterec(*sco, fp);
          return TRUE;
          
        case FP_DELETEREC:
          FSdeleterec(*sco, fp);
          return TRUE;
          
        case FP_ISEEK:
          FSseekrec(*sco, fp);
          return TRUE;
          
        case FP_IUPDATE:
          FSupdate(*sco, fp);
          return TRUE;
          
        case FP_IATOMICUPDATE:
          FSatomicupdate(*sco, fp);
          return TRUE;
          
        case FP_IINSERT:
          FSinsert(*sco, fp);
          return TRUE;
          
        case FP_IDELETE:
          FSdelete(*sco, fp);
          return TRUE;

        case FP_ISELECT:
          FSselect(*sco, fp);
          return TRUE;
          
        case FP_FIRSTREC:
          FSfirstrec(*sco, fp);
          return TRUE;
          
        case FP_LASTREC:
          FSlastrec(*sco, fp);
          return TRUE;
          
        case FP_SEEKREC:
          FSseekrecno(*sco, fp);
          return TRUE;
          
        case FP_NEXTREC:
          FSnextrec(*sco, fp);
          return TRUE;
          
        case FP_PREVREC:
          FSprevrec(*sco, fp);
          return TRUE;

        case FP_BSEEK:
          FSBseek(*sco, fp);
          return TRUE;

        case FP_BINSERT:
          FSBinsert(*sco, fp);
          return TRUE;
          
        case FP_BDELETE:
          FSBdelete(*sco, fp);
          return TRUE;
          
        case FP_REINDEX:
          FSreindex(*sco, fp);
          return TRUE;
          
        case FP_IATOMICINSERT :
          FSatomicinsert(*sco, fp);
          return TRUE;
          
        case FP_BATOMICINSERT:
          FSBatomicinsert(*sco, fp);
          return TRUE;
          
        default:
          ls(NOTEERR) << "Unhandled command packet: " << (int) fp.command << endl;
          fr.rc = FR_NOTIMPLEMENTED;
          break;
      }
    ret = sendfr(*sco, fr); // check for errors on this?
    if (ret == FR_FAIL)
      if (check_connection() == FALSE)
        return FALSE;
    return TRUE;
  }
  
fresp FlatServer::getfp(socketio &sco, FlatPacket &fp)
  {
    // if you want to flip endian, this is the place
    int ret;
    ret = sco.getdata(&fp, sizeof(FlatPacket));
    if (ret == FALSE)
      return FR_FAIL;
    if (fp.protocolid != PROTOCOL_ID)
      return FR_PROTOCOLERROR;
    return FR_OK;
  }

fresp FlatServer::sendfr(socketio &sco, FlatResponse &fr)
  {
    // if you want to flip endian, this is the place
    fr.protocolid = PROTOCOL_ID;
    int ret = sco.senddata(&fr, sizeof(FlatResponse));
    if (ret == FALSE)
      return FR_FAIL;
    return FR_OK;
  } 
  
int FlatServer::check_connection(void)
  {
    /* This is the fun and neat one. We've got an error.
       We check errno. If there is an error that inferrs connection
       loss, we return FALSE, else we return TRUE;
       This is the flag that we should end the session on error
    */
    int err = sco->errorstate();
    if (err == 0)
      {
        ls(NOTEERR) << "check connection with errno == 0" << endl;
        return TRUE;
      }
    ls (NOTEERR) << "Error on socket: (" << err << ") " << strerror(err) << endl;
    closeconn(); // free up the files
    return FALSE;
  }

int FlatServer::closeconn(void)
  {
    // close all files using this connection.
    // we don't send a response, duh
    ls (NOTEINF) << "Closing all files on connection" << endl;
    int max = fileary.MaxItems();
    int lp;
    for (lp = 0; lp < max; lp++)
      {
        FFilePtr *f = (FFilePtr *)fileary.GetItem(lp);
        if (f != NULL) // there can be holes
          fileary.DelPos(lp); // deletes pointer which decs refcount which cl
      }
    return 0;
  }
  
fresp FlatServer::openfile(socketio &sco, FlatResponse &fr)
  {
    /* See if we already have this file open, if so, make another
       pointer to it, and add to the list. If not, make a
       new flathand file and pointer and add to the list. */
    // Get params.
    
    int ret;
    Memory fn;
    ULONG headersize, recordsize, index, keysize;
    ret = sco.getstr(fn);
    ret &= sco.getint(headersize);
    ret &= sco.getint(recordsize);
    ret &= sco.getint(index);
    ret &= sco.getint(keysize);
    if (ret == FALSE)
      {
        check_connection();
        return FR_FAIL;
      }
    ls(NOTEDIA) << "Open: " << (char *)fn << endl;
    // look through our list
    int max = fileary.MaxItems();
    int lp;
    for (lp = 0; lp < max; lp++)
      {
        FFilePtr *f = (FFilePtr *)fileary.GetItem(lp);
        if (f != NULL) // there can be holes
          {
            FFile *fptr = f->getffile();

            /* Here we look for the file by name to see if it already
               exists. If we find a match, we use the same mutex, so if
               you get at the same file with a different name the
               serialized reading-writing will not work. You have
               been warned. */
            if (strcmp(fptr->getfname(), fn) == 0)
              {
                // make a new pointer reference to what's there
                FFilePtr *np = new FFilePtr(fptr, sco.getfd());
                int pos = fileary.Add(np);
                fr.data.id = pos; // very important this gets passed back. it's our future
                return FR_OK;
              }
         }
      }
    // If we're here, open a new file.
    
    FLATHAND fh;
    fh = flatopen(fn, headersize, recordsize, index, keysize);
    if (fh == NULL)
      { 
        ls(NOTEERR) << "Unable to open file: " << (char *)fn << " : " << errno << " " << strerror(errno) << endl;
        return FR_FAIL;
      }
    // if we got it, add it to the list
    FFile *ff = new FFile(fn, fh, headersize, recordsize, index, keysize);
    // check mutex create
    if (ff->getfh() == NULL) // error on mutex create
      {
        ls(NOTEERR) << "Unable to create mutex for file: " << (char *)fn << " : " << errno << " " << strerror(errno) << endl;
        delete ff;
        return FR_FAIL;
      }
    FFilePtr *np = new FFilePtr(ff, sco.getfd());
    int pos = fileary.Add(np);
    // set the array index to ID
    fr.data.id = pos; // very important this gets passed back. it's our future
    return FR_OK;
  }
          
fresp FlatServer::closefile(socketio &sco, FlatPacket &fp)
  {
    /* Delete this guys pointer, the cascading deletes will
       close the file if necessary */
    int id = fp.fileid;

    ls(NOTEDIA) << "Close file" << endl;
    if (fileary.GetItem(id) == NULL)
      {
        ls(NOTEERR) << "Trying to close a closed file id: " << id << endl;
        return FR_FAIL;
      }
      
    fileary.DelPos(id); // ary takes care of errant nulls. 
    return FR_OK;
  }
                                                                           
FLATHAND FlatServer::getfile(FlatPacket &fp, FFilePtr *&sffp)
  {
    int id = fp.fileid;
    if ((id < 0) || (id > fileary.MaxItems()))
      {
        ls(NOTEERR) << "Invalid file id: " << id << endl;
        return NULL;
      }
    FFilePtr *ffp = (FFilePtr *)fileary.GetItem(id);
    if (ffp == NULL)
      {
        ls(NOTEERR) << "File ID is null: " << id << endl;
        return NULL;
      }
    sffp = ffp;
    ls(NOTEDIA) << "Getfile: " << (char *)ffp->getffile()->getfname() << endl;
    FLATHAND x = ffp->getfh();
    return x;
  }

void FlatServer::szfix(Memory &me, UINT size)
  {
    /* make this m match the size */
    /* Actually, we only have to make it longer if it ends up being too short. */
    char *m = me;
    ULONG sz;

    memcpy (&sz, m-4, 4); /* get long back */
    if (sz >= size)
      return;
    int add = size - sz;
    Memory addstr(add+1); // room for nt
    memset (addstr, ' ', add);
    *(addstr+add) = '\0';
    me.addstr(addstr); // pad with spaces
  }
  
void FlatServer::FSgetheader(socketio &sco, FlatPacket &fp)
  {
    // first we send an FP if it looks like we can do this,
    // then we send the data
    FlatResponse fr;
    FFilePtr *ffp;
    
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Get header." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    // go get the header and send it back...
    Memory out(ffp->getheadersize());
    int ret;
    { MutexLockRead x(ffp->getlock()); // read lock the read
      ret = getheader(fh, out);
    }
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for getheader fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    ret &= sco.sendstr(out, ffp->getheadersize()); // with size so other end knows how much to accept
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for getheader success." << endl;
  
    return;    
  }

void FlatServer::FSsetheader(socketio &sco, FlatPacket &fp)
  {
    // first we send an FP if it looks like we can do this,
    // then we send the data
    FlatResponse fr;

    Memory in;
    sco.getint(); // length of key is sent, we don't use it.
    int ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get header string." << endl;
        return;
      }

    FFilePtr *ffp;
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Set header." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    szfix(in, ffp->getheadersize());
    { MutexLockWrite x(ffp->getlock()); // write lock the write
      ret = setheader(fh, in);
    }
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for setheader fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for setheader success." << endl;
  
    return;    
  }

void FlatServer::FSreadrec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Read rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    Memory out(ffp->getrecordsize());

    int ret;
    { MutexLockRead x(ffp->getlock());
      ret = flatread(fh, fp.recno, out);
    }
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for readrec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    ret &= sco.sendstr(out, ffp->getrecordsize()); // with size so other end knows how much to accept
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for getheader success." << endl;
  
    return;    
  }

void FlatServer::FSwriterec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;

    Memory in;
    int ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get writerec string." << endl;
        return;
      }

    FFilePtr *ffp;
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Write rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    szfix(in, ffp->getrecordsize());
    
    { MutexLockWrite x(ffp->getlock());
      ret = flatwrite(fh, fp.recno, in);
    }
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for writerec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for writerec success." << endl;
  
    return;    
  }

void FlatServer::FSdeleterec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;

    FFilePtr *ffp;
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Delete rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    int ret;
    { MutexLockWrite x(ffp->getlock());
      ret = flatreaddel(fh, fp.recno);
    }
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for deleterec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for writerec success." << endl;
  
    return;    
  }
   
void FlatServer::FSseekrec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    Memory key;
    UINT keylen = sco.getint(); // 1/9/01 // get 'key' length 
    int ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Seek rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen);
    
    Memory out(ffp->getrecordsize());

    { MutexLockRead x(ffp->getlock());
      ret = flatseek(fh, key, keylen, out);
    }
    if (ret != 0)
      {
        if (ret == 2)
          fr.rc = FR_NOTFOUND;
        else
          fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for seekrec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    ret &= sco.sendstr(out, ffp->getrecordsize()); // with size so other end knows how much to accept
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for seekrec success." << endl;
  
    return;    
  }

void FlatServer::FSselect(socketio &sco, FlatPacket &fp)
  {
    /* 1/12/01 We get a keylength, and a key. If the keylength is 0
       then we match all records. Otherwise, we seek on the key
       and keep going next until no more, returning a length
       and the record string ending in a zero length. */
       
    FlatResponse fr;
    FFilePtr *ffp;
    
    Memory key;
    UINT keylen = sco.getint(); // 1/9/01 // get 'key' length 
    int ret = 1;
    if (keylen > 0)
      ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Select rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen);
    
    Memory out(ffp->getrecordsize());
    int lp;

    { MutexLockRead x(ffp->getlock()); // lock whole select as atomic operation
    
    if (keylen == 0) // get first record
      lp = firstrec(fh);
    else // seek to first key matching record
      lp = flatseekrec(fh, key, keylen);        

    ret = -1;  
    if (lp != -1)
      ret = flatread(fh, lp, out);
    if (ret != 0)
      lp = (UINT)-1;
    
    if (lp == -1)
      {
        fr.rc = FR_NOTFOUND; // or bad, what do we know.
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for seekrec fail." << endl;
        return;
      }
      
    fr.rc = FR_OK;
    ret = sendfr(sco, fr); // send back the okay
    if (ret != FR_OK)
      {
        ls(NOTEERR) << "Can't send response to client for sendfr fail." << endl;
        return;
      }
    // now start whipping back records.
    while (1)
      {
        ret = sco.putint(1); // just has to be non-zero
        if (ret != FR_OK)
          {
            ls(NOTEERR) << "Can't send response to client for putrec fail." << endl;
            return;
          }
        ret &= sco.sendstr(out, ffp->getrecordsize()); // with size so other end knows how much to accept
        if (ret != FR_OK)
          {
            ls(NOTEERR) << "Can't send response to client for sendstr fail." << endl;
            return;
          }
        lp = nextrec(fh, lp);
        if (lp == -1)
          break;
        ret = flatread(fh, lp, out);
        if (ret != 0)
          {
            ls(NOTEERR) << "Can't send response flatread returned bad: " << ret << endl;
            break; // send a "that's it"
          }
        // see if they key still matches
        if (keylen == 0)
          continue; // don't check, do everything
        // hate to do this, it's cheating, but I don't wanna expose
        // that deep function in the api just for this...
        if (memcmp(out, key, keylen) != 0)
          break; // we're done
      }
    } // mutex read lock
    ret &= sco.putint(0); // we outta here.
    return;    
  }

void FlatServer::FSupdate(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;

    Memory key;
    Memory in;
    UINT keylen = sco.getint(); // 1/9/01
    int ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }
    sco.getint(); // 1/9/01 length not needed here
    ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get update string." << endl;
        return;
      }

    FFilePtr *ffp;
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Update rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen); // 1/9/01
    
    szfix(in, ffp->getrecordsize());

    { MutexLockWrite x(ffp->getlock());
      ret = flatupdate(fh, key, in);
    }
    if (ret != 0)
      {
        if (ret == 2)
          fr.rc = FR_NOTFOUND;
        else
          fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for update fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for update success." << endl;
  
    return;    
  }

void FlatServer::FSatomicupdate(socketio &sco, FlatPacket &fp)
  {
    /* I just noticed this horribly missing. It was probably missing
       a long time ago and I forgot about it. 1/2/03 */
    /* Ahhh... and what makes it interesting is that you only check for
       the duplicate part if some part of the key changed. */
    /* Kinda sucksies, to avoid reading the record in (which flat probably
       does anyway) we can only check the new record versus the length
       of the key they sent us */
    FlatResponse fr;

    Memory key;
    Memory in;
    UINT keylen = sco.getint(); // 1/9/01
    int ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }
    sco.getint(); // 1/9/01 length not needed here
    ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get update string." << endl;
        return;
      }

    FFilePtr *ffp;
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Update rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
      
    /* this will break something if they don't send us the full key length
       to work with */
    szfix(key, keylen); // 1/9/01
    
    szfix(in, ffp->getrecordsize());
 
    /* The presumption is if they're using atomic update
       they've sent the whole key length to check, because
       that's what we're going to do our seek on. */
    // make room for seek...
    Memory out(ffp->getrecordsize());
    Memory destkey(in); // get a copy of the string, it will have the index in the front of it
    szfix(destkey, ffp->getkeysize()); // make sure this matches up to check against
    { MutexLockWrite x(ffp->getlock()); // lock whole operation, seek and insert
    
    /* only check for duplicate if the primary key part is changing. And
       we check that what we're going to change TO doesn't exist. */
    if (memcmp(key, destkey, keylen) != 0) // check the size of the key they want to update vs the record they're sending us
      {    
        ret = flatseek(fh, destkey, ffp->getkeysize(), out); // xxxz check this
        if (ret == 0) // finding it is bad...
          {
            fr.rc = FR_DUPLICATE;
            if (sendfr(sco, fr) != FR_OK)
              ls(NOTEERR) << "Can't send response to client for seekrec fail." << endl;
            return;
          }
      } // else the key is the same and it's okay to update.
      ret = flatupdate(fh, key, in);
    }
    if (ret != 0)
      {
        if (ret == 2)
          fr.rc = FR_NOTFOUND;
        else
          fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for update fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for update success." << endl;
  
    return;    
  }

void FlatServer::FSinsert(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    Memory in;
    
    sco.getint(); // 1/9/01 unused key length
    int ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get insert string." << endl;
        return;
      }
      
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Insert rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    szfix(in, ffp->getrecordsize());
    
    { MutexLockWrite x(ffp->getlock());
      ret = flatappend(fh, in);
    }
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for update fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for insert success." << endl;
  
    return;    
  }

void FlatServer::FSatomicinsert(socketio &sco, FlatPacket &fp)
  {
    /* same thing like insert, but we do a seek first, and if
       it's there, we return FR_DUPLICATE */
       
    FlatResponse fr;
    FFilePtr *ffp;
    Memory in;
    
    sco.getint(); // unused key length
    int ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get insert string." << endl;
        return;
      }
      
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Atomic Insert rec." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }

    // make room for seek...
    Memory out(ffp->getrecordsize());
    Memory key(in); // get a copy of the string, it will have the index in the front of it
    { MutexLockWrite x(ffp->getlock()); // lock whole operation, seek and insert
    ret = flatseek(fh, key, ffp->getkeysize(), out); // xxxz check this
    if (ret == 0) // finding it is bad...
      {
        fr.rc = FR_DUPLICATE;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for seekrec fail." << endl;
        return;
      }
    
    szfix(in, ffp->getrecordsize());
    
    ret = flatappend(fh, in);
    } // mutex lock
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for update fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for insert success." << endl;
  
    return;    
  }

void FlatServer::FSdelete(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    Memory key;

    UINT keylen = sco.getint();
    int ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Delete key." << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen); // 1/9/01
    
    { MutexLockWrite x(ffp->getlock());
      ret = flatseekdel(fh, key, keylen);
    }
    if (ret != 0)
      {
        if (ret == 2)
          fr.rc = FR_NOTFOUND;
        else
          fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for seekdel fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for seekdel success." << endl;
  
    return;    
  }

void FlatServer::FSfirstrec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "First rec" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    int ret;
    { MutexLockRead x(ffp->getlock());
      ret = firstrec(fh);
    }
    if (ret == -1)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for firstrec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    fr.data.recno = ret; // pass this back
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for firstrec success." << endl;
  
    return;    
  }

void FlatServer::FSlastrec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Last rec" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    int ret;
    { MutexLockRead x(ffp->getlock());
      ret = lastrec(fh);
    }
    if (ret == -1)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for lastrec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    fr.data.recno = ret; // pass this back
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for lastrec success." << endl;
  
    return;    
  }

void FlatServer::FSseekrecno(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    Memory key;
    
    UINT keylen = sco.getint(); // 1/9/01
    int ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Seek key" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen);
    
    { MutexLockRead x(ffp->getlock());
      ret = flatseekrec(fh, key, keylen); // 1/9/01
    }
    if (ret == -1)
      {
        fr.rc = FR_NOTFOUND; // I dunno why, this could also be a fail
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for seekrecno fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    fr.data.recno = ret; // pass this back
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for lastrec success." << endl;
  
    return;    
  }

void FlatServer::FSnextrec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Next key rec" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    int ret;
    { MutexLockRead x(ffp->getlock());
      ret = nextrec(fh, fp.recno);
    }
    if (ret == -1)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for nextrec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    fr.data.recno = ret; // pass this back
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for nextrec success." << endl;
  
    return;    
  }

void FlatServer::FSprevrec(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Prev key rec" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    int ret;
    { MutexLockRead x(ffp->getlock());
      ret = prevrec(fh, fp.recno);
    }
    if (ret == -1)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for prevrec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    fr.data.recno = ret; // pass this back
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for prevrec success." << endl;
  
    return;    
  }

// the blob functions are pretty much identical to their
// record counterparts 

void FlatServer::FSBseek(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    Memory key;
    UINT keylen = sco.getint();
    int ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Blob seek" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen);  // 1/9/01
    
    long sz;
    Memory out;
    { MutexLockRead x(ffp->getlock());
    // this is really inefficient, we call this twice
    sz = blobsize (fh, key);

    if (sz == -1) // not found
      {
        fr.rc = FR_NOTFOUND;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for seekrec fail." << endl;
        return;
      }
    
    Memory outbuf(sz);
    ret = blobseek(fh, key, outbuf);
    out = outbuf; // out of mutex scope copy
    } // mutex lock
    if (ret != 0)
      {
        if (ret == 2)
          fr.rc = FR_NOTFOUND;
        else
          fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for seekrec fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    fr.data.recno = sz; // so they know the size they get back
    ret = sendfr(sco, fr);
    ret &= sco.sendstr(out, sz); // with size so other end knows how much to accept
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for seekrec success." << endl;
  
    return;    
  }

void FlatServer::FSBinsert(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    Memory in;
    
    Memory key;
    UINT keylen = sco.getint(); // 1/9/01
    int ret = sco.getstr(key); // blob inserts specify their own key
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    sco.getint(); // 1/9/01 unused key length
    ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get blob insert string." << endl;
        return;
      }
      
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Blob insert" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen); 

    { MutexLockWrite x(ffp->getlock());
      ret = blobappend(fh, key, in, fp.recno); // easiest way to get size info
    }
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for blob insert fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for blob insert success." << endl;
  
    return;    
  }

void FlatServer::FSBatomicinsert(socketio &sco, FlatPacket &fp)
  {
    /* Similar to atomicinsert. If there's already a blob there,
       don't delete/replace it, report a duplicate. */
       
    FlatResponse fr;
    FFilePtr *ffp;
    Memory in;
    
    Memory key;
    UINT keylen = sco.getint(); // 1/9/01
    int ret = sco.getstr(key); // blob inserts specify their own key
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    sco.getint(); // unused key size
    ret = sco.getstr(in);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get blob insert string." << endl;
        return;
      }
      
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Blob atomic insert" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen); // perhaps this should be left alone max out key should match in blobs

    long sz;
    { MutexLockWrite x(ffp->getlock());
    // do the check here first.
    sz = blobsize (fh, key);

    if (sz != -1) // it didn't not find it.
      {
        fr.rc = FR_DUPLICATE;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for blobatomicinsert fail." << endl;
        return;
      }

    ret = blobappend(fh, key, in, fp.recno); // easiest way to get size info
    } // mutex lock
    if (ret != 0)
      {
        fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for blob insert fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for blob insert success." << endl;
  
    return;    
  }

void FlatServer::FSBdelete(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    Memory key;
    UINT keylen = sco.getint(); // 1/9/01
    int ret = sco.getstr(key);
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Can't get key string." << endl;
        return;
      }

    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Blob delete" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    szfix(key, keylen); // 1/9/01 maybe should be ffp->getkeysize());
    
    { MutexLockWrite x(ffp->getlock());
      ret = blobdelete(fh, key);
    }
    if (ret != 0)
      {
        if (ret == 2)
          fr.rc = FR_NOTFOUND;
        else
          fr.rc = FR_FAIL; // I dunno why...
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for blob delete fail." << endl;
        return;
      }
    fr.rc = FR_OK;
    ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for blob delete success." << endl;
  
    return;    
  }
          
void FlatServer::FSreindex(socketio &sco, FlatPacket &fp)
  {
    FlatResponse fr;
    FFilePtr *ffp;
    
    FLATHAND fh = getfile(fp, ffp);
    ls(NOTEDIA) << "Reindex" << endl;
    if (fh == NULL)
      {
        fr.rc = FR_BADFILEID;
        if (sendfr(sco, fr) != FR_OK)
          ls(NOTEERR) << "Can't send response to client for bad fileid." << endl;
        return;
      }
    
    { MutexLockWrite x(ffp->getlock()); // this one's a doozie
      reindex(fh); // doesn't return errors.
    }
    fr.rc = FR_OK;
    int ret = sendfr(sco, fr);
    if (ret != FR_OK)
      ls(NOTEERR) << "Can't send response to client for reindex success." << endl;
  
    return;    
  }

/* We need a normalize function here. To balance the index tree. */
