/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>

#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/acceptor.h>
#include <Q/socketio.h>
#include <Q/notecraft.h>
#include <Q/configcraft.h>
#include <Q/filelist.h>
#include <flat/fc.h>

#include "release.h"
#include "global.h"
#include "db.h"
#include "io.h"
#include "dbfinder.h"
#include "utils.h"
#include "msgstore.h"
#include "admin.h"

#define NAME "admin"

/* This is the admin command line program for the O3 mail system. By Stu Mark.

   For adding, updating, and deleting users, domains and other
   system related information.
*/
   
int main(int num, char *opts[])
  {
    ConfigCraft ar(O3NAME, num, opts, TRUE); // look for /etc/o3.cf

    addconfigdefaults(ar); /* we have to do this after refresh as well. */

    NoteCraft ls(NAME, ar[TXT_LOGFILE], val(ar[TXT_LOGLEVEL]));
 
    ioinit(ar, ls); // ugly, but it works, and there's only one.
    Memory chk;
    getdbrootdir(chk);
    if (strcmp(chk, "/") == 0)
      {
        printf ("Need config file to work with. Pass -M<cfgfilename> on command line\n");
        exit (1);
      }

    Memory str;
    str = "220 ";
    str.addstr("O3 admin. Build ");
    Memory relstr;
    getrelease(relstr);
    str.addstr(relstr);
    str.addstr(" " __DATE__ " " __TIME__);
    printf ("%s\n", (char *)str);

    /* Read off the command line, the command and function as
       appropriate. */
       
    if (num < 2)
      {
        help();
      }
    else
      {
        admin go(ar, ls, num, opts);
        go.run();
      }
    printf (".\n");
    return 0;
  }

void addconfigdefaults(ConfigCraft &ar)
  {
    ar.adddefault(TXT_LOGFILE, NAME ".log");
    ar.adddefault(TXT_LOGLEVEL, "0");
    ar.adddefault(TXT_SOCKETTIMEOUT, "30");
  }

int help(void)
  {
    printf ("500 Insufficient parameters\n");
    printf ("adddomain <domain>                        - add a domain to the list O3 accepts\n");
    printf ("removedomain <domain>                     - remove a domain from the list O3 accepts\n");
    printf ("listdomains                               - list all of the domains that are local to this O3 install\n");
    printf ("relaydomain <domain> <host:port>          - make this domain a relay for host:port\n");
    printf ("unrelaydomain <domain>                    - remove relay for this domain.\n");
    printf ("\n");
    printf ("listusers                                 - list users\n");
    printf ("showuser <username> <domain>              - show user information\n");
    printf ("adduser <username> <domain> <defstore>    - create a user account in this domain and mailstore\n");
    printf ("chpassword <username> <domain> <oldpasswd> <newpassword>\n");
    printf ("                                          - set user's password\n");
    printf ("chtype <username> <domain> <type> [alias|username|path|smtp]\n");
    printf ("       type=mail|alias|program|file|smtp\n");
    printf ("            mail                          - normal mail account\n");
    printf ("            alias <username@domain>       - alias to other user in same domain\n");
    printf ("            program <filename>            - program to run for each mail received\n");
    printf ("            file <filename>               - append mail to this file\n");
    printf ("            smtp <smtp recip@server:port> - deliver mail to this mail server\n");
    printf ("chstatus <username> <domain> <password> <e|d>\n");
    printf ("                                          - set account status Enabled or Disabled\n");
    printf ("chquota <username> <domain> <quotamax>    - max space allowed for users mailbox 0=no limit\n");
    printf ("chsmtpallow <username> <domain> <y|n>     - allow user to send mail\n");
    printf ("removeuser <username> <domain>            - remove a user and all their mail\n");
    printf ("\n");
    printf ("getmsgdefstore <mailboxdefstore>          - get the current message defstore from this defstore\n");
    printf ("setmsgdefstore <mailboxdefstore> <message defstore>\n");
    printf ("                                          - set the current message defstore from this defstore\n");
    printf ("reindexdirectory                          - rebuild the index for the directory and resource table only\n");
    printf ("reindex <defstore>                        - rebuild the indexes for the mailstore on this defstore\n");
    return 0;
  }


/* it is worth noting that if we split to multiple data stores we
   have to either copy the resource table to each instance
   or keep it in the prime store with the directory. */
/* I think we decided to keep one copy with the directory. */
/* Actually, I have a better answer. Have two resource tables.
   The one in the primedefstore has global stuff, the one
   on the mailstore has specific stuff (like current msgstore defstore)
   and if they happen to be the same, so be it, the information will
   be in the same table. */

admin::admin(ConfigCraft &Ar, NoteCraft &Ls, int Num, char **Opts) : ar(Ar), ls(Ls)
  {
    num = Num;
    opts = Opts;
    finder = NULL;
    dirholder = NULL;
  }

admin::~admin(void)
  {
    if (dirholder != NULL)
      delete dirholder;
    if (finder != NULL)
      delete finder;
  }  
  
int admin::run(void)
  {
    if (checkfinder(ar, ls, finder, dirholder) == FALSE)
      {
        printf ("Unable to connect to database.\n");
        return 0;
      }

    int n = 1;
    /* zero param commands */
    if (strcasecmp(opts[n], "listdomains") == 0)
      return listdomains();
    else if (strcasecmp(opts[n], "listusers") == 0)
      return listusers();
    else if (strcasecmp(opts[n], "reindexdirectory") == 0)
      return reindexdirectory();

    if (num < 3)
      {
        printf ("500 unknown command\n");
        return help();
      }

    /* one param commands */
    if (strcasecmp(opts[n], "adddomain") == 0)
      return adddomain(opts[n+1]);
    else if (strcasecmp(opts[n], "removedomain") == 0)
      return removedomain(opts[n+1]);
    if (strcasecmp(opts[n], "unrelaydomain") == 0)
      return unrelaydomain(opts[n+1]);
    else if (strcasecmp(opts[n], "reindex") == 0)
      return reindex(opts[n+1]); 
    else if (strcasecmp(opts[n], "getmsgdefstore") == 0)
      return getmessagedefstore(opts[n+1]); 

    if (num < 4)
      {
        printf ("500 unknown command\n");
        return help();
      }

    /* two param commands */
    if (strcasecmp(opts[n], "showuser") == 0)
      return showuserbyname(opts[n+1], opts[n+2]); // by username
    else if (strcasecmp(opts[n], "removeuser") == 0)
      return removeuser(opts[n+1], opts[n+2]);
    else if (strcasecmp(opts[n], "relaydomain") == 0)
      return relaydomain(opts[n+1], opts[n+2]);
    else if (strcasecmp(opts[n], "setmsgdefstore") == 0)
      return setmessagedefstore(opts[n+1], opts[n+2]);

    if (num < 5)
      {
        printf ("500 unknown command\n");
        return help();
      }

    /* three param commands */
    if (strcasecmp(opts[n], "adduser") == 0)
      return adduser(opts[n+1], opts[n+2], opts[n+3]);
    else if (strcasecmp(opts[n], "chquota") == 0)
      return chquota(opts[n+1], opts[n+2], opts[n+3]); 
    else if (strcasecmp(opts[n], "chsmtpallow") == 0)
      return chsmtpallow(opts[n+1], opts[n+2], opts[n+3]); 
    else if (strcasecmp(opts[n], "chtype") == 0) // variable number for this
           {
             if (num == 5)
               return chtype(opts[n+1], opts[n+2], opts[n+3], NULL); 
             else
               return chtype(opts[n+1], opts[n+2], opts[n+3], opts[n+4]); 
           }

    if (num < 6)
      {
        printf ("500 unknown command\n");
        return help();
      }

    if (strcasecmp(opts[n], "chpassword") == 0)
      return chpassword(opts[n+1], opts[n+2], opts[n+3], opts[n+4]); 
    else if (strcasecmp(opts[n], "chstatus") == 0)
      return chstatus(opts[n+1], opts[n+2], opts[n+3], opts[n+4]); 

    printf ("500 unknown command\n");
    return help();
  }





/* -----------------------------  Domain functions ------------------------------- */
int admin::adddomain(char *dm)
  {
    /* do some minimal checking to see if domain name is okay. */
    Memory n(dm);
    trim(n);
    lowdo(n);
    UINT max = strlen(n);
    int bad = FALSE;
    UINT lp;
    char *valid = "abcdefghijklmnopqrstuvwxyz-.0123456789";
    char *badenders = ".-";
    
    if (max == 0) // must have something
      bad = TRUE;
    if (bad == FALSE)
      for (lp = 0; lp < max; lp++) // must be in valid chars.
        {
          if (strchr (valid, n[lp]) == NULL)
            bad = TRUE;
        }
    if (strchr(badenders, n[0]) != NULL)
      bad = TRUE;
    if (strchr(badenders, n[max-1]) != NULL)
      bad = TRUE;
    max = SZ_RESKEY-strlen(TXT_DOMAINPREFIX); // see if we can fit it. 
    if (strlen(n) >= max)
      bad = TRUE;

    if (bad)
      {
        printf("500 %s Invalid domain name\n", (char *)n);
        return 0;
      }
    // if it gets here, we add it.
    tresource res(*finder->getprimeconn(), ar, ls);

    // let's see if it already exists.
    Memory k = TXT_DOMAINPREFIX;
    k.addstr(n);
    int ret = res.lookup(k);
    if (ret == RET_OK)
      {
        printf("400 %s already exists\n", (char *)n);
        return 0;
      }

    memorytofield(res.getdata().key, k, sizeof(res.getdata().key));
    memorytofield(res.getdata().value, n, sizeof(res.getdata().value));
    ret = res.add(k);
    if (ret == RET_OK)
      printf ("200 %s domain added\n", (char *)n);
    else
      printf ("500 %s unable to add domain\n", (char *)n);
    return 0;
  }

int admin::removedomain(char *dm)
  {
    Memory n(dm);
    trim(n);
    lowdo(n);
    int max = strlen(n);
    
    if (max == 0) // must have something
      {
        printf ("400 bad domain\n");
        return 0;
      }

    tresource res(*finder->getprimeconn(), ar, ls);

    Memory k = TXT_RELAYDOMAINPREFIX;
    k.addstr(n);
    res.del(k); // if there's a relay record, delete it.
    
    k = TXT_DOMAINPREFIX;
    k.addstr(n);
    int ret = res.del(k);
    if (ret == RET_OK)
      printf ("200 %s domain removed\n", (char *)n);
    else
      printf ("500 %s unable to remove domain\n", (char *)n);
    return 0;
  }
  
int admin::listdomains(void)
  {
    zlist domainlist;
    
    tresource res(*finder->getprimeconn(), ar, ls);
    res.getdomainlist(domainlist);
    UINT lp;
    UINT max = domainlist.ListCount();
    printf ("200 %d in list\n", max);
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)domainlist.GetItem(lp);
        Memory dm = s->getstr();
        // look up and see if there's a relay record for this domain
        Memory k = TXT_RELAYDOMAINPREFIX;
        k.addstr(dm);
        int ret = res.lookup(k);
        Memory sup;
        if (ret == RET_OK)
          {
            Memory val;
            fieldtomemory(val, res.getdata().value, sizeof(res.getdata().value));
            sup = " will relay to ";
            sup.addstr(val);
          }
        printf ("%s%s\n", (char *)s->getstr(), (char *)sup);
      }
    return 0;
  }

int admin::relaydomain(char *dm, char *hostport)
  {
    /* add a resource record for this domain TXT_RELAYDOMAINPREFIX
       if there is a value for this, it is the host:port string
       for the host we're supposed to bounce all mail from this domain
       to. */

    Memory host;
    int port;
    Memory junk;

    splithostportpath(hostport, host, port, junk);
    
    if (strlen(host) == 0)
      {
        printf("500 %s Invalid host name\n", hostport);
        return 0;
      }
    
    if (port == 0)
      {
        printf("500 %s Invalid port number\n", hostport);
        return 0;
      }
    
    tresource res(*finder->getprimeconn(), ar, ls);

    // let's see if it already exists.
    Memory k = TXT_RELAYDOMAINPREFIX;
    k.addstr(dm);
    int ret = res.lookup(k);
    if (ret == RET_OK)
      {
        printf("400 %s relay domain already exists\n", dm);
        return 0;
      }

    memorytofield(res.getdata().key, k, sizeof(res.getdata().key));
    memorytofield(res.getdata().value, hostport, sizeof(res.getdata().value));
    ret = res.add(k);
    if (ret == RET_OK)
      printf ("200 %s relay domain added\n", hostport);
    else
      printf ("500 %s unable to add domain\n", hostport);
    return 0;
  }

int admin::unrelaydomain(char *dm)
  {
    /* see if there's a resource record for this domain TXT_RELAYDOMAINPREFIX
       delete it if it's there. */

    tresource res(*finder->getprimeconn(), ar, ls);

    // let's see if it already exists.
    Memory k = TXT_RELAYDOMAINPREFIX;
    k.addstr(dm);
    int ret = res.lookup(k);
    if (ret != RET_OK)
      {
        printf("400 %s relay domain not found\n", dm);
        return 0;
      }
    ret = res.del(k); // if there's a relay record, delete it.
    if (ret == RET_OK)
      printf ("200 %s relay domain removed\n", dm);
    else
      printf ("500 %s unable to remove relay domain\n", dm);
    return 0;
  }

/* -----------------------------  Other resource functions ------------------------------- */

/* Following the idea that we will someday have multiple resource tables, one
   in the primedefstore for the domain list, possibly others for the current
   message store defstores foe each mailbox defstore, we require the defstore
   to look up. It can be the prime mailbox defstore if you only have one mailbox defstore */

int admin::getmessagedefstore(char *defstore)
  {
    /* Get the current message defstore setting out of this mailstore defstore */

    FlatConnection *mailstore = finder->getconn(defstore);
    if (mailstore == NULL)
      {
        printf ("500 %s Unable to find defstore\n", (char *)defstore);
        return 0;
      }    

    tresource res(*mailstore, ar, ls);
    int ret = res.lookup(TXT_MESSAGE_DEFSTORE);
    if (ret == RET_NOTFOUND)
      {
        printf ("400 %s Messagestore defstore not set on this mailbox defstore\n", defstore);
        return 0;
      }
    if (ret != RET_OK)
      {
        printf ("500 %s Unable to read resource table on this mailbox defstore\n", defstore);
        return 0;
      }

    Memory msgdstore;
    fieldtomemory (msgdstore, res.getdata().value, sizeof(res.getdata().value));
    trim(msgdstore);
    printf ("200 %s is the current message defstore for mailbox defstore %s\n", (char *)msgdstore, (char *)defstore);
    return 0;
  }

int admin::setmessagedefstore(char *defstore, char *newmsgdefstore)
  {
    /* Set the current message defstore for this mailstore defstore with newmsgdefstore
       Do some checking on it. */
       
    Memory msgdstore(newmsgdefstore);
    trim(msgdstore);
    if (strlen(msgdstore) != SZ_DEFSTORE)
      {
        printf ("400 %s is not a valid message defstore - must be %d characters in length\n", (char *)msgdstore, SZ_DEFSTORE);
        return 0;
      }

    FlatConnection *mailstore = finder->getconn(defstore);
    if (mailstore == NULL)
      {
        printf ("500 %s Unable to find defstore\n", (char *)defstore);
        return 0;
      }    

    tresource res(*mailstore, ar, ls);
    int ret = res.del(TXT_MESSAGE_DEFSTORE);
    if ((ret != RET_NOTFOUND) && (ret != RET_OK))
      { // some database error
        printf ("500 %s Unable to remove old message defstore from this mailbox defstore\n", defstore);
        return 0;
      }

    // now add the new one
    memorytofield(res.getdata().value, msgdstore, sizeof(res.getdata().value));
    ret = res.add(TXT_MESSAGE_DEFSTORE);
    if (ret != RET_OK)
      {
        printf ("500 %s Unable to add new message defstore from this mailbox defstore\n", defstore);
        return 0;
      }
    printf ("200 %s The new message defstore for mailbox defstore is set.\n", (char *)msgdstore);
    return 0;
  }


/* -----------------------------  User functions ------------------------------- */

int admin::listusers(void)
  {
    // this could be painful...
    zlist userlist;
    
    tdirectory dir(*finder, ar, ls);
    dir.getuserlist(userlist);
    
    UINT lp;
    UINT max = userlist.ListCount();
    printf ("200 %d in list\n", max);
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)userlist.GetItem(lp);
        dirrec *d = (dirrec *)(char *)s->getstr();
        Memory user;
        Memory domain;
        fieldtomemory(user, d->username, sizeof(d->username));
        fieldtomemory(domain, d->domain, sizeof(d->domain));
        printf ("%s %s\n", (char *)user, (char *)domain);
      }
    return 0;
  }

int admin::adduser(char *username, char *domain, char *defstore)
  {
    /* See if this record already exists in the directory, if not,
       validate the defstore and add it. */

    // see if domain exists.
    tresource res(*finder->getprimeconn(), ar, ls);
    Memory k = TXT_DOMAINPREFIX;
    k.addstr(domain);
    int ret = res.lookup(k);
    if (ret != RET_OK)
      {
        printf("400 %s domain doesn't exist in domain table\n", domain);
        return 0;
      }
    
    tdirectory dir(*finder, ar, ls);
    ret = dir.lookup(username, domain);
    if (ret == RET_OK)
      {
        printf ("400 %s@%s User already exists\n", username, domain);
        return 0;
      }
    if (ret != RET_NOTFOUND)
      {
        printf ("500 Unable to check database\n");
        return 0;
      }
      
    FlatConnection *mailstore = finder->getconn(defstore);
    if (mailstore == NULL)
      {
        printf ("500 %s Unable to find defstore\n", defstore);
        return 0;
      }    

    // so now we try and add it. We add a directory, a user, and an inbox.
    memorytofield(dir.getdata().defstore, defstore, sizeof(dir.getdata().defstore));
    ret = dir.add(username, domain);
    if (ret != RET_OK)
      {
        printf ("500 %s@%s Unable to add user\n", username, domain);
        return 0;
      }
    
    UINT userid = dir.getdata().userid; // this guys brand new userid

    tuser u(*mailstore, ar, ls);
    u.getdata().status = USERSTATUS_DISABLED;
    u.getdata().type = USERTYPE_MAIL;
    u.getdata().access = 0;
    // default pw is "." so it can be changed by mkuser.sh
    memorytofield(u.getdata().password, ".", sizeof(u.getdata().password));
    memorytofield(u.getdata().alt, "", sizeof(u.getdata().alt));
    u.getdata().quotamax = 0;
    u.getdata().quotaused = 0;
    u.getdata().allowsmtp = ALLOWSMTP_NO;
    
    ret = u.add(userid);
    if (ret != RET_OK)
      {
        printf ("500 %s@%s Unable to add user, account in broken state\n", username, domain);
        return 0;
      }

    // add an inbox
    tfolder f(*mailstore, ar, ls);
    ret = f.add(userid, TXT_INBOX);
    if (ret != RET_OK)
      {
        printf ("500 %s@%s Unable to add user, account in broken state\n", username, domain);
        return 0;
      }
    printf ("200 %s@%s User added\n", username, domain);
    return 0;
  }  

FlatConnection *admin::getuser(char *username, char *domain, tuser *&u)
  {
    /* Everybody uses this to load a user.
       Given the name and domain it will lookup in the directory, find the
       defstore and load the user record. If it returns non null then 
       u is valid and needs to be deleted. */
       
    u = NULL; // this is kinda hokey, but don't make object if can't find user
    tdirectory dir(*finder, ar, ls);
    int ret = dir.lookup(username, domain);
    if (ret == RET_NOTFOUND)
      {
        printf ("400 %s@%s User not found\n", username, domain);
        return NULL;
      }
    if (ret != RET_OK)
      {
        printf ("500 %s@%s Unable to look up user\n", username, domain);
        return NULL;
      }
    UINT userid = dir.getdata().userid;
      
    Memory defstore;
    /* Get his storage location */
    fieldtomemory (defstore, dir.getdata().defstore, sizeof(dir.getdata().defstore));

    FlatConnection *mailstore = finder->getconn(defstore);
    if (mailstore == NULL)
      {
        printf ("500 %s Unable to find defstore\n", (char *)defstore);
        return NULL;
      }    

    u = new tuser(*mailstore, ar, ls);
    ret = u->lookup(userid);
    if (ret == RET_NOTFOUND)
      {
        printf ("400 %s@%s (%d) User not found\n", username, domain, userid);
        delete u; 
        u = NULL;
        return NULL;
      }
    if (ret != RET_OK)
      {
        printf ("500 %s@%s (%d) Unable to look up user\n", username, domain, userid);
        delete u;
        u = NULL;
        return NULL;
      }

    return mailstore;
  }

int admin::showuserbyname(char *username, char *domain)
  { /* heh, we can't do showuserbyid, because we have no way of looking up the username
       by the id, just the name. */

    tuser *u;
    FlatConnection *mailstore = getuser(username, domain, u);
    if (mailstore == NULL)
      return 0;

    /* This is a double effort, but, but... well... */
    tdirectory dir(*finder, ar, ls);
    dir.lookup(username, domain); // we know it's there, we just looked it up ;-)
      
    printf ("200 User information follows\n");

    Memory d;
    fieldtomemory(d, dir.getdata().username, sizeof(dir.getdata().username));
    printf ("Username: %s\n", (char *)d);
    
    fieldtomemory(d, dir.getdata().domain, sizeof(dir.getdata().domain));
    printf ("Domain: %s\n", (char *)d);
    
    printf ("Userid: %d\n", dir.getdata().userid);
    
    fieldtomemory(d, dir.getdata().defstore, sizeof(dir.getdata().defstore));
    printf ("Defstore: %s\n", (char *)d);

    if (u->getdata().status == USERSTATUS_ENABLED)
      d = "Enabled";
    else
      d = "Disabled";
    printf ("Status: %s\n", (char *)d);

    switch (u->getdata().type)
      {
        case USERTYPE_MAIL:
          d = "Mail account"; break;
        case USERTYPE_ALIAS:
          d = "Mail alias"; break;
        case USERTYPE_PROGRAM:
          d = "Program"; break;
        case USERTYPE_FILE:
          d = "File"; break;
        case USERTYPE_SMTP:
          d = "SMTP server"; break;
        default:
          d = "Unknown";
      }

    printf ("Type: %s\n", (char *)d);

    printf ("Access: %d\n", u->getdata().access);

    fieldtomemory(d, u->getdata().alt, sizeof(u->getdata().alt));
    printf ("Alt: %s\n", (char *)d);
    
    printf ("Quota max: %d\n", u->getdata().quotamax);
    printf ("Quota used: %d (%d%%)\n", u->getdata().quotaused, u->getdata().quotaused*100/u->getdata().quotamax);

    if (u->getdata().allowsmtp == ALLOWSMTP_YES)
      d = "Yes";
    else
      d = "No";
    printf ("Can SMTP: %s\n", (char *)d);

    delete u;
    return 0;
  }

int admin::chpassword(char *username, char *domain, char *oldpw, char *newpassword)
  {
    tuser *u;
    FlatConnection *mailstore = getuser(username, domain, u);
    if (mailstore == NULL)
      return 0;

    /* Check password. */
    Memory pw;
    fieldtomemory (pw, u->getdata().password, sizeof(u->getdata().password));
    if (strcmp(oldpw, pw) != 0)
      {
        printf ("500 Invalid password\n");
        delete u;
        return 0;
      }
    memorytofield (u->getdata().password, newpassword, sizeof(u->getdata().password));
    
    int ret = u->update(u->getdata().userid);
    if (ret == RET_OK)
      printf ("200 Password update completed\n");
    else
      printf ("500 Unable to update password\n");
    delete u;
    return 0;
  }

int admin::chtype(char *username, char *domain, char *newtype, char *altval)
  {
    /* Check type */
    UINT type = 0;
    if (strcasecmp(newtype, "mail") == 0)
      type = USERTYPE_MAIL;
    if (strcasecmp(newtype, "alias") == 0)
      type = USERTYPE_ALIAS;
    if (strcasecmp(newtype, "program") == 0)
      type = USERTYPE_PROGRAM;
    if (strcasecmp(newtype, "file") == 0)
      type = USERTYPE_FILE;
    if (strcasecmp(newtype, "smtp") == 0)
      type = USERTYPE_SMTP;
    if (type == 0)
      {
        printf ("400 %s invalid user type (mail/alias/program/file/smtp)\n", newtype);
        return 0;
      }
    Memory alt = altval;
    trim(alt);
    if (type != USERTYPE_MAIL)
      {
        if (strlen(alt) == 0)
          {
            printf ("400 %s parameter required for type\n", newtype);
            return 0;
          }
      }
    else 
      alt = "";

    tuser *u;
    FlatConnection *mailstore = getuser(username, domain, u);
    if (mailstore == NULL)
      return 0;
      
    memorytofield (u->getdata().alt, alt, sizeof(u->getdata().alt));
    u->getdata().type = type;
    
    int ret = u->update(u->getdata().userid);
    if (ret == RET_OK)
      printf ("200 Type update completed\n");
    else
      printf ("500 Unable to update type\n");
    delete u;
    return 0;
  }

int admin::chstatus(char *username, char *domain, char *password, char *newstatus)
  {
    tuser *u;
    FlatConnection *mailstore = getuser(username, domain, u);
    if (mailstore == NULL)
      return 0;

    /* Check password. */
    Memory pw;
    fieldtomemory (pw, u->getdata().password, sizeof(u->getdata().password));
    if (strcmp(password, pw) != 0)
      {
        printf ("500 Invalid password\n");
        delete u;
        return 0;
      }

    UINT status = 0;
    Memory s = newstatus;
    updo(s);
    if (s[0] == 'D')
      status = USERSTATUS_DISABLED;
    if (s[0] == 'E')
      status = USERSTATUS_ENABLED;
      
    if (status == 0)
      {
        printf ("400 %s invalid status (enabled/disabled)\n", newstatus);
        delete u;
        return 0;
      }

    u->getdata().status = status;
    int ret = u->update(u->getdata().userid);
    if (ret == RET_OK)
      printf ("200 Status update completed\n");
    else
      printf ("500 Unable to update status\n");
      
    delete u;
    return 0;
  }

int admin::chquota(char *username, char *domain, char *newquota)
  {
    tuser *u;
    FlatConnection *mailstore = getuser(username, domain, u);
    if (mailstore == NULL)
      return 0;

    Memory q = newquota;
    trim(q);
    UINT quota = val(q);
    if (quota == 0)
      {
        if (q[0] != '0')
          {
            printf ("400 %s invalid quota value\n", newquota);
            delete u;
            return 0;
          }
      }

    u->getdata().quotamax = quota;
    int ret = u->update(u->getdata().userid);
    if (ret == RET_OK)
      printf ("200 Quota max update completed\n");
    else
      printf ("500 Unable to update quota max\n");
      
    delete u;
    return 0;
  }

int admin::chsmtpallow(char *username, char *domain, char *yn)
  {
    tuser *u;
    FlatConnection *mailstore = getuser(username, domain, u);
    if (mailstore == NULL)
      return 0;

    UINT allow = 0;
    Memory a = yn;
    updo(a);
    if (a[0] == 'Y')
      allow = ALLOWSMTP_YES;
    if (a[0] == 'N')
      allow = ALLOWSMTP_NO;
      
    if (allow == 0)
      {
        printf ("400 %s invalid setting (yes/no)\n", yn);
        delete u;
        return 0;
      }

    u->getdata().allowsmtp = allow;
    int ret = u->update(u->getdata().userid);
    if (ret == RET_OK)
      printf ("200 Allowsmtp update completed\n");
    else
      printf ("500 Unable to update allowsmtp\n");
      
    delete u;
    return 0;
  }

int admin::removeuser(char *username, char *domain)
  {
    /* Here's where all that effort in cascading delete in io
       pays off... */

    tdirectory dir(*finder, ar, ls);
    int ret = dir.lookup(username, domain);
    if (ret == RET_NOTFOUND)
      {
        printf ("500 %s@%s User doesn't exist\n", username, domain);
        return 0;
      }

    if (ret != RET_OK)
      {
        printf ("500 %s@%s Unable to access user directory\n", username, domain);
        return 0;
      }

    Memory defstore; /* get mailstore locartion */
    fieldtomemory (defstore, dir.getdata().defstore, sizeof(dir.getdata().defstore));
    FlatConnection *mailstore = finder->getconn(defstore);
    if (mailstore == NULL)
      {
        printf ("500 %s Unable to find defstore\n", (char *)defstore);
        return 0;
      }    

    ret = dir.del(username, domain); // this should be fun to watch
    if (ret == RET_OK)
      printf ("200 %s@%s User deleted\n", username, domain);
    else
      if (ret == RET_PARTIAL)
        printf ("300 %s@%s User was in broken state, but removed\n", username, domain);
      else
        printf ("500 %s@%s Unable to delete user\n", username, domain);
    return 0;
  }


/* -----------------------------  System functions ------------------------------- */

/* We *NEED* a normalizing function (put it in flat) */

int admin::reindexdirectory(void)
  {
    /* Do reindex of just the directory in primedefstore */
    tdirectory dir(*finder, ar, ls);
    tresource res(*finder->getprimeconn(), ar, ls);
    int ret1 = dir.reindex();
    int ret2 = res.reindex();
    if ((ret1 == RET_OK) && (ret2 == RET_OK))
      {
        printf ("200 Directory and resource reindex completed\n");
      }
    else
      printf ("500 Unable to reindex directory or resource\n");
    return 0;
  }

int admin::reindex(char *defstore)
  {
    /* Do reindex of all the tables in this defstore. */
    FlatConnection *mailstore = finder->getconn(defstore);
    if (mailstore == NULL)
      {
        printf ("500 %s Unable to find defstore\n", defstore);
        return 0;
      }
      
    tuser user(*mailstore, ar, ls);
    tfolder folder(*mailstore, ar, ls);
    tmsgptr msgptr(*mailstore, ar, ls);
    tresource res(*mailstore, ar, ls);
    
    int ret1 = user.reindex();
    int ret2 = folder.reindex();
    int ret3 = msgptr.reindex();
    int ret4 = res.reindex();

    if ((ret1 == RET_OK) && (ret2 == RET_OK) && (ret3 == RET_OK) && (ret4 == RET_OK))
      {
        printf ("200 %s Mailstore reindex completed\n", defstore);
      }
    else
      printf ("500 %s Unable to reindex mailstore\n", defstore);
    return 0;
  }





