/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <errno.h>

#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/acceptor.h>
#include <Q/configcraft.h>
#include <Q/notecraft.h>
#include <Q/mutex.h>
#include <flat/fc.h>

#include "global.h"
#include "msgstore.h"
#include "io.h"
#include "dbfinder.h"
#include "utils.h"

Memory DBROOTDIR;

/* Must call this before you do anything to the mail system's io */
void ioinit(ConfigCraft &ar, NoteCraft &ls)
  {
    ls(NOTEINF) << "Initializing io." << endl;
    DBROOTDIR = ar[TXT_DBROOTDIR];
    addslash(DBROOTDIR);
    ls(NOTEINF) << "Mail system data directory: " << (char *)DBROOTDIR << endl;
    umask(0066); // global system wide setting for all programs
  }

void getdbrootdir(Memory &rd)
  {
    rd = DBROOTDIR;
  }

int frtoret(fresp ret)
  {
    switch (ret)
      {
        case FR_NOTFOUND:
          return RET_NOTFOUND;

        case FR_DUPLICATE:
          return RET_DUPLICATE;

        case FR_OK:
          return RET_OK; // hmmm...
        
        default:
          break;
      }
    return RET_ERROR; // all others are error
  }

/*********************** Directory table ************************/

tdirectory::tdirectory(dbfinder &dbf, ConfigCraft &Ar, NoteCraft &Ls) : finder(dbf), dbc(*dbf.getprimeconn()), ar(Ar), ls(Ls)
  {
    /* The directory is always in the primedefstore, so since we get the finder
       we can just ask it where the directory is. */
    fn = DBROOTDIR;
    fn.addstr(TXT_DIRECTORYFILE);

    fdirectory = dbc.Open(fn, sizeof(dirheader), sizeof(dirrec), TRUE,
                          getkeysize());
  }

void tdirectory::checkversion(void)
  {
    // make sure the version in the file matches what
    // we expect, if not, log and hang. If we exit,
    // the parent will just start looping forever
    Memory header;
    fdirectory.getheader(header); // guarenteed to return a buffer of at least header length
    dirheader *h = (dirheader *)(char *)header;
    if (h->version == 0) // empty buffer is zeroed out
      {
        // file is new, set it to the file version we're going to create
        /* You might be getting upset because there's no locking here,
           but this happens only once per file and chances are you're
           doing admin, not accepting mail */
        h->version = TXT_DIRECTORYFILEVERSION;
        fdirectory.setheader(header); /* write it back out */
      }
    if (h->version != TXT_DIRECTORYFILEVERSION)
      {
        ls(NOTESEV) << "***" << endl;
        ls(NOTESEV) << "Incorrect file version for directory. Blocking." << endl;
        ls(NOTESEV) << "***" << endl;
        while (1)
          sleep(60);
      }
  }
  
int tdirectory::getkeysize(void)
  {
    return sizeof(data.username) + sizeof(data.domain);
  }

void tdirectory::makedirectorykey(Memory &key, char *Name, char *Domain, int &keysz)
  {
    /* Make the key to search for out of name and domain */
    // all names and domains in lower case
    key = Name;
    Memory domain(Domain);
    trim(key);
    trim(domain);
    
    pad(key, SZ_USERNAME);
    pad(domain, SZ_DOMAIN);
    
    key.addstr(domain);
    lowdo(key);
    keysz = getkeysize(); // not including trailing \0, key is not a string
  }

int tdirectory::lookup(char *Name, char *Domain)
  {
    /* Given this name and domain, find the record and load it. */
       
    Memory key;
    int keysz;
    makedirectorykey(key, Name, Domain, keysz);
    
    Memory resp;
    fresp ret = fdirectory.iseek(key, resp, keysz);
    if (ret == FR_OK)
      {
        memcpy (&data, (char *)resp, sizeof(data));
        return RET_OK;
      }
    if (ret == FR_NOTFOUND)
      return RET_NOTFOUND;
    return RET_ERROR;
  }

int tdirectory::lookupalias(char *Name, char *Domain)
  {
    /* This is like lookup execpt it follows aliases if it can 
       If it aliases to something that doesn't exist, we return
       not found. 
       What not might make much sense is that fact that denoting an 
       account as an alias is in the user record (because the 
       directory is JUST a directory) but that means we have to
       find the defstore go there, get the user and follow the alias.
       It also means there's a folder record for this person too, but 
       so what. */
       
    int recurse = MAX_ALIAS_DEPTH; /* Max number of alias dereferences */
    Memory name(Name);
    Memory domain(Domain);
    
    while (recurse > 0)
      {
        recurse--;

        int ret = lookup(name, domain);
        if (ret == RET_NOTFOUND)
          return RET_NOTFOUND;
        if (ret != RET_OK)
          return ret;

        // load the user
        UINT userid = data.userid;
        
        Memory defstore; /* Get his defstorage location */
        fieldtomemory (defstore, data.defstore, sizeof(data.defstore));
        FlatConnection *mailstore = finder.getconn(defstore);
        if (mailstore == NULL)
          return RET_ERROR;
        tuser user(*mailstore, ar, ls);
        ret = user.lookup(userid);
        if (ret == RET_NOTFOUND)
          return RET_NOTFOUND;
        if (ret != RET_OK)
          return ret;

        if (user.getdata().type != USERTYPE_ALIAS)
          return RET_OK; // we got something that isn't an alias
        
        /* See if the alias has been disabled */
        if (user.getdata().status != USERSTATUS_ENABLED)
          return RET_OK; /* We stop here, and the caller will find
                            out and report that the account is disabled */
        // process the alias
        /* get the name and domain from alt */
        Memory n;
        fieldtomemory(n, user.getdata().alt, sizeof(user.getdata().alt));
        namefromaddress(name, n);
        domainfromaddress(domain, n);
      } // process alias, recurse while traversing aliases...
      
    return RET_NOTFOUND; // we got an error or not found on last iteration of alias finding
  }

int tdirectory::add(char *name, char *domain)
  {
    /* take all the fields in the data dirrec, get
       the next userid number, set it, and then insert 
       the record. */
    /* We make sure name and domain are set to what they pass us. */
    /* Multiple directorys will screw this up, must work up
       a peer system if we're going to have multiple directories
       (or a master and a bunch of slaves) */
    /* See if user already exists, return RET_DUPLICATE if so */
    
    /* This mutex is to get a unique sequence id from the database.
       This means only one machine can add users. No way to enforce
       this across machines unless we add it to the flatserver. */
    /* We don't need to do the duplicate check because atomic insert does
       that. */
    MutexLock x(MUTEX_DIRADD);
    
    memorytofield(data.username, name, sizeof(data.username), LOWER); // these keys must be lower case
    memorytofield(data.domain, domain, sizeof(data.domain), LOWER); 
    
    /* use diraddmutex to get next higher id */
    Memory header;
    fdirectory.getheader(header);
    dirheader *h = (dirheader *)(char *)header;
    h->highestid++;
    data.userid = h->highestid;
    fdirectory.setheader(header); /* write it back out */
    
    /* Now add the row. */
    memset (data.filler, 0, sizeof(data.filler)); /* for future use make it zero */
    Memory rec((char *)&data, sizeof(data));
    fresp fret = fdirectory.iatomicinsert(rec);
    // if this fails then the user already exists, which shouldn't happen,
    // the sequencer must be messed up or somebody on another machine
    // added a user at the same time and beat me to it.
    return frtoret(fret);
  }
                                                        
int tdirectory::del(char *name, char *domain)
  {
    /* Look up this name and domain record and delete it */
    /* Then find the defstore and delete everything there cascading from user. */
    int ret = lookup(name, domain); // look up the row
    if (ret == RET_OK)
      {
        /* If it's here we can delete it */
        /* We are deleting a user. We must delete all their messages,
           pointers and folders. */
        /* Delete the user record and it will cascade, just like a real database :-) */
        Memory defstore;
        fieldtomemory (defstore, data.defstore, sizeof(data.defstore));
        FlatConnection *mailstore = finder.getconn(defstore);
        if (mailstore == NULL)
          return RET_ERROR;

        int partial = FALSE;
        tuser u(*mailstore, ar, ls);
        int ret = u.del(data.userid);

        if (ret == RET_PARTIAL)
          partial = TRUE;
        else
          if (ret != RET_OK)
            return ret; // don't kill the dir entry if can't delete the sub parts

        Memory key;
        int keysz;
        makedirectorykey(key, name, domain, keysz);
        
        fresp fret = fdirectory.idelete(key, keysz);
        if (partial)
          return RET_PARTIAL;
        return frtoret(fret);
      }
    return ret; // not found or error
  }

int tdirectory::update(char *kname, char *kdomain)
  {
    /* In case they wanna change the username or domain
       update this record with information in datarec.
       So kname is the source, and data. is the dest.
       
       Check first and make sure new user doesn't already exist,
       if it's different. */

    /* See if user already exists, return RET_DUPLICATE if it does.. */
       
    
    Memory Name;
    Memory Domain;
    fieldtomemory(Name, data.username, sizeof(data.username));
    fieldtomemory(Domain, data.domain, sizeof(data.domain));

    /* Only check the name/domain if it's different */
    int keysz;
    Memory srckey;
    makedirectorykey(srckey, kname, kdomain, keysz);

    /* otherwise, update based on src key, replace with data in rec */
    memset (data.filler, 0, sizeof(data.filler)); /* for future use make it zero */
    Memory rec((char *)&data, sizeof(data));
    fresp ret = fdirectory.iatomicupdate(srckey, rec); // it knows key size
    return frtoret(ret);
  }

int tdirectory::getuserlist(zlist &list)
  {
    /* Do a select on everything in the directory and select it to list */
    
    fresp ret = fdirectory.iselect("", list, 0); // keylength = 0
    return frtoret(ret);
  }

/*********************** User table ************************/

tuser::tuser(FlatConnection &f, ConfigCraft &Ar, NoteCraft &Ls) : dbc(f), ar(Ar), ls(Ls)
  {
    fn = DBROOTDIR;
    fn.addstr(TXT_USERFILE);

    fuser = dbc.Open(fn, sizeof(userheader), sizeof(userrec), TRUE, 
                     getkeysize());
  }

void tuser::checkversion(void)
  {
    // make sure the version in the file matches what we expect.
    Memory header;
    fuser.getheader(header); 
    userheader *h = (userheader *)(char *)header;
    if (h->version == 0) 
      {
        h->version = TXT_USERFILEVERSION;
        fuser.setheader(header); /* write it back out */
      }
    if (h->version != TXT_USERFILEVERSION)
      {
        ls(NOTESEV) << "***" << endl;
        ls(NOTESEV) << "Incorrect file version for user file. Blocking." << endl;
        ls(NOTESEV) << "***" << endl;
        while (1)
          sleep(60);
      }
  }

int tuser::getkeysize(void)
  {
    return sizeof(data.userid);
  }

void tuser::makeuserkey(Memory &key, UINT userid, int &keysz)
  {
    // the key is the userid
    keysz = getkeysize();
    Memory k(keysz);
    memcpy (k, &userid, sizeof(userid));
    key = k;
  }
  
int tuser::lookup(UINT userid)
  {
    /* Look up this userid, returns RET_OK if it worked
       and will fill data structure.
       RET_NOT_FOUND or RET_ERROR */

    Memory key;
    int keysz;
    makeuserkey(key, userid, keysz);

    Memory resp;
    fresp ret = fuser.iseek(key, resp, keysz);
    if (ret == FR_OK)
      {
        memcpy (&data, (char *)resp, sizeof(data));
        return RET_OK;
      }
    if (ret == FR_NOTFOUND)
      return RET_NOTFOUND;
    return RET_ERROR;
  }

int tuser::add(UINT userid)
  {
    /* take all the fields in the data userrec, set the userid
       and then insert the record. */

    /* We don't need to mutex lock because we're sure that the caller
       gave us a unique ID, right? */    

    /* The atomic insert command will yield a duplicate if there's
       already something with the same key there. */
       
    data.userid = userid; // set the key field
    
    /* Now add the row. */
    memset (data.filler, 0, sizeof(data.filler)); /* for future use make it zero */
    Memory rec((char *)&data, sizeof(data));
    fresp fret = fuser.iatomicinsert(rec);

    return frtoret(fret);
  }
                                                        
int tuser::del(UINT userid)
  {
    /* Look up this user and delete it */

    int ret = lookup(userid); // look up the row
    if (ret == RET_OK)
      {
        /* If it's here we can delete it */
        /* We are deleting a user. We must delete all their messages,
           pointers and folders. */
        /* Delete the folder and it will cascade, just like a real database :-) */
        tfolder f(dbc, ar, ls); // this is the connection of the defstore for this guy, not the directory defstore
        ret = f.deluser(userid);
        if (ret != RET_PARTIAL)
          if (ret != RET_OK)
            return ret;

        Memory key;
        int keysz;
        makeuserkey(key, userid, keysz);
        
        fresp fret = fuser.idelete(key, keysz);
        if (ret == RET_PARTIAL)
          return ret;
        return frtoret(fret);
      }
    else
      if (ret == RET_NOTFOUND)
        return RET_PARTIAL;
    return ret; // not found or error
  }

int tuser::update(UINT userid)
  {
    /* They can change anything but the userid */
    Memory key;
    int keysz;
    makeuserkey(key, userid, keysz);
    
    /* otherwise, update based on src key, replace with data in rec */
    memset (data.filler, 0, sizeof(data.filler)); /* for future use make it zero */
    Memory rec((char *)&data, sizeof(data));
    fresp fret = fuser.iatomicupdate(key, rec); // update knows key size
    return frtoret(fret);
  }


/*********************** Folder table ************************/

tfolder::tfolder(FlatConnection &f, ConfigCraft &Ar, NoteCraft &Ls) : dbc(f), ar(Ar), ls(Ls)
  {
    fn = DBROOTDIR;
    fn.addstr(TXT_FOLDERFILE);

    ffolder = dbc.Open(fn, sizeof(folderheader), sizeof(folderrec), TRUE,
                       getkeysize());
  }

void tfolder::checkversion(void)
  {
    // make sure the version in the file matches what we expect.
    Memory header;
    ffolder.getheader(header); 
    folderheader *h = (folderheader *)(char *)header;
    if (h->version == 0) 
      {
        h->version = TXT_FOLDERFILEVERSION;
        ffolder.setheader(header); /* write it back out */
      }
    if (h->version != TXT_FOLDERFILEVERSION)
      {
        ls(NOTESEV) << "***" << endl;
        ls(NOTESEV) << "Incorrect file version for folder file. Blocking." << endl;
        ls(NOTESEV) << "***" << endl;
        while (1)
          sleep(60);
      }
  }

int tfolder::getkeysize(void)
  {
    return sizeof(data.userid) + sizeof(data.foldernamekey);
  }

int tfolder::lookup(UINT userid, char *Foldername, int exact)
  {
    /* Look up this foldername for this user
       returns RET_OK if it worked and will fill data structure.
       RET_NOT_FOUND or RET_ERROR
       If exact is false, then doa fuzzy search */

    Memory key;
    int keysz;
    makefolderkey(key, userid, Foldername, keysz);
    if (exact == FALSE)
      {
        // lookup key only up to end of foldername so we can look
        // for this folder or any children, this is for delete folder
        keysz = sizeof(data.userid) + strlen(Foldername); // only look as far as this foldername
      }
    Memory resp;
    fresp fret = ffolder.iseek(key, resp, keysz);
    if (fret == FR_OK)
      memcpy (&data, (char *)resp, sizeof(data));
    return frtoret(fret);
  }

void tfolder::makefolderkey(Memory &key, UINT userid, char *Foldername, int &keysz)
  {
    keysz = getkeysize();
    Memory k(keysz);
    memcpy (k, &userid, sizeof(userid));
    Memory fn(Foldername);
    lowdo(fn); // key is in lower case
    memorytofield(k+sizeof(userid), fn, sizeof(data.foldernamekey));
    key = k;
  }

int tfolder::add(UINT userid, char *Foldername)
  {
    /* Get a new folderid, see if this folder exists,
       and if not, then add it. */
 
    /* If folder is already there, atomic insert will return duplicate */    
      
    Memory fn(Foldername);
    lowdo(fn); // for key
    
    data.userid = userid;
    memorytofield(data.foldernamekey, fn, sizeof(data.foldernamekey), LOWER);
    memorytofield(data.foldername, Foldername, sizeof(data.foldername));
    
    /* use folderaddmutex to get next higher id */
    Memory header;
    ffolder.getheader(header);
    folderheader *h = (folderheader *)(char *)header;
    h->highestid++;
    data.folderid = h->highestid;
    ffolder.setheader(header); /* write it back out */
    
    /* Now add the row. */
    memset (data.filler, 0, sizeof(data.filler)); /* for future use make it zero */
    Memory rec((char *)&data, sizeof(data));
    fresp fret = ffolder.iatomicinsert(rec);

    return frtoret(fret);
  }

int tfolder::del(UINT userid, char *Foldername)
  {
    /* Look up this foldername record and delete it */
    /* I realize a boo-boo here. This folder can have subfolders
       I should delete those too just as I am deleting all
       of this folder's messages. */

    int ret = lookup(userid, Foldername, FALSE); // look up the row, fuzzy search
    while (ret == RET_OK)
      {
        /* If it's here we can delete it */
        /* Delete all of his messages first */
        tmsgptr m(dbc, ar, ls); // defstore connection for user
        ret = m.deluserfolder(userid, data.folderid);
        if (ret != RET_OK)
          return ret; // should we try to do the rest?

        Memory key;
        int keysz;
        makefolderkey(key, userid, Foldername, keysz);
        
        fresp fret = ffolder.idelete(key, keysz);
        if (fret != FR_OK)
          return frtoret(fret);
          
        // see if there's any other folders with this root foldername
        ret = lookup(userid, Foldername, FALSE); // look up the row, fuzzy
      }
    if (ret == RET_NOTFOUND)
      ret = RET_OK; // running out of folders is a good thing
    return ret; // otherwise return error
  }

int tfolder::deluser(UINT userid)
  {
    /* Delete all the folders for this user */

    /* Delete all of his messages first */
    tmsgptr m(dbc, ar, ls);
    int ret = m.deluser(userid);
    if (ret != RET_PARTIAL)
      if (ret != RET_OK)
        return ret;

    /* Delete all of his folder records. */
    Memory key(sizeof(userid));
    memcpy (key, &userid, sizeof(userid));
    fresp fret = ffolder.idelete(key, sizeof(userid)); // specify shorter than default length

    /* Of course it would be quicker to make one
       big delete request to the flatserver, but not today. */

    /* This will eventually yield FR_NOTFOUND or ERROR */
    while (fret == FR_OK)
      fret = ffolder.idelete(key, sizeof(userid));

    if (ret == RET_PARTIAL) // from above
      return RET_PARTIAL;
    ret = frtoret(fret);
    if (ret == RET_NOTFOUND) // end of folder list is okay
      ret = RET_OK;
    return ret;
  }

int tfolder::update(UINT userid, char *Foldername)
  {
    /* The only thing you can update is the foldername 
       Check first and make sure new folder doesn't already exist,
       if it's different. */

    /* See if user already exists, return RET_DUPLICATE if yes.. */
    
    if (strcasecmp(Foldername, data.foldername) == 0)
      {
        /* If they're not changeing the name, nothing to do. */
        return RET_OK;
      }
      
    /* check for existance of new name folder */
    Memory fn; // the destination
    fieldtomemory(fn, data.foldername, sizeof(data.foldername));

    /* otherwise, update with src key */
    Memory srckey;
    int keysz;
    makefolderkey(srckey, userid, Foldername, keysz);
    Memory rec((char *)&data, sizeof(data));
    fresp fret = ffolder.iatomicupdate(srckey, rec);
    return frtoret(fret);
  }

int tfolder::getfolderlist(UINT userid, zlist &list)
  {
    /* Do a select on everything in the folder list for this user */

    Memory key;
    int sz;
    makefolderkey(key, userid, "", sz);
    fresp ret = ffolder.iselect(key, list, sizeof(userid)); // only match the userid
    return frtoret(ret);
  }


/*********************** Message pointer table ************************/

tmsgptr::tmsgptr(FlatConnection &f, ConfigCraft &Ar, NoteCraft &Ls) : dbc(f), ar(Ar), ls(Ls)
  {
    fn = DBROOTDIR;
    fn.addstr(TXT_MSGPTRFILE);

    fmsgptr = dbc.Open(fn, sizeof(msgptrheader), sizeof(msgptrrec), TRUE,
                       getkeysize());
  }

void tmsgptr::checkversion(void)
  {
    // make sure the version in the file matches what we expect.
    Memory header;
    fmsgptr.getheader(header); 
    msgptrheader *h = (msgptrheader *)(char *)header;
    if (h->version == 0) 
      {
        h->version = TXT_MSGPTRFILEVERSION;
        fmsgptr.setheader(header); /* write it back out */
      }
    if (h->version != TXT_MSGPTRFILEVERSION)
      {
        ls(NOTESEV) << "***" << endl;
        ls(NOTESEV) << "Incorrect file version for msgptr file. Blocking." << endl;
        ls(NOTESEV) << "***" << endl;
        while (1)
          sleep(60);
      }
  }

int tmsgptr::getkeysize(void)
  { 
    return sizeof(data.userid) + sizeof(data.folderid) + sizeof(data.msgid);
  }

int tmsgptr::lookup(UINT userid, UINT folderid, char *msgid, int exact)
  {
    /* Look up this msgptr for this user
       returns RET_OK if it worked and will fill data structure.
       RET_NOT_FOUND or RET_ERROR */

    Memory key;
    int keysz;
    makemsgptrkey(key, userid, folderid, msgid, keysz);
    if (exact == FALSE)
      {
        // lookup key only up to userid and folderid so we can look
        // quickly to see if the folder's empty
        keysz = sizeof(data.userid) + sizeof(data.folderid);
      }
    
    Memory resp;
    fresp fret = fmsgptr.iseek(key, resp, keysz);
    if (fret == FR_OK)
      memcpy (&data, (char *)resp, sizeof(data));
    return frtoret(fret);
  }

int tmsgptr::lookupuserfolder(UINT userid, UINT folderid, zlist &list)
  {
    /* For loading a mailbox. Do a select on userid and folderid and fill up
       msglist with the records. */

    int ksize = sizeof(userid) + sizeof(folderid);
    Memory key(ksize);
    memcpy (key, &userid, sizeof(userid));
    memcpy (key+sizeof(userid), &folderid, sizeof(folderid));

    fresp ret = fmsgptr.iselect(key, list, ksize);
    if ((ret != FR_OK) && (ret != FR_NOTFOUND))
      return frtoret(ret);
    
    return RET_OK;
  }

void tmsgptr::makemsgptrkey(Memory &key, UINT userid, UINT folderid, char *msgid, int &keysz)
  {
    keysz = getkeysize();
    Memory k(keysz);
    memcpy (k, &userid, sizeof(userid));
    memcpy (k + sizeof(userid), &folderid, sizeof(folderid));
    Memory id(msgid);
    memorytofield(k + sizeof(userid) + sizeof(folderid), id, sizeof(data.msgid));
    key = k;
  }

int tmsgptr::add(UINT userid, UINT folderid, char *msgid)
  {
    /* See if this msgptr exists, if not, then add it. */
    /* atomic insert takes care of duplicate checking for us */
    data.userid = userid;
    data.folderid = folderid;
    memorytofield(data.msgid, msgid, sizeof(data.msgid));
    // all other fields are presumed set.
    
    // no sequence number we presume msgid is world unique
    /* Now add the row. */
    memset (data.filler, 0, sizeof(data.filler)); /* for future use make it zero */
    Memory rec((char *)&data, sizeof(data));
    fresp fret = fmsgptr.iatomicinsert(rec);

    return frtoret(fret);
  }

int tmsgptr::del(UINT userid, UINT folderid, char *msgid, int dofilealso)
  {
    /* Look up this message and delete it */

    int ret = lookup(userid, folderid, msgid); // look up the row
    if (ret == RET_OK)
      {
        /* If it's here we can delete it */
        /* We must tell the message storage manager to delete
           our reference to the message. */
        /* when recovering from a copy error we need to just delete
           the row and not rename/delete the file */
        if (dofilealso != FALSE)
          {
            Memory msgdefstore;
            Memory msgid;
            fieldtomemory(msgdefstore, data.defstore, sizeof(data.defstore));
            fieldtomemory(msgid, data.msgid, sizeof(data.msgid));
            msgstore ms(ar, ls, msgdefstore, msgid, FALSE); /* independant of database */
            ret = ms.remove(); /* world unique, remember? */
            if (ret != RET_OK)
              return ret;
          }

        Memory key;
        int keysz;
        makemsgptrkey(key, userid, folderid, msgid, keysz);
        
        fresp fret = fmsgptr.idelete(key, keysz);
        return frtoret(fret);
      }
    return ret; // not found or error
  }

int tmsgptr::deluser(UINT userid)
  {
    /* Delete all the msgptrs for this user */
    /* Delete the associated messages as we do it.
       This is slow and annoying because we have to load
       the record so we can get its defstore. */

    Memory key(sizeof(userid));
    memcpy (key, &userid, sizeof(userid));

    Memory resp;
    fresp fret = fmsgptr.iseek(key, resp, sizeof(userid)); // specify shorter key length than default
    while (fret == FR_OK)
      {
        memcpy (&data, (char *)resp, sizeof(data));

        Memory msgdefstore;
        Memory msgid;
        fieldtomemory(msgdefstore, data.defstore, sizeof(data.defstore));
        fieldtomemory(msgid, data.msgid, sizeof(data.msgid));
        msgstore ms(ar, ls, msgdefstore, msgid, FALSE); /* independant of database */
        int ret = ms.remove();
        if (ret != RET_OK)
          return RET_MSGSTORE_ERROR;

        Memory key; // make key for this specific message we just deleted
        int keysz;
        makemsgptrkey(key, data.userid, data.folderid, data.msgid, keysz);
        fret = fmsgptr.idelete(key, keysz); 

        /* Find the next one. */
        fret = fmsgptr.iseek(key, resp, sizeof(userid)); // but seek for just any records for this user
      } 
    int ret = frtoret(fret);
    if (ret == RET_NOTFOUND)
      ret = RET_OK; // when bulk deleting it's not okay to have no folders, but it's okay to have no messages
    return ret; // ok or error, no not found
  }

int tmsgptr::deluserfolder(UINT userid, UINT folderid)
  {
    /* Delete all the msgptrs for this user in this folder */
    /* Delete the associated messages as we do it. */

    Memory key(sizeof(userid) + sizeof(folderid));
    memcpy (key, &userid, sizeof(userid));
    memcpy (key+sizeof(userid), &folderid, sizeof(folderid));

    Memory resp;
    fresp fret = fmsgptr.iseek(key, resp, sizeof(userid)+sizeof(folderid)); // specify shorter key length than default
    while (fret == FR_OK)
      {
        /* Here, we seek for anything in the folder, then get the whole key, so we 
           can delete THAT one's message and msgptr. */
        memcpy (&data, (char *)resp, sizeof(data));
        Memory msgdefstore;
        Memory msgid;
        fieldtomemory(msgdefstore, data.defstore, sizeof(data.defstore));
        fieldtomemory(msgid, data.msgid, sizeof(data.msgid));
        msgstore ms(ar, ls, msgdefstore, msgid, FALSE); /* independant of database */
        int ret = ms.remove();
        if (ret != RET_OK)
          return RET_MSGSTORE_ERROR;
        Memory key; // delete this specific message
        int keysz;
        makemsgptrkey(key, data.userid, data.folderid, data.msgid, keysz);
        fret = fmsgptr.idelete(key, keysz); // specify enyire key

        /* Find the next one. */
        fret = fmsgptr.iseek(key, resp, sizeof(userid)+sizeof(folderid));
      }
    int ret = frtoret(fret);
    if (ret == RET_NOTFOUND)
      return RET_OK; // when bulk deleting it's not okay to have no folders, but it's okay to have no messages
    return frtoret(fret); // not found or error
  }

int tmsgptr::update(UINT userid, UINT folderid, char *msgid)
  {
    /* The three fields passed in are the key, and 
       form the unique id for this message for this user in
       this folder and therefore can not be changed.
       Hmm... I suppose if you want to move the message to
       another folder I should allow that.
       But the userid and msgid HAVE to stay the same. */
    /* Because of all of the problems it causes, we will
       not allow more than one copy of the same exact message
       in the same folder. */

    /* See if message already exists, return RET_DUPLICATE if does..
       In this case, we need the mutex because there's a small
       window between where we check and where we do the update.
       I now realize the problem is that there's no iatomicupdate in 
       the flatserver which should check for the update about to cause
       a duplicate key. I think I'll add that. */

    /* I had it backwards, you call this fucntion with the src key fields
       you want to update (The old ones) and data has the new fields
       to update it with */

    /* These must remain the same. */
    data.userid = userid; 
    /* They can screw us here by passing a short buffer, how to check? */
    /* But they're not allowed to change the message id */
    memorytofield(data.msgid, msgid, sizeof(data.msgid));

    /* They're updating something or moving to somewhere safe,
       do the update */

    /* otherwise, update with src key */
    Memory srckey;
    int keysz;
    // data already had correct data to be inserted, make key based on values
    // passed in.
    makemsgptrkey(srckey, userid, folderid, msgid, keysz);
    Memory rec((char *)&data, sizeof(data));
    fresp fret = fmsgptr.iatomicupdate(srckey, rec);
    return frtoret(fret);
  }

/*********************** Resource table ************************/

tresource::tresource(FlatConnection &f, ConfigCraft &Ar, NoteCraft &Ls) : dbc(f), ar(Ar), ls(Ls)
  {
    fn = DBROOTDIR;
    fn.addstr(TXT_RESOURCEFILE);

    fresource = dbc.Open(fn, sizeof(resourceheader), sizeof(resourcerec), 
                         TRUE, getkeysize());
  }

void tresource::checkversion(void)
  {
    // make sure the version in the file matches what we expect.
    Memory header;
    fresource.getheader(header); 
    resourceheader *h = (resourceheader *)(char *)header;
    if (h->version == 0) 
      {
        h->version = TXT_RESOURCEFILEVERSION;
        fresource.setheader(header); /* write it back out */
      }
    if (h->version != TXT_RESOURCEFILEVERSION)
      {
        ls(NOTESEV) << "***" << endl;
        ls(NOTESEV) << "Incorrect file version for resource file. Blocking." << endl;
        ls(NOTESEV) << "***" << endl;
        while (1)
          sleep(60);
      }
  }

int tresource::getkeysize(void)
  {
    return sizeof(data.key);
  }

int tresource::lookup(char *rkey)
  {
    /* Look up this resource by key
       returns RET_OK if it worked and will fill data structure.
       RET_NOT_FOUND or RET_ERROR */

    Memory key;
    int keysz;
    makeresourcekey(key, rkey, keysz);
    
    Memory resp;
    fresp ret = fresource.iseek(key, resp, keysz);
    if (ret == FR_OK)
      memcpy (&data, (char *)resp, sizeof(data));
    return frtoret(ret);
  }

void tresource::makeresourcekey(Memory &key, char *rkey, int &keysz)
  {
    keysz = getkeysize();
    Memory k(keysz);
    Memory fn(rkey);
    lowdo(fn); // key is in lower case
    memorytofield(k, fn, sizeof(data.key));
    key = k;
  }

int tresource::add(char *rkey)
  {
    /* See if this resource exists,and if not, then add it. */
    
    Memory fn(rkey);
    lowdo(fn); // for key
    
    memorytofield(data.key, fn, sizeof(data.key));
    
    /* Now add the row. */
    memset (data.filler, 0, sizeof(data.filler)); /* for future use make it zero */
    Memory rec((char *)&data, sizeof(data));
    fresp fret = fresource.iatomicinsert(rec);

    return frtoret(fret);
  }

int tresource::del(char *rkey)
  {
    /* Look up this key and delete it */

    int ret = lookup(rkey); // look up the row
    if (ret == RET_OK)
      {
        /* If it's here we can delete it */
        Memory key;
        int keysz;
        makeresourcekey(key, rkey, keysz);
        
        fresp fret = fresource.idelete(key, keysz);
        return frtoret(fret);
      }
    return ret; // not found or error
  }

int tresource::update(char *rkey)
  {
    /* Update value for this key. */

    Memory key;
    int keysz;
    makeresourcekey(key, rkey, keysz);
    memorytofield(data.key, rkey, sizeof(data.key));

    Memory rec((char *)&data, sizeof(data));
    fresp fret = fresource.iatomicupdate(key, rec);
    return frtoret(fret);
  }

int tresource::getdomainlist(zlist &list)
  {
    /* Do a select on DOMAIN- and load up list with it. */
    /* The value has just the name of the domain, the key has it also. */
    // DOMAIN-nyti.spaghetti.com        nyti.spaghetti.com 
    
    zlist dt;
    fresp ret = fresource.iselect(TXT_DOMAINPREFIX, dt, strlen(TXT_DOMAINPREFIX)); // get every domain record
    if ((ret != FR_OK) && (ret != FR_NOTFOUND))
      return frtoret(ret);
    
    // now copy in just the domain names without the prefix.
    list.WipeList();
    int lp;
    for (lp = 0; lp < dt.ListCount(); lp++)
      {
        stringentry *s = (stringentry *)dt.GetItem(lp);
        resourcerec *r = (resourcerec *)(char *)s->getstr();

        Memory dom;
        fieldtomemory(dom, r->value, sizeof(r->value));
        lowdo(dom); // all domains are lowercase
        stringentry *z = new stringentry (dom);
        list.Add(z);
      }
    return RET_OK;
  }





