/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <dirent.h>
#include <sys/stat.h>
#include <errno.h>

#include <Q/memm.h>
#include <Q/configcraft.h>
#include <Q/notecraft.h>
#include "global.h"
#include "msgstore.h"


/* Our message store mechanism.
   File system based. Given a 6 character locator and a message id (SZ_MSGID=100)
   we can read from and write to our message store.
   We support reference counts to save on multiple delivery.
   The store command will return you yes or no if you have to send data
   or did we just up the ref count. */
   

msgstore::msgstore(ConfigCraft &ar, NoteCraft &Ls, char *Defstore, char *Msgid, int flushoften) : ls(Ls)
  {
    f = NULL;
    /* We figure out the root directory by looking it up in the
       config file. by the message defstore, we don't need ar otherwise */
    rootdir = ar[Defstore]; 
    trim(rootdir);
    if (strlen(rootdir) == 0)
      ls(NOTEERR) << "msgdefstore " << Defstore << " is not configured. " << endl;
    addslash(rootdir);
    defstore = Defstore;
    msgid = Msgid;
    reading = TRUE; // assume this
    forceflush = flushoften;
    /* in the case where we wrote a blank file with the idea that we
       were going to be writing to it soon, but nobody ever did. */
    badfile = FALSE;
    badfilename = "";
  }
        
msgstore::~msgstore(void)
  {
    if (f)
      fclose(f);
    if (badfile)
      unlink(badfilename);
  }

void msgstore::statbuild(char *fn)
  {
    /* See if this directory exists, if not, create it. */
    struct stat s;
    int ret = stat(fn, &s);
    if (ret == -1)
      if ((errno == ENOENT) || (errno == ENOTDIR))
        { // then create it.
          if (mkdir(fn, CREATEMODE) == -1)
            ls(NOTEERR) << "Unable to create: " << fn << " Error: " << errno << " " << strerror(errno) << endl;
        }
    // in case of other error they'll find out when
    // they go to open the file.
  }

void msgstore::getrootdir(Memory &rd, int checkdir)
  {
    // generate the directory name that the msg file will go in
    /* The last character of the 6 in defstore is the number of 
       directory levels depth to go. */
    /* If checkdir is true then we're being called for a first
       time write, verify that all of the directories exist
       and make them if they don't. We presume the rootdir exists. */
    Memory fn = rootdir;
    int n = 0;
    if (strlen(defstore) >= 6)
      {
        char s[2];
        s[0] = *(defstore+5);
        s[1] = '\0';
        n = val(s);
      }
    int lp;
    for (lp = 0; lp < n; lp++)
      {
        fn.addchar(*(msgid+lp));
        fn.addchar('/');
        if (checkdir != FALSE)
          statbuild(fn);
      }
    rd = fn;
  }

void msgstore::getfn(Memory &fn, int checkdir)
  {
    /* Get the full filename of the message file starting
       with the text of the message id.
       I.E. if msgid = "abcdefg" find "abcdefg-refcount" */
       
    Memory rd;
    getrootdir(rd, checkdir);
    fn = "";
    /* Now look it up and find the ref count info */
    DIR *dir = opendir (rd);
    if (dir == NULL)
      { 
        ls(NOTEERR) << "Unable to open directory: " << rd << " Error: " << errno << " " << strerror(errno) << endl;
        return;
      }

    struct dirent *p = readdir(dir);
    int n = strlen(msgid); // just check the prefix part
    while (p)
      {
        if (strncmp(p->d_name, msgid, n) == 0)
          {
            fn = rd;
            fn.addstr(p->d_name);
            closedir(dir);
            return;
          }
        p = readdir(dir);
      }
    closedir(dir);
   
    /* If we didn't find anything, it wasn't there. Make a new one. */
    fn = rd;
    fn.addstr(msgid);
    fn.addstr("-0"); // refcount = 0, so we can tell it's new
  }

int msgstore::getrefcount(Memory &fn, Memory *rootfn)
  {
    /* return the refcount value at the end of the string
       and optionnally return the file without the
       extention. If rootfn != NULL */
    char *p;
    if (rootfn != NULL)
      {
        *rootfn = fn; // make copy that we can mangle.
        p = strrchr (*rootfn, '-');
        if (p != NULL)
          *p = '\0'; // make it end of string
      }
    else
      p = strrchr (fn, '-');
    if (p == NULL)
      {
        // we've got problems.
        return 0;
      }
    return val(p+1); 
  }

int msgstore::increfcount(Memory &fn, int *newref)
  { /* move the filename to the next higher number ref count */
    /* return the new filename and the new refcount if newref not NULL */
    char buf[10];
    Memory nfn;
    Memory oldfn = fn;
    int c = getrefcount(fn, &nfn);
    c++;
    if (newref != NULL)
      *newref = c;
    sprintf(buf, "-%d", c);
    nfn.addstr(buf);
    fn = nfn; 
    if (c == 1) // file doesn't exist yet to rename
      return RET_OK; // we're done in this case.
    int ret = rename (oldfn, nfn);
    if (ret == 0)
      return RET_OK;
    ls(NOTEERR) << "Unable to rename " << (char *)oldfn << " to " << (char *)nfn << " Error: " << errno << " " << strerror(errno) << endl;
    return RET_ERROR;
  }
  
int msgstore::decrefcount(Memory &fn, int *newref)
  {
    /* drop the refcount on this file if it is
       or hits zero, delete the file. If not null,
       return the new value in newref */
    char buf[10];
    Memory nfn;
    int ret;
    
    int c = getrefcount(fn, &nfn);
    if (c < 2) // 1 or 0 already?
      {
        if (newref != NULL)
          *newref = 0;
        ret = unlink(fn);
        if (ret == 0)
          return RET_OK;
        ls(NOTEERR) << "Unable to unlink " << (char *)fn << " Error: " << errno << " " << strerror(errno) << endl;
        return RET_ERROR;
      }
    c--;
    if (newref != NULL)
      *newref = c;
    sprintf(buf, "-%d", c);
    nfn.addstr(buf);
    ret = rename (fn, nfn);
    if (ret == 0)
      return RET_OK;
    ls(NOTEERR) << "Unable to rename " << (char *)fn << " to " << (char *)nfn << " Error: " << errno << " " << strerror(errno) << endl;
    return RET_ERROR;
  }

int msgstore::store(void)
  {
    /* A process wishing to store a message (delivery, copy) should
       call this first. If it returns RET_OK then the store was done
       (we inced the refcount) if it returned RET_NEEDDATA then
       the caller must call the write function to store the data. */
    /* It is also here that we force the directory creation if need be */
    /* If we need data We also write a blank file and put it the
       bad state in case they never come back and write to it. */
    Memory fn;
    int count;
    
    getfn(fn, TRUE); // build dir tree
    if (strlen(fn) == 0) // error of some kind 
      return RET_ERROR;

    int ret = increfcount(fn, &count); // renames file if count > 1
    if (ret == RET_OK)
      {
        if (count == 1) // new file
          {
            reading = FALSE;
            // create the file and put it in a bad state
            // we can also get a head start here since we're pretty sure
            // they're going to write soon.
            f = fopen(fn, TXT_FOPENWRITE); // with the -1 ref count
            /* If this is NULL, then when we go to ms.write() it will
               make a new file ref count 0 not 1, so let them know now
               that they can't write. */
            if (f == NULL)
              {
                ls(NOTEERR) << "Error creating message file: " << (char *)fn << " Error " << errno << " " << strerror(errno) << endl;
                return RET_ERROR;
              }
            badfilename = fn;
            badfile = TRUE;
            return RET_NEEDDATA;
          }
        return RET_OK; // file already exists, we're done
      }
    return ret;
  }
  
int msgstore::remove(void)
  { /* drops ref count, deletes if at 0 */
    Memory fn;
    getfn(fn);
    if (strlen(fn) == 0) // error of some kind 
      return RET_ERROR;
    return decrefcount(fn, NULL);
  }
  
int msgstore::getsize(UINT &size)
  {
    // return the data size of the message
    Memory fn;
    getfn(fn);
    if (strlen(fn) == 0) // error of some kind 
      return RET_ERROR;

    struct stat buf;
    int ret = stat(fn, &buf);
    if (ret == -1)
      return RET_ERROR;
      
    size = buf.st_size;
    return RET_OK;
  }

int msgstore::seek(UINT pos)
  { // from beginning, forces read mode
    
    if (f != NULL)
      if (reading == FALSE)
        { // if writing, then close, so we can reopen for read
          fclose (f);
          f = NULL;
        }
    
    reading = TRUE;
    if (f == NULL)
      {
        Memory fn;
        getfn(fn);
        f = fopen(fn, TXT_FOPENREAD);
        if (f == NULL)
          {
            ls(NOTEERR) << "Unable to open " << (char *)fn << " Error: " << errno << " " << strerror(errno) << endl;
            return RET_ERROR;
          }
      }
    int ret = fseek(f, pos, SEEK_SET);
    if (ret == 0)
      return RET_OK;
    ls(NOTEERR) << "msg seek error " << " Error: " << errno << " " << strerror(errno) << endl;
    return RET_ERROR;
  }
    
int msgstore::read(char *outbuf, int amount, int &got)
  {
    /* Read a chunk of file into outbuf of amount amount.
       If this is the first time in, or we were writing, 
       we start from the beginning. */
    if (f != NULL)
      if (reading == FALSE)
        { // if writing, then close, so we can reopen for read
          fclose (f);
          f = NULL;
        }
    
    reading = TRUE;
    if (f == NULL)
      {
        Memory fn;
        getfn(fn);
        f = fopen(fn, TXT_FOPENREAD);
        if (f == NULL)
          {
            ls(NOTEERR) << "Unable to open: " << (char *)fn << " Error: " << errno << " " << strerror(errno) << endl;
            return RET_ERROR;
          }
      }
    int ret = fread(outbuf, 1, amount, f);
    got = ret; // valid only if RET_OK or RET_PARTIAL
    if (ret == amount)
      return RET_OK;
    if (ret < amount)
      {
        if (feof(f))
          return RET_PARTIAL; // okay but not all
      }
    // all other cases are error.
    ls(NOTEERR) << "Unable to read from message file. Error: " << errno << " " << strerror(errno) << endl;
    return RET_ERROR;
  }
  
int msgstore::write(char *inbuf, int amount)
  {
    /* Open for write or append from current write position. */
    if (f != NULL)
      if (reading != FALSE)
        { // if reading, then close, so we can reopen for write
          // this will wipe the file. YOU ONLY DO THIS ONCE
          fclose (f);
          f = NULL;
        }
    
    reading = FALSE;
    if (f == NULL)
      {
        Memory fn;
        getfn(fn);
        f = fopen(fn, TXT_FOPENWRITE);
        if (f == NULL)
          {
            ls(NOTEERR) << "Unable to create: " << (char *)fn << " Error: " << errno << " " << strerror(errno) << endl;
            return RET_ERROR;
          }
      }
    int ret = fwrite(inbuf, 1, amount, f);
    if (ret == amount)
      return RET_OK;
    ls(NOTEERR) << "Unable to write to message file. Error: " << errno << " " << strerror(errno) << endl;
    return RET_ERROR;
  }

int msgstore::close(void)
  {
    int ret;
    badfile = FALSE; // I can't think of a better time, the caller thinks its okay
    if (f == NULL)
      return RET_OK;
    ret = fclose(f);
    f = NULL;
    reading = TRUE;
    if (ret == 0)
      return RET_OK;
    ls(NOTEERR) << "Unable to close message file. Error: " << errno << " " << strerror(errno) << endl;
    return RET_ERROR;
  }


/* This will be a brail like class someday. */

int msgstore::readline(Memory &mline)
  { 
    /* read a line from the message, return RET_OK if we got a valid
       line, and RET_ERROR if not. */
    
    UINT count = 0;

    ULONG lsize;
    memcpy (&lsize, ((char *)mline)-4, 4); /* get long back. We're cheating, we know how Memory works. */
    char *line = mline; // movable positional pointer
    int retval = RET_OK;
    while (1)
      {
        char b;
        int got;
        int ret = read(&b, 1, got);
        if (ret == RET_PARTIAL)
          {
            // we got nothing back, and that's the end of the file
            retval = RET_PARTIAL;
            break;
          }
        if (ret != RET_OK) 
          return RET_ERROR; // that means we got a real error
          
        count++;
        if (b == '\n')
          break;
        if (b == '\r')
          {
            read(&b, 1, got); /* get the assumed \n */
            break;
          } 

        // check the buffer size
        if (count >= lsize)
          {
            UINT pos = line - mline; // take note of where we are in case the ptr changes
            ULONG need = lsize * 2; // double each time.
            mline.resize(need);
            memcpy (&lsize, ((char *)mline)-4, 4); /* get new size back. */
            line = mline + pos; // put the ptr back where it was in (possibly) new buffer
          }
        *(line++) = b;
      } 
    *line = '\0'; /* n.t. */
    return retval; 
  } 
