/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>

#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/acceptor.h>
#include <Q/socketio.h>
#include <Q/notecraft.h>
#include <Q/configcraft.h>
#include <Q/filelist.h>
#include <Q/hashtable.h>
#include <flat/fc.h>

#include "release.h"
#include "global.h"
#include "db.h"
#include "io.h"
#include "dbfinder.h"
#include "utils.h"
#include "msgstore.h"
#include "pop.h"


/* This is the heavily recursing mime parser.
   It took me a lot longer than it should be it seems to work.
   The basic idea of numbering scheme is that you add another 
   dot level when you recurse, and you recurse when you get 
   to a multipart or a message content-type.
   In this way, you can pull out a whole message, a whole embedded
   message or a part of an [embedded] message. */

/* this should really be its own class, but it's not.   
   if it was its own class, we wouldn't have to pass so many parameters.
  
   I'm not terribly happy about this, I may clean it up since I decided to 
   put some variables in the pop class anyway.
   But it works.
   The basic idea is that it has two modes: display part list and
   display part content. Except for the if's to function in
   a specific mode, it is hack free, if somewhat noisy.
   No butchering of data, and if you follow the flow, where I make
   exceptions for the oddball way of dealing with mime (submessages
   are recursed into) everything is laid consistently to that end.
   And best of all, it's relativly compact and it works.
   Someday I'll rip it out to be standalone.
   
   optimizations:
   1) don't recalc current part number each line
   2) use a brail version of message line reader
   
   Every request for a partial message requires the entire message being scanned.
   This could be fixed up to stop short, or if we ever go nutz, build a mime 
   parser index. and have all the lookups use that.
*/

/****************************** POP 4 *************************/

/* This function will romp through the message looking and recursing 
   for all of the parts and display their current message part number */
   
void pop::processlprt(Memory &l)
  {
    char *p = strchr (l, ' '); // find msg number
    int badparm = FALSE;
    Memory mid;
    
    if (gettoken(p, mid) == FALSE)
      badparm = TRUE;
      
    if (badparm)
      {
        response ("-ERR LPRT command requires one parameter.");
        return;
      }

    int max = msglist.ListCount();
    int v = getmsgpos(mid); // get the number or uidl converted to a number
    if ((v < 1) || (v > max))
      {
        response ("-ERR Invalid msg id argument to LPRT command.");
        return;
      }

    stringentry *s = (stringentry *)msglist.GetItem(v-1); // zlist counts from 0
    msgptrrec *m = (msgptrrec *)(char *)s->getstr();
    if ((m->status & MSGSTATUS_TMPDEL) != 0)
      {
        Memory resp(100);
        sprintf (resp, "-ERR Message %d has been deleted.", v);
        response (resp);
        return;
      }

    /* Make sure we can get at least some part of the message before we say okay. */
    Memory msgdefstore;
    fieldtomemory(msgdefstore, m->defstore, sizeof(m->defstore)); // this is the MESSAGE's defstore not the user's mailstore defstore
    Memory msgid;
    fieldtomemory(msgid, m->msgid, sizeof(m->msgid));

    ls(NOTEINF) << "Retrieving partslist for message: " << (char *)msgid << " on msgdefstore " << (char *)msgdefstore << endl;
    msgstore ms(ar, ls, msgdefstore, msgid, FALSE);
    
    /* As a major optomization, it would be a good idea to write a brail 
       message store layer class. Wouldn't be too hard, since we have brail
       and msgstore implements all the functions that brail calls... */

    Memory readbuffer; // for unfolding
    int nextret = RET_OK; // for unfolding work
    // we have to prime the unfold buffer
    int ret = ms.readline(readbuffer); // slow slow ugh slow.
    // and while doing that, make sure we can read the file at all.
    if ((ret != RET_OK) && (ret != RET_PARTIAL))
      {
        ls (NOTESEV) << "Error reading message " << (char *)msgid << " from message store " << (char *)msgdefstore << endl;
        Memory resp(100);
        sprintf (resp, "-ERR Temporary error reading message %s.", (char *)msgid);
        response (resp);
        return;
      }

    response ("+OK part list follows.");
    Memory prefix = ""; // start with nothing, partlist will add 1.
    Memory eboundary = ""; // denote to stop at end of file
    int insubmessage = FALSE; // first message is not submessage
    Memory lastline; // never really matters.
    ret = partlist(nextret, readbuffer, ms, prefix, eboundary, insubmessage, lastline);
    if ((ret != RET_OK) && (ret != RET_PARTIAL))
      {
        response ("-ERR Error retrieving message.");
        return;
      }
    response(".");
  }

void pop::processretrpart(int istop, msgstore &ms, int lines, char *findpartname)
  {
    /* this function will prime the partlist function for output of
       partname, but it otherwise just calls partname like lprt does. */
    ls(NOTEINF) << "Retrieving message part: " << findpartname << endl;
    
    Memory readbuffer; // for unfolding
    int nextret = RET_OK; // for unfolding work
    // we have to prime the unfold buffer
    int ret = ms.readline(readbuffer); // slow slow ugh slow.
    // and while doing that, make sure we can read the file at all.
    if ((ret != RET_OK) && (ret != RET_PARTIAL))
      {
        ls (NOTESEV) << "Error reading message" << endl;
        Memory resp(100);
        sprintf (resp, "-ERR Temporary error reading message.");
        response (resp);
        return;
      }

    // we don't say okay until we've actually found that part of the message they're asking for
    foundpart = FALSE;
    linesout = 0; // how many we've output so far
    
    Memory prefix = ""; // start with nothing, partlist will add 1.
    Memory eboundary = ""; // denote to stop at end of file
    int insubmessage = FALSE; // first message is not submessage
    Memory lastline; // never really matters.
    ret = partlist(nextret, readbuffer, ms, prefix, eboundary, insubmessage, lastline, findpartname, lines);
    if ((ret != RET_OK) && (ret != RET_PARTIAL))
      {
        response ("-ERR Error retrieving message.");
        return;
      }
    if (foundpart == FALSE)
      response("-ERR Part not found.");
    else
      response(".");
  }
   
int pop::partlist(int &nextret, Memory &readbuffer, msgstore &ms, Memory prefix, Memory boundary, int insubmessage, Memory &returnlastline, char *findpartname, int lines)
  {
    /* We are at the top of a message, so we are going to start
       with adding 1 and then 1. to prefix and printing it
       we will recurse when we hit a email attachment
       with the current prefix ("2.") and the boundary line to
       return at. */
    
    /* The last two params are a hack, because it was either that or make a callback
       This makes my nice clean mime parser ugly, because it now does two
       different things. Oh well. It was nice while it lasted. */

    Memory l;
    int ret;
    int inheader = TRUE;
    int partcount = 1;
    int thisissubmessage = FALSE; // keep tabs on this new part's submessage-ness
    // these get defined as we hit the content type of our [sub]part (that
    // we recursed into) 
    Memory ctype;
    Memory dispo;
    int multipart = FALSE;
    /* In main message part, boundary will be blank */
    Memory childboundary;
    Memory endboundary(boundary);
    endboundary.addstr("--");

    /* If we are in retr mode then we have to keep track of the headers for this part
       in case this is the one we want to display, we won't know until we reach the
       end of the headers. Upon exiting the headers, if this is it, dump out the 
       header list and display the rest of this part. */

    while (1)
      {
        Memory realline;
        ret = unfoldreadline(nextret, readbuffer, ms, l, inheader, realline); 
        if (ret == RET_PARTIAL)
          {
            // last line, not much we can do.
            returnlastline = l;
            return RET_PARTIAL; // If this happens when we're nested, then the message is not valid mime.
          }
        if (ret != RET_OK)
          {
            ls (NOTESEV) << "Error reading from message store " << endl;
            return RET_ERROR; // should cause denoting of an error of some kind.
          }

        if ((inheader) && (strlen(l) == 0))
          { // exiting header, see what this part is...
            inheader = FALSE;

            Memory pfx (prefix);
            if (strlen(pfx) > 0)
              pfx.addchar('.'); // first one is 1, second one is 1.2
            pfx.addnum(partcount); 
            
            /* If you're just exiting the header, and we've been
               flagged as multipart, that means we're about to
               recurse, so skip to the first boundary.
               then recurse into the sub part, if it's multpart.
               You'll be skipping the 'this is a multipart mime message' text
            */

            if (multipart != FALSE) // this is only true if this PART is multipart
              { // it gets reset on the next boundary.
                ret = RET_OK;
                while (ret == RET_OK)
                  {
                    // here we look for the first child boundary and then we can recurse to it
                    // whereas we use boundary when looking for siblings

                    // if this is part of a submessage that we want to display, show this too
                    if (findpartname) // the blank line denoting end of header
                      showthisline(prefix, partcount, findpartname, realline, lines, inheader);
                      
                    ret = unfoldreadline(nextret, readbuffer, ms, l, inheader, realline); 

                    if (strcmp(l, childboundary) == 0)
                      break;
                  }
                if (ret != RET_OK)
                  return ret;

                if (findpartname) // the child boundary
                  showthisline(prefix, partcount, findpartname, realline, lines, inheader);
              }

            /* Here is where we print out the part names, we're at
               the end of the header and we know the content type
               to print out. 
               We don't do this after skipping the mime blurb
               because that isn't a separate part, it's
               part of the primary message part. */
            
            if (findpartname == NULL)
              {
                Memory resp(pfx);
                resp.addchar(' ');
                resp.addstr(ctype); // add content type string
                if (strlen(dispo) > 0)
                  {
                    resp.addstr("; " TXT_CONTENT_DISPOSITION " ");
                    resp.addstr(dispo); // this can get long, should we fold it?
                  }
                foldout(resp); // this will be a long line, fold as neccesary
              }
              
            Memory lastline;
            /* If this part is a multipart, then we recurse when we end
               the headers because there will be subpieces for this part
               Recurse to the child piece, since the body is where another
               part can be at */
            if ((multipart) || (thisissubmessage))
              {
                // pass it the boundary, that the siblings have to look for
                /* The fact that we're recursing into an rfc822 message is just
                   a well designed hack. To make it look like we're not,
                   we have to exit on parent's boundary, and point the
                   prefix to be the parents, not the childs. */
                if (thisissubmessage) // submessage has to end on parent boundary. multipart will always be false here...
                  ret = partlist(nextret, readbuffer, ms, pfx, boundary, TRUE, lastline, findpartname, lines);
                else
                  ret = partlist(nextret, readbuffer, ms, pfx, childboundary, FALSE, lastline, findpartname, lines);
                // do not set l to last line because we don't want to reprocess that end marker
                if (ret != RET_OK)
                  return ret;
                l = lastline; // so we can end on the boundary the submessage hit
                              // since it wasn't for it, it was for the next level up.
                // if it was okay, keep looking for siblings, because
                // we're past child parts
              } 
            /* If it's not multipart we just cruise through the sibling parts */ 
          }

        if (inheader)
          {
            // check this line for interesting header information
            char *p = strchr (l, ' ');
            if (p != NULL)
              {
                Memory h(l, p-l); // get header name, lose the space
                if (strcasecmp(h, TXT_CONTENT_DISPOSITION) == 0)
                  {
                    dispo = p; 
                    trim(dispo);
                  }
                if (strcasecmp(h, TXT_CONTENT_TYPE) == 0)
                  {
                    ctype = p;
                    trim(ctype); // save this for printing.
                    parsecontenttype(ctype, childboundary, multipart, thisissubmessage);
                  }
              }
          }

        /* Check and see if this line is our ending border, must
           do this before checking for part boundary because both
           will match.
           We must do it after the recursion call, so when it returns
           we can pick up on end boundaries that recursion came out
           of because. */

        /* Check and see if this line is a mime part boundary */
        int skipleadingborder = FALSE;
        if (strlen(l) > 0)
          {
            if (strcmp(l, endboundary) == 0)
              {
                returnlastline = l; // in case of submessage
                return RET_OK; // we successfully ended this message part.
              }

            if (strcmp(l, boundary) == 0)
              {
                if (insubmessage) // if we came in to this sibling as a child message, we return on part end
                  {
                    returnlastline = l; // in case of submessage
                    return RET_OK;
                  }

                // write out that we found a part
                partcount++;
                /* note that we're at the top of a new message part
                   we have to scan headers, and possibly
                   recurse if this part is complex */
                inheader = TRUE; 
                // clean out fields for new part
                multipart = FALSE;
                thisissubmessage = FALSE;
                ctype = "";
                dispo = "";
                childboundary = "";
                if (foundpart == FALSE)  
                  skipleadingborder = TRUE; // we don't want to print out the border that made us change part numbers.
              }
          }

        /* And last but not least, see if we're in retr mode and if this is us, print it out */
        if (skipleadingborder == FALSE)
          if (findpartname)
            { // see if this should be output as part of the request 
              showthisline(prefix, partcount, findpartname, realline, lines, inheader);
            }

      } // while (1)
  }

void pop::showthisline(char *prefix, int partcount, char *findpartname, char *l, int lines, int inheader) 
  {
    /* when checking a part, we have to include this part and all of it's children.
       In the headers it doesn't matter because we hit the correct exact 
       match first, so it works, but for non header parts, we have to
       take all parts of all children */
    // it would be nice not to do this each time, optomize this
    Memory pfx (prefix);
    if (strlen(pfx) > 0)
      pfx.addchar('.'); // first one is 1, second one is 1.2
    pfx.addnum(partcount); 
    // now cut it to match
    UINT xl = strlen(findpartname);
    if (strlen(pfx) >= xl)
      {
        *(pfx + xl) = '\0';
        if (strcmp(findpartname, pfx) == 0)
          {
            if (foundpart == FALSE)
              { 
                foundpart = TRUE;
                response ("+OK Message part follows.");
              }
            if (lines == -1) // in retr command, show all lines
              response(l); // the current line
            else
              { // deal with top command
                /* Simple idea, if we've counted zero lines, and we're in header,
                   we output, as soon as we're at zero and not in header 
                   (keeping in mind we're only here because we are in a viewable
                   part of the message) we start counting. When we hit lines, stop.
                   Heh. Stop. That's going to be interesting. How do we stop? */
                if (linesout >= 0)
                  {
                    response(l); // top counts lines after header, all headers are free
                    if (linesout == 0)
                      {
                        if (inheader == FALSE)
                          linesout++;
                      }
                    else // once we are past header, start counting
                      {
                        linesout++;
                        if (linesout > lines)
                          linesout = -1; // flag to say we're done.
                      } // these
                  } // are
              } // unneccesary
          } // but they make
      } // a nice flowing
  } // pattern :-)

void pop::parsecontenttype(Memory &ct, Memory &boundary, int &multipart, int &thisissubmessage)
  {
    /* the data for the content type header is in ct,
       read the boundary and anything else you want out of it. */
    boundary = "";
    multipart = FALSE;
    
    char *s = stristr(ct, TXT_BOUNDARY "=");
    if (s != NULL)
      {
        int hasquote = FALSE;
        boundary = "--";
        char *bpos = s+strlen(TXT_BOUNDARY "=");
        if (*bpos == '\"')
          {
            hasquote = TRUE;
            bpos++; // skip quote
          }
        boundary.addstr(bpos); // first char after quote
        
        // find ending pos of boundary
        if (hasquote)
          s = strchr(boundary, '\"'); // find ending quote
        else // find a semicolon in case there are more fields
          s = strchr(boundary, ';');
        if (s != NULL)
          *s = '\0';
      }
    trim (boundary);
    /* If this large part is denoting that it's sub parts are pieces of a 
       multipart, then we recurse again, because it's another level of
       nesting. */
    if (stristr(ct, TXT_MULTIPART) != NULL) // of course some moron could make it part of the boundary string
      multipart = TRUE;
   
    /* this one's special. If it's a message, it may be a simple message,
       but we have to recurse into it, but end on our parent's boundary. */   
    if (stristr(ct, TXT_MESSAGE_RFC822) != NULL)
      {
        thisissubmessage = TRUE; // denote special case where we don't recurse but reset inheader
      }
  }

int pop::unfoldreadline(int &nextret, Memory &buffer, msgstore &ms, Memory &l, int inheader, Memory &realline)
  {
    /* If inheader is true we unfold lines as we get them and
       only return one header line, otherwise we just
       read a raw line of the file and return */
    // next ret is like buffer, it's the next line's info
    // also return realline (the amalgamation of partial lines we unfold
    // so we can reproduce the original form of the message, not our unfolded version for headers
    realline = "";
    l = buffer;
    int ret = nextret; // get ready to return the line and read status from the line we read last time
    nextret = ms.readline(buffer); // preload the buffer...
    if (strlen(realline) > 0)
      realline.addstr(CRLF);
    realline.addstr(l);

    if (inheader == FALSE)
      return ret; 

    if (strlen(l) == 0) // don't unfold onto a blank line
      return ret; 

    while ((ret == RET_OK) || (ret == RET_PARTIAL))
      {
        if ((*buffer != ' ') && (*buffer != '\t')) // not folded line
          { 
            return RET_OK; // return l
          }
       
        realline.addstr(CRLF); realline.addstr(buffer);

        l.addchar(' ');
        trim(buffer);
        l.addstr(buffer);

        nextret = ms.readline(buffer); 
      }
    return ret; // error
  }

void pop::foldout(char *respin)
  {
    /* Take this line and fold it over multiple lines if it's too long */
    
    /* I amaze myself sometimes. I already wrote a foldline function.
       It's in utils.cpp and writes out to a file, maybe should be
       a template... this one is smarter and doesn't break on
       quoted information, whereas the other one is not likely to
       hit upon quoted data. */
    
    Memory mresp(respin); // we need to butcher it, make our own copy
    char *resp = mresp;
    int indent = FALSE;
    int inquote = FALSE;
    
    while (1)
      {
        int pos = 0;
        int goodbreakpos = 0;
        while (pos < RFCLINESIZE)
          {
            char c = *(resp+pos);
            if (c == '\0')
              { // we hit the end of the string before it got too long write and exit
                Memory out;
                if (indent) // indent all but first line
                  out = "\t";
                out.addstr(resp);
                response(out);
                return;
              }
            if (inquote == FALSE)
              if (c == ' ')
                goodbreakpos = pos;
            if (c == '\"')
              inquote = (inquote == FALSE) ? TRUE : FALSE;
            pos++;
          }
        if (goodbreakpos == 0)
          { 
            // there was nowhere good to break, then write out everything and bail
            Memory out;
            if (indent) // indent all but first line
              out = "\t";
            out.addstr(resp);
            response(out);
            return;
          }
        *(resp + goodbreakpos) = '\0'; // chop off the line here
        Memory out;
        if (indent) // indent all but first line
          out = "\t";
        out.addstr(resp);
        response(out);
        resp += goodbreakpos+1;
        inquote = FALSE; // restore this to situation at break point
        // actually it has to be false since we'd never mark a good break
        // point in the middle of a quoted string
        indent = TRUE;
      }
  }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// And That Ladies and Gentlemen, Is The Third Half Of That.


