/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>

#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/acceptor.h>
#include <Q/socketio.h>
#include <Q/notecraft.h>
#include <Q/configcraft.h>
#include <Q/filelist.h>
#include <Q/hashtable.h>
#include <flat/fc.h>

#include "release.h"
#include "global.h"
#include "db.h"
#include "io.h"
#include "dbfinder.h"
#include "utils.h"
#include "msgstore.h"
#include "pop.h"

#define NAME "pop"

#define BUFSZ 300 // local buf sizes (gethostname)

/* 
 
   This is the POP3 and POP4 server for the O3 mail system. By Stu Mark.

*/


   
int main(int num, char *opts[])
  {
    ConfigCraft ar(O3NAME, num, opts, TRUE); // look for /etc/o3.cf

    addconfigdefaults(ar); /* we have to do this after refresh as well. */

    NoteCraft ls(NAME, ar[TXT_LOGFILE], val(ar[TXT_LOGLEVEL]));
 
    ioinit(ar, ls); 
    Memory chk;
    getdbrootdir(chk);
    if (strcmp(chk, "/") == 0)
      {
        printf ("Need config file to work with. Pass -M<cfgfilename> on command line\n");
        exit (1);
      }

    ls(NOTEINF) << "Starting " << NAME << endl;
    ls(NOTEINF) << "Listening on port " << ar[NAME "-" TXT_PORT] << endl;
    Memory bindaddr = ar[NAME "-" TXT_BINDADDR];
    if (strlen(bindaddr) > 0)
      ls(NOTEINF) << "Binding to address " << ar[NAME "-" TXT_BINDADDR] << endl;
    ls(NOTEINF) << "Kids wanted " << ar[NAME "-" TXT_KIDS] << endl;

    fixsigpipe(ls);    
    Acceptor ac(NAME, val(ar[NAME "-" TXT_PORT]), opts, ls, bindaddr); // make a socket and listen in on this port
    if (ac.errorstate())
      {
        ls(NOTESEV) << "Unable to make listen socket: " << ac.errstr() << endl;
        return 1;
      }
    ac.sethostsallowlist(ar[NAME "-" TXT_HOSTS_ALLOW]);
        
    KidHandler kd(val(ar[NAME "-" TXT_KIDS]), ar, ac, runner);
    kd.run(); // parent never comes back kids need cleanup

    return 0;
  }

void addconfigdefaults(ConfigCraft &ar)
  {
    ar.adddefault(NAME "-" TXT_KIDS, "0");
    ar.adddefault(NAME "-" TXT_PORT, "110");
    ar.adddefault(NAME "-" TXT_DEFAULT_DOMAIN, ""); // Hmmm... what should we put here. 
    ar.adddefault(NAME "-" TXT_BAD_SLEEP_SECONDS, "5"); // number of seconds to wait after a bad auth
    ar.adddefault(NAME "-" TXT_BAD_AUTH_MAX, "3"); // number of bad auth attempts before we drop them
    ar.adddefault(TXT_LOGFILE, NAME ".log");
    ar.adddefault(TXT_LOGLEVEL, "0");
    ar.adddefault(TXT_SOCKETTIMEOUT, "30"); // wait 30 seconds for client to respond
  }


/* The application */

int runner(int num, ConfigCraft &ar, Acceptor &ac)
  {
    /* One of these gets run per process. */
    NoteCraft &ls = ac.getNC();
    
    pop go(ar, ac, ls, num);
    return go.run();
  }

                
pop::pop (ConfigCraft &Ar, Acceptor &Ac, NoteCraft &Ls, int num) : ar(Ar), ac(Ac), ls(Ls)
  {
    ls(NOTEINF) << "Kid starting." << endl;

    char buf[BUFSZ];
    gethostname(buf, sizeof(buf)-1);
    xhostname = buf;
    mailstore = NULL; // finder actually owns connection, we just need to keep track of where this guy is.
    user = NULL;
    folder = NULL;
    originalstat = NULL;
    uidlpos = NULL;
    finder = NULL;
    dirholder = NULL;
    persistent = FALSE;
  }

pop::~pop(void)
  {
    /* This never really happens, but is here because I'm anal */
    if (originalstat != NULL)
      freemem (originalstat);
    if (uidlpos != NULL)
      delete uidlpos;
    if (folder != NULL)
      delete folder;
    if (user != NULL)
      delete user;

    if (dirholder != NULL)
      delete dirholder;
    if (finder != NULL)
      delete finder; // close all database connections. 
  }

int pop::run()
  {
    int failures = 0;
    while (1) // loop for each session
      {
        SOCKET inout;
        inout = ac.accept(); /* block forever until we get a connection */
        if (inout == -1)
          {
            ls(NOTESEV) << "Error on accept: " << ac.errstr() << endl;
            sleep(5);
            failures++;
            if (failures >= 5)
              return 1; // restart
            continue; // got any other ideas?
          }

        // start a pop session.
        socketio Sco (inout, inout, val(ar[TXT_SOCKETTIMEOUT])); // takes ownership of the socket/FD
        sco = &Sco;
        ls (NOTEINF) << "Session starting." << endl;
        startsession();
        ls (NOTEINF) << "Session ending." << endl;
      } // closes socket, start a new session.
    
    return 0; // exits program
  }
                                  
void pop::startsession(void)
  {
    state = needauth;
    badauth = 0;
    user = NULL;
    folder = NULL;
    heading(); // spit out heading.
    int stout = val(ar[TXT_SOCKETTIMEOUT]);
    
    while (1) // main command accept loop
      {
        Memory l;
        ls(NOTEDEB) << "Memleak: " << memleak << endl;
        
        /* persistent mode means we don't timeout on command read, but we will
           still time out on unresponsiveness after a command has been read. */
        if (persistent)
          sco->settimeout(-1); // a really really long time
        
        int ret = sco->bufferedgetline(l);
        if (ret == FALSE)
          {
            int r = sco->errorstate();
            ls(NOTESEV) << "Error reading from socket: " << r << " " << strerror(r) << endl;
            break; // from while so we can do session cleanup.
          }
        trim(l);
        ls(NOTEDEB) << (char *) l << endl;

        if (persistent) // put this back for the rest of the session.
          sco->settimeout(stout);

        /* do the primestore connection check here, keep trying 
           to reconnect if it's not working. */
        if (checkfinder(ar, ls, finder, dirholder) == FALSE)
          {
            ls(NOTESEV) << "Unable to connect to prime defstore." << endl;
            response("400 Temporary datastore connection failure. Try later.");
            continue;
          }

        if (strncasecmp(l, TXT_USER, sizeof(TXT_USER)-1) == 0)
          {
            if (state != needauth) 
              response(expect());
            else
              processuser(l);
          }
        else if (strncasecmp(l, TXT_PASS, sizeof(TXT_PASS)-1) == 0)
          {
            if (state != needpass)
              {
                response(expect());
                state = needauth; // reset this on bad pw 1939 is vague on this
              }
            else
              if (processpass(l) == FALSE)
                break; // critical failure
          }
        else if (strncasecmp(l, TXT_APOP, sizeof(TXT_APOP)-1) == 0)
          {
            if (state != needauth)
              response(expect());
            else
              if (processapop(l) == FALSE)
                break; // critical failure, end session.
          }
        else if (strncasecmp(l, TXT_RSET, sizeof(TXT_RSET)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processreset();
          }
        else if (strncasecmp(l, TXT_STAT, sizeof(TXT_STAT)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processstat();
          }
        else if (strncasecmp(l, TXT_LIST, sizeof(TXT_LIST)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processlist(l);
          }
        else if (strncasecmp(l, TXT_RETR, sizeof(TXT_RETR)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processretr(l);
          }
        else if (strncasecmp(l, TXT_DELE, sizeof(TXT_DELE)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processdele(l);
          }
        else if (strncasecmp(l, TXT_TOP, sizeof(TXT_TOP)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processtop(l);
          }
        else if (strncasecmp(l, TXT_UIDL, sizeof(TXT_UIDL)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processuidl(l);
          }
        else if (strncasecmp(l, TXT_QUIT, sizeof(TXT_QUIT)-1) == 0)
          {
            processquit(); // in any state
            break; // end session.
          }
        else if (strncasecmp(l, TXT_LPRT, sizeof(TXT_LPRT)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processlprt(l);
          }
        else if (strncasecmp(l, TXT_HELP, sizeof(TXT_HELP)-1) == 0)
          {
            processhelp(); // in any state
          }
        else if (strncasecmp(l, TXT_CAPA, sizeof(TXT_CAPA)-1) == 0)
          {
            processcapa(); // in any state
          }
        else if (strncasecmp(l, TXT_LAST, sizeof(TXT_LAST)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processlast();
          }
        else if (strncasecmp(l, TXT_LFLD, sizeof(TXT_LFLD)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processlfld(l);
          }
        else if (strncasecmp(l, TXT_SFLD, sizeof(TXT_SFLD)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processsfld(l);
          }
        else if (strncasecmp(l, TXT_CFLD, sizeof(TXT_CFLD)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processcfld(l);
          }
        else if (strncasecmp(l, TXT_DFLD, sizeof(TXT_DFLD)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processdfld(l);
          }
        else if (strncasecmp(l, TXT_MOVE, sizeof(TXT_MOVE)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              if (processmove(l) == FALSE) // if this is false, there was a socket error
                break;
          }
        else if (strncasecmp(l, TXT_COPY, sizeof(TXT_COPY)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              if (processcopy(l) == FALSE) // socket error
                break;
          }
        else if (strncasecmp(l, TXT_PERS, sizeof(TXT_PERS)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processpers();
          }
        else if (strncasecmp(l, TXT_SETF, sizeof(TXT_SETF)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processsetf(l);
          }
        else if (strncasecmp(l, TXT_GETF, sizeof(TXT_GETF)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processgetf(l);
          }
        else if (strncasecmp(l, TXT_QUOT, sizeof(TXT_QUOT)-1) == 0)
          {
            if (state != neednothing)
              response(expect());
            else
              processquot(l);
          }
        else if (strncasecmp(l, TXT_NOOP, sizeof(TXT_NOOP)-1) == 0)
          {
            response ("+OK Nooping.");
          }
        else
          {
            response("-ERR Unrecognized command.");
          }
      
        if (badauth > val(ar[NAME "-" TXT_BAD_AUTH_MAX]))
          {
            ls(NOTEINF) << "Ending session, too many failed authentication attempts." << endl;
            response("-ERR Too many failed authentication attempts.");
            break;
          }
      } // end command loop
      
    // do any session cleanup here
  }
  
void pop::response(char *s)
  {
    /* write this to the client, we make a function so we can do any
       other logging we want to. */
    ls(NOTEDEB) << s << endl;
    sco->sendline(s);
  }

char *pop::expect(void)
  {
    /* make a string about what mode we're in to expect a command based on state. */
    switch (state)
      {
        case needauth:
          xmsg = "-ERR Expecting USER or APOP.";
          break;
        case needpass:
          xmsg = "-ERR Expecting PASS.";
          break;
        case neednothing:
          xmsg = "-ERR Invalid command.";
          break;
        default:
          xmsg = "-ERR Unexpected state. Expecting QUIT.";
          break;
      }
    return xmsg;
  }
  
void pop::heading(void)
  {
    /* our logo... */
    /* By putting the text "POP4 rev" in the banner the client knows this is a
       pop4 capable server. Much like apop, thus it is NOT listed in
       capabilities. */
    
    char buf[BUFSZ];
    Memory str;
    str = "+OK ";
    str.addstr(" O3 POP3 POP4 rev 0 Server. Build ");
    Memory relstr;
    getrelease(relstr);
    str.addstr(relstr);
    str.addstr(" " __DATE__ " " __TIME__ "; ");
    // time_t is a long int, somtimes, AIX complains
    sprintf(buf, "<%d.%ld@%s>", getpid(), time(NULL), (char *)xhostname);
    apopheading = buf; // save for apop
    str.addstr(buf);
    sco->sendline(str);
  }



/* The rules about passing in a domain are this:
   username  will assume the default domain in the config file
   username@domain will use the domain name specified after the @
   domain/username will use the domain name specified before the /
   While this may seem backwards, it follows form of other systems
   supporting multi-domain authentication. */

/* Any database failures will cause an immediate end of session. */

void pop::userdomaincheck(void)
  {
    // look for domain @ first
    char *p = strchr(username, '@');
    if (p != NULL)
      {
        domain = p+1;
        *p = '\0';
      }
    // domain/ overrides @domain
    p = strchr (username, '/');
    if (p != NULL)
      {
        *p = '\0';
        domain = username;
        Memory temp = p+1;
        username = temp;
      }
  }
  
void pop::processuser(Memory &l)
  {
    // Get the username and possible / or @ domain.
      
    domain = ar[NAME "-" TXT_DEFAULT_DOMAIN]; // presume this domain
    char *p = strchr (l, ' '); // find username
    if (p == NULL)
      {
        response ("-ERR USER command requires a parameter.");
        return;
      } 
    username = p+1;
    trim(username);
    // check for other spaces
    p = strchr(username, ' ');
    if (p != NULL)
      *p = '\0';

    userdomaincheck(); // check for the / and @
    trim(domain);
    
    // do the actual lookup when the enter password
    ls(NOTEINF) << "User: " << (char *)username << "@" << (char *)domain << endl;
    Memory r = "+OK Hello ";
    r.addstr(username);
    r.addchar('@');
    r.addstr(domain);
    response(r);
    state = needpass;
  }

UINT pop::lookupuser(int &bail) // shared for pass and apop
  { /* Return 0 if there was an error output, load up user-> if they're there.
       set bail to true if we had a critical error with the datastore */
  
    bail = FALSE;
    tdirectory dir(*finder, ar, ls);
    /* There is some question as to if we should go to a real box
       here or follow the alias, but really, if it is an alias,
       the box will never get mail. */
    int ret = dir.lookupalias(username, domain);
    if (ret == RET_NOTFOUND)
      {
        response ("-ERR Authentication failed.");
        badauth++;
        int mx = val(ar[NAME "-" TXT_BAD_SLEEP_SECONDS]);
        ls(NOTEDIA) << "Bad auth, sleeping: " << mx << endl;
        sleep (mx);
        return 0;
      }
    if (ret != RET_OK)
      {
        ls(NOTEERR) << "Error from alias lookup for user: " << (char *)username << "@" << (char *)domain << endl;
        response ("-ERR Temporary failure authenticating user. Try later.");
        bail = TRUE;
        return 0;
      }
    UINT userid = dir.getdata().userid;
    
    // get machine, and checkum passwordum
    Memory defstore; /* Get his storage location */
    fieldtomemory (defstore, dir.getdata().defstore, sizeof(dir.getdata().defstore));

    mailstore = finder->getconn(defstore);
    if (mailstore == NULL)
      {
        ls(NOTESEV) << "Unable to get connection to mail defstore: " << (char *)defstore << endl;
        response ("-ERR Unable to find user mailstore.");
        bail = TRUE;
        return 0;
      }

    // user must be null here, we never reenter this function in one session
    user = new tuser(*mailstore, ar, ls);
    ret = user->lookup(userid);
    if (ret == RET_NOTFOUND)
      {
        ls(NOTESEV) << "Sanity failure. Can't find user " << userid << " in mail defstore " << (char *)defstore << endl;
        response ("-ERR Unable to find user in mailstore.");
        delete user; user = NULL;
        bail = TRUE;
        return 0;
      }
    if (ret != RET_OK)
      {
        ls(NOTESEV) << "Database failure. Can't lookup user " << userid << " in mail defstore " << (char *)defstore << endl;
        response ("-ERR Temporary failure in user mailstore. Try later.");
        delete user; user = NULL;
        bail = TRUE;
        return 0;
      }
    return userid;
  }

int pop::processpass(Memory &l)
  {
    /* this is the checker and loader actually since apop does it to,
       I'll write a separate function for that. 
       return FALSE if the session should end */
  
    Memory pass;
          
    char *p = strchr (l, ' '); // find param
    if (p == NULL)
      {
        response ("-ERR PASS command requires a parameter.");
        return TRUE;
      } 
    pass = p+1;
    // spec says there can be spaces in password, okay, whatever. just not before or after
    trim(pass);
    
    // look up user
    int bail = FALSE;
    
    UINT userid = lookupuser(bail);
    if (userid == 0)
      if (bail)
        return FALSE;
      else
        return TRUE;

    // any failures after here require that you delete user

    /* Check account status before pw, because we might have a disabled alias. */
    if (user->getdata().status != USERSTATUS_ENABLED)
      {
        delete user; user = NULL;
        response ("-ERR Account is disabled.");
        return TRUE; // keep going
      }
      
    Memory dbpass; /* Get his real password */
    fieldtomemory (dbpass, user->getdata().password, sizeof(user->getdata().password));
    if (strcmp(pass, dbpass) != 0) // they say case sensitive
      {
        delete user; user = NULL;
        response ("-ERR Authentication failed.");
        badauth++;
        int mx = val(ar[NAME "-" TXT_BAD_SLEEP_SECONDS]);
        ls(NOTEDIA) << "Bad auth, sleeping: " << mx << endl;
        sleep (mx);
        return TRUE; // keep going
      }
    // I guess it's him.
    loadmailbox(TXT_INBOX);
    response("+OK " TXT_INBOX " selected.");
    state = neednothing;
    return TRUE;
  }

int insertdatesort (const void *i1, const void *i2)
  {
    /* these are strentries of msgptrrecs, sort by insertdate */
    stringentry *s1 = *(stringentry **)i1;
    msgptrrec *m1 = (msgptrrec *)(char *)s1->getstr();
    stringentry *s2 = *(stringentry **)i2;
    msgptrrec *m2 = (msgptrrec *)(char *)s2->getstr();
    if (m1->insertdate < m2->insertdate)
      return -1;
    if (m1->insertdate > m2->insertdate)
      return 1;
    return 0;
  }

void pop::loadmailbox(char *foldername)
  {
    /* user is set, load up the copy of the folder to reset to. */
    
    /* look up this folder, and load the message list for it.
       The null/not-null ness of the folder object is important for if there
       Is a currently selected folder. It shouldn't happen, (if you delete
       current folder, you pop back to inbox, which can't be deleted, but...)
       but mind that situation. */

    ls(NOTEINF) << "Loading folder: " << foldername << endl;
    if (user == NULL)
      return;
    UINT userid = user->getdata().userid;
    
    msglist.WipeList();
    if (folder != NULL)
      delete folder; // sfld calls loadmailbox, it already exists.
    folder = new tfolder(*mailstore, ar, ls);
    int ret = folder->lookup(userid, foldername);
    if (ret == RET_NOTFOUND)
      {
        delete folder; 
        folder = NULL;
        currentfoldername = "";
        return;
      }
    UINT folderid = folder->getdata().folderid;
    
    /* Now that we have the userid and folderid, make a messagelist 
       for that folder. */
       
    tmsgptr mp(*mailstore, ar, ls);
    ret = mp.lookupuserfolder(userid, folderid, msglist);
    if (ret != RET_OK)
      {
        ls(NOTESEV) << "Error from database looking for message pointers for user/folder: " << userid << "/" << folderid << endl;
        delete folder;
        folder = NULL; /* There's a difference between we can't read and empty. Don't let them update on a can't read. */
        currentfoldername = "";
        return;
      }
      
    /* The msgptrrec is useful enough that we can use it in its raw mode for
       a pop session. Anything that it doesn't do for pop4, probably should be
       added to the mailstore msgptr definition. */

    /* it is important to return this in consistent order. By the index
       it will come in message order so we sort by insertdate. */   
    msglist.Sort(insertdatesort);
    
    /* Make the backup copy of the statuses to check for quit. Array's existance is in
       sync with folder object */
    if (originalstat != NULL)
      freemem(originalstat);
    // also load the hashlist with the uidls and position number
    if (uidlpos != NULL)
      delete uidlpos;

    int lp;
    int max = msglist.ListCount();

    uidlpos = new hashtable<int>(max);
    originalstat = (UINT *)getmem(sizeof(UINT) * max);
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)msglist.GetItem(lp);
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        originalstat[lp] = m->status; // make a copy so we can detect difference
        Memory uid;
        fieldtomemory (uid, m->msgid, sizeof(m->msgid));
        uidlpos->set(uid, lp+1); // make the hashpile, we put the pop position (0= retr 1)
      }
    // save the database case-preserving version (this also won't have trailing delimeter on it 
    Memory t;
    fieldtomemory(t, folder->getdata().foldername, sizeof(folder->getdata().foldername));
    currentfoldername = t; // take the database (pretty) version and hold on to it.
    // so now we're done.
  }

int pop::processapop(Memory &l)
  {
    /* Get the username and md5 digest and run the gamut,
       return FALSE if session should end because of unrecoverable error */
    
    domain = ar[NAME "-" TXT_DEFAULT_DOMAIN]; // presume this domain
    char *p = strchr (l, ' '); // find username

    int badparm = FALSE;
    Memory username;
    Memory digest;
    
    if (gettoken(p, username) == FALSE)
      badparm = TRUE;
    if (gettoken(p, digest) == FALSE)
      badparm = TRUE;
  
    if (badparm)
      {
        response ("-ERR APOP command requires two parameters.");
        return TRUE;
      } 
    username = p+1;

    userdomaincheck(); // check for the / and @
    
    ls(NOTEINF) << "APOP User: " << (char *)username << "@" << (char *)domain << endl;

    // now load them and their password and do the md5 hash on it
    int bail;
    UINT userid = lookupuser(bail);
    if (userid == 0)
      if (bail)
        return FALSE;
      else
        return TRUE;
        
    // any failures after here require that you delete user

    /* Check account status in case it's a disabled alias. */
    if (user->getdata().status != USERSTATUS_ENABLED)
      {
        delete user; user = NULL;
        response ("-ERR Account is disabled.");
        return TRUE; // keep going
      }
      
    Memory dbpass; /* Get his real password */
    fieldtomemory (dbpass, user->getdata().password, sizeof(user->getdata().password));

    // do md5 hash check
    Memory hashthis = apopheading;
    hashthis.addstr(dbpass); /* this is what they should have hashed */
    
    Memory hashout;
    md5hash (hashout, hashthis);
    
    if (strcasecmp(hashout, digest) != 0) // in case they're not in lowercase...
      {
        delete user; user = NULL;
        response ("-ERR Authentication failed.");
        badauth++;
        int mx = val(ar[NAME "-" TXT_BAD_SLEEP_SECONDS]);
        ls(NOTEDIA) << "Bad apop auth, sleeping: " << mx << endl;
        sleep (mx);
        return TRUE;
      }

    // I guess it's him.
    loadmailbox(TXT_INBOX);
    response("+OK " TXT_INBOX " selected.");
    state = neednothing;
    return TRUE;
  }

void pop::processreset(void) 
  { 
    // reset must be only in working/transaction state, undeletes messages.
    // which is basically clear the tmpdel flags
    int lp;
    int max = msglist.ListCount();
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)msglist.GetItem(lp);
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        m->status = originalstat[lp]; // reset deletes and reads.
      }
    response ("+OK Folder reset.");
  }

void pop::processstat(void)
  {
    /* write out the number of messages in this folder and the total size of them */
    /* The spec doesn't allow for error here. */
    UINT total = 0;
    int lp;
    int count = 0;
    int max = msglist.ListCount();
    
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)msglist.GetItem(lp);
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        if ((m->status & MSGSTATUS_TMPDEL) == 0)
          {
            total += m->msgsize;
            count++;
          }
      }
    Memory resp(50);
    sprintf (resp, "+OK %d %d", count, total);
    response(resp);
  }
  
void pop::processlist(Memory &l)
  {
    msginfo(l, FALSE); // not UIDL
  }

void pop::processuidl(Memory &l)
  {
    msginfo(l, TRUE); // UIDL command
  }
  
/* Here we get into one of those programming design issues. There are 17 ways to do it
   but I come down to copying code for similar functions, or have one function with a flag.
   I seem to be leaning towards the flag. */
   
void pop::msginfo(Memory &l, int isuidl)
  {
    /* Get optional argument message number */
    /* This could be so cute if one off case output similar to the
       full list case, but nooooooooooooooooo. */
    char *p = strchr (l, ' '); // find param if any
    // position is consistent throughout session so the index number is the message number
    int max = msglist.ListCount();
    if (p != NULL)
      {
        int v = getmsgpos(p+1); // get the number or uidl converted to a number
        if ((v < 1) || (v > max))
          {
            if (isuidl)
              response ("-ERR Invalid argument to UIDL command.");
            else
              response ("-ERR Invalid argument to LIST command.");
            return;
          }
        // see if it has been deleted. 
        stringentry *s = (stringentry *)msglist.GetItem(v-1); // zlist counts from 0
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        if ((m->status & MSGSTATUS_TMPDEL) != 0)
          {
            response ("-ERR Message doesn't exist.");
            return;
          }
        Memory resp(50);
        if (isuidl)
          {
            Memory uid;
            fieldtomemory (uid, m->msgid, sizeof(m->msgid));
            sprintf (resp, "+OK %d %s", v, (char *)uid);
          }
        else
          sprintf (resp, "+OK %d %d", v, m->msgsize);
        response (resp);
        return;
      } 

    /* The entire list. Minus deleted */
    Memory resp(SZ_MSGID+50); // just in case
    
    /* Ahhh, here's some poor design for you, we have to run through
       the loop just to get a count of non deleted messages. */
       
    int lp;
    int count = 0;
    for (lp = 1; lp < max+1; lp++)
      {
        stringentry *s = (stringentry *)msglist.GetItem(lp-1); // zlist counts from 0, the for counts from 1
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        if ((m->status & MSGSTATUS_TMPDEL) == 0)
          count++;
      }
    
    sprintf (resp, "+OK %d messages.", count);
    response(resp);
    for (lp = 1; lp < max+1; lp++)
      {
        stringentry *s = (stringentry *)msglist.GetItem(lp-1); // zlist counts from 0, the for counts from 1
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        if ((m->status & MSGSTATUS_TMPDEL) == 0)
          {
            if (isuidl)
              {
                Memory uid;
                fieldtomemory (uid, m->msgid, sizeof(m->msgid));
                sprintf (resp, "%d %s", lp, (char *)uid);
              }
            else
              sprintf (resp, "%d %d", lp, m->msgsize);
            response (resp);
          }
      }
    response (".");
  }
  
void pop::processretr(Memory &l)
  {
    goretr (l, FALSE);
  }

void pop::processtop(Memory &l)
  {
    goretr (l, TRUE);
  }

void pop::goretr(Memory &l, int istop)
  {
    /* Get the message number to retrieve. Make sure it exists and isn't
       deleted, and shoot it out. If istop, check for second param and
       only show those lines. */

    /* Now I have a problem. should I use two different functions for
       message output depending on whether they ask for a part or not?
       I think the answer is yes, because if it's a pop3 client or 
       somebody who just wants the whole message, use the faster function
       to dump it out, the parser is going to slow things down.
       Don't penalize all retrievals just because we support partial
       messages. */

    char *p = strchr (l, ' '); // find msg num
    int badparm = FALSE;
    Memory mid;
    Memory mlines;
    Memory findpartname;
    
    if (gettoken(p, mid) == FALSE)
      badparm = TRUE;
    if (istop)
      {
        if (gettoken(p, mlines) == FALSE)
          badparm = TRUE;
      }
    
    gettoken(p, findpartname);
  
    if (badparm)
      {
        if (istop)
          response ("-ERR TOP command requires two or three parameters.");
        else
          response ("-ERR RETR command requires one or two parameter.");
        return;
      }
      
    int lines = -1; // show all
    if (istop)
      {
        lines = val(mlines); // at worst this hits a \0
        if (lines < 0)
          {
            response ("-ERR Line count must be positive.");
            return;
          }
      }
        
    // check the message for existance
    int max = msglist.ListCount();
    int v = getmsgpos(mid); // get the number or uidl converted to a number
    if ((v < 1) || (v > max))
      {
        if (istop)
          response ("-ERR Invalid argument to TOP command.");
        else
          response ("-ERR Invalid argument to RETR command.");
        return;
      }
    // see if it has been deleted. 
    stringentry *s = (stringentry *)msglist.GetItem(v-1); // zlist counts from 0
    msgptrrec *m = (msgptrrec *)(char *)s->getstr();
    if ((m->status & MSGSTATUS_TMPDEL) != 0)
      {
        Memory resp(100);
        sprintf (resp, "-ERR Message %d has been deleted.", v);
        response (resp);
        return;
      }
 
    /* Make sure we can get at least some part of the message before we say okay. */
    Memory msgdefstore;
    fieldtomemory(msgdefstore, m->defstore, sizeof(m->defstore)); // this is the MESSAGE's defstore not the user's mailstore defstore

    Memory msgid;
    fieldtomemory(msgid, m->msgid, sizeof(m->msgid));
    ls(NOTEINF) << "Retrieving message: " << (char *)msgid << " on msgdefstore " << (char *)msgdefstore << endl;

    msgstore ms(ar, ls, msgdefstore, msgid, FALSE); // don't worry about flushing, we're only reading

    if (strlen(findpartname) > 0)
      {
        processretrpart(istop, ms, lines, findpartname);
        return;
      }
    
    Memory buf(TRANSFER_BUF); // transfer how much at a time. Oh, if only I had brail-ized the messagestore
    /* Read in the first chunk */
    int got;
    int ret = ms.read(buf, TRANSFER_BUF, got);
    if ((ret != RET_OK) && (ret != RET_PARTIAL))
      {
        ls (NOTESEV) << "Error reading message " << (char *)msgid << " from message store " << (char *)msgdefstore << endl;
        Memory resp(100);
        sprintf (resp, "-ERR Temporary error reading message %d.", v);
        response (resp);
        return;
      }

    Memory resp(50);
    sprintf (resp, "+OK %d octets", m->msgsize);
    response (resp);
    
    int pastheader = FALSE;
    int linecount = 0;
    char last1 = 0;
    char last2 = 0; // safety check for trailing \r\n
    char last3 = 0;
    /* Read in message bytes, count lines if istop */
    int pos = 0;
    while (1)
      {
        last3 = last2; // for \n\r\n we could use cr flag, but this works too.
        last2 = last1;
        last1 = buf[pos];
        if (sco->senddata(&last1, 1) == FALSE)
          break; // error

        if (pastheader) // only count lines in body, not header
          if (last1 == '\n')
            linecount++;

        if ((last3 == '\n') && (last2 == '\r') && (last1 == '\n')) // cheezy but functional
          pastheader = TRUE;
          
        if (lines >= 0) // -1 means all
          if (pastheader) // keep lines==0 from blocking header
            if (linecount >= lines) // we were one over 3/30/03
              break; // we're done here
            
        if ((last1 == '.') && (last2 == '\n'))
          if (sco->senddata(&last1, 1) == FALSE) // . escape line leader.
            break; // socket error

        pos++;
        if (pos >= got) // past end of buffer
          { // read in next chunk
            ret = ms.read(buf, TRANSFER_BUF, got);
            if ((ret != RET_OK) && (ret != RET_PARTIAL))
              {
                ls (NOTESEV) << "Error reading message " << (char *)msgid << " from message store " << (char *)msgdefstore << endl;
                Memory resp(100);
                sprintf (resp, "-ERR Temporary error reading message %d.", v);
                response (resp);
                return;
              }
            else // if we sucessfully read
              if (got == 0) // eof?
                break; // no more data. We could count chars and read until we got msgsize, but lets not hang if we're out of sync
            pos = 0;
          }
      }
    
    if ((last2 != '\r') || (last1 != '\n'))
      response (""); // force a trailing \r\n if there isn't one on the message
    response (".");

    if (istop == FALSE)
      {
        // mark the message as read if they used the retr command
        m->status |= MSGSTATUS_READ;
      }
  }

void pop::processdele(Memory &l)
  {
    /* Mark the param 1 numbered message as deleted. */
    char *p = strchr (l, ' '); // find msg number
    if (p == NULL)
      {
        response ("-ERR DELE command requires a parameter.");
        return;
      } 
    int max = msglist.ListCount();
    int v = getmsgpos(p+1); // get the number or uidl converted to a number
    if ((v < 1) || (v > max))
      {
        response ("-ERR Invalid argument to DELE command.");
        return;
      }
    // see if it has already been deleted. 
    stringentry *s = (stringentry *)msglist.GetItem(v-1); // zlist counts from 0
    msgptrrec *m = (msgptrrec *)(char *)s->getstr();
    if ((m->status & MSGSTATUS_TMPDEL) != 0)
      {
        response ("-ERR Message has been deleted.");
        return;
      }
    m->status |= MSGSTATUS_TMPDEL;
    Memory resp(50);
    sprintf (resp, "+OK Message %d deleted.", v);
    response (resp);
    return;
  }

void pop::statusflush(void)
  {
    // this gets called on quit and on select folder.
    // and on move and copy, since they require a folder refresh afterwards
    // do the database stuff, flush changes to status and delete messages
    // as necessary.
    /* we make one msgptr object to do the work of all deletes and updates
       because it's quicker than reopening and closing for each hit. */
 /// xxxz reload backup list with new list
 
    tmsgptr worker(*mailstore, ar, ls);

    int lp;
    int max = msglist.ListCount();
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)msglist.GetItem(lp);
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        if (originalstat[lp] != m->status)
          {
            // something has changed, update/delete this record.
            Memory msgid;
            fieldtomemory(msgid, m->msgid, sizeof(m->msgid));

            if ((m->status & MSGSTATUS_TMPDEL) != 0)
              { // delete this msgptr
                /* Again, another case where the overzealous io work takes care of removing
                   the actual message (or dropping the ref count) for us, we just say delete. */
                int ret = worker.del(m->userid, m->folderid, msgid, TRUE); // work on record and file
                if (ret != RET_OK) // not that the user cares...
                  ls(NOTEERR) << "Unable to delete record during statusflush user/fld/msg/retcode " << m->userid << " " << m->folderid << " " << (char *)msgid << " " << ret << endl;
                else
                  {
                    // lower quota in user record
                    adjustquota(m->msgsize);
                  }
              }
            else
              { // update this msgptr with the new status
                // copy the field data to the updater worker msgptr object
                memcpy (&worker.getdata(), m, sizeof(msgptrrec));
                int ret = worker.update(m->userid, m->folderid, msgid);
                if (ret != RET_OK) // not that the user cares...
                  ls(NOTEINF) << "Unable to update record during statusflush: " << m->userid << " " << m->folderid << " " << (char *)msgid << endl;
              } // if update
          } // if status changed on this message
      } // looping through message list
  } 
    
void pop::processquit(void)
  {
    /* This is one of the more complicated ones.
       We see if we're logged in, if so, see if any 
       statuses have changed, if so, update all the appropriate
       database records. Failure is NOT an option :-) */
    if (state == neednothing)
      {
        statusflush(); // just like when we select a new folder
      } // if in the logged in state.

    // do this now in case something blows up?
    response ("+OK Quitting.");
    
    // clean up all session related information
    apopheading = "";
    username = "";
    domain = "";
    if (originalstat != NULL)
      freemem (originalstat);
    if (folder != NULL)
      delete folder;
    if (user != NULL)
      delete user;
      
    originalstat = NULL;
    folder = NULL;
    user = NULL;
    /* we don't own the mailstore, it's just a pointer into the dbfinder list.
       The dbfinder gets cleaned up on process exit (which never really happens) */
  }

void pop::adjustquota(UINT sz)
  {
    // subtract this amount from the user's quota used value
    if (user == NULL)
      return;
    user->getdata().quotaused -= sz;
    int ret = user->update(user->getdata().userid); // if this fails, it's not a problem, the recalc will catch it. 
    if (ret != RET_OK)
      {
        ls(NOTESEV) << "Unable to adjust quota for " << user->getdata().userid << "." << endl;
      }
  }

void pop::processlast(void)
  {
    /* Print out the number of the last message read */
    int lp;
    int max = msglist.ListCount();
    for (lp = max; lp > 0; lp--) // go backwards to find the first read
      {
        stringentry *s = (stringentry *)msglist.GetItem(lp-1); // from max-1 to 0
        msgptrrec *m = (msgptrrec *)(char *)s->getstr();
        if ((m->status & MSGSTATUS_READ) != 0)
          break;
      }
    if (lp == 0) 
      lp = 1;
    Memory resp(50);
    sprintf (resp, "+OK %d", lp); // lp is 1-max, not 0
    response(resp);
  }

int pop::getmsgpos(char *p)
  { /* from this point in cmd line, get the number or if
       string is led with '-' then lookup the uidl in the hashlist and
       return the position number. return 0 if not found */
    Memory r(p);
    trim(r);
    p = strchr(r, ' '); 
    if (p != NULL)
      *p = '\0'; // strip off any other parameters
    if (*r == '-')
      {
        Memory uidl(r+1);
        int pos;
        int ret = uidlpos->get(uidl, pos);
        if (ret == FALSE)
          pos = 0;
        return pos;
      }
    else
      return val(r);
  }

void pop::processhelp(void)
  {
    response ("+OK Help list follows.");
    response ("HELP - O3 POP3 POP4 server command list");
    response ("USER <username>             - Authentication username");
    response ("PASS <password>             - Authentication password (after USER)");
    response ("APOP <username> <md5digest> - APOP authentication");
    response ("RSET                        - reset session to logged in state");
    response ("STAT                        - Message count and total size");
    response ("LIST [msg n]                - show msg size of message n or entire folder");
    response ("UIDL [msg n]                - show unique id of message n or entire folder");
    response ("RETR <msg n> [part p]       - retrieve message n, display part p");
    response ("TOP <msg n> <lines l> [part p]");
    response ("                            - retrieve message header plus l lines of part p");
    response ("DELE <msg n>                - delete message n");
    response ("LAST                        - display the last message number retrieved");
    response ("NOOP                        - no operation");
    response ("CAPA                        - display server capabilities");
    response ("QUIT                        - commit changes and exit session");

    response ("LFLD [folder pathname]      - list folders");
    response ("SFLD <folder pathname>      - select folder");
    response ("CFLD <folder pathname>      - create folder");
    response ("DFLD <folder pathname>      - delete folder");

    response ("MOVE <destination folder>   - move messages to folder, followed by");
    response ("                              dot terminated message list");
    response ("COPY <destination folder>   - copy messages to folder, followed by");
    response ("                              dot terminated message list");

    response ("SETF <msg n> <flagname> <on or off>");
    response ("                            - set or clear flagname (or number) for msg n.");
    response ("GETF <msg n>                - get set flag list for msg n.");

    response ("QUOT                        - show user information");

    response ("LPRT <msg n>                - List parts of a message");

    response ("PERS                        - Enable persistent connection");
 
    response ("All msg n references can be message number or UIDL if preceeded by -");

    response (".");
  }

void pop::processcapa(void)
  {
    response ("+OK Capability list follows.");
    response ("TOP");
    response ("USER");
// not yet    response ("RESP-CODES");
// not yet    response ("LOGIN-DELAY 0");
    response ("PIPELINING"); // this is taken care of by our buffered getline 
    response ("EXPIRE NEVER");
    response ("UIDL");
    response ("IMPLEMENTATION O3-Mail-System");
    response ("PERSISTENT"); // pop4 specific
    response ("QUOTA"); // pop4 specific
    response (".");
  }

void pop::processpers(void)
  {
    /* Turn on persistent mode. Just means we don't time them out. */
    persistent = TRUE;
    response("+OK Persistent mode enabled.");
  }
  
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// And That Ladies and Gentlemen, Is The First Half Of That.


