/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>

#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/acceptor.h>
#include <Q/socketio.h>
#include <Q/notecraft.h>
#include <Q/configcraft.h>
#include <Q/filelist.h>
#include <Q/hashtable.h>
#include <flat/fc.h>

#include "release.h"
#include "global.h"
#include "db.h"
#include "io.h"
#include "dbfinder.h"
#include "utils.h"
#include "msgstore.h"
#include "pop.h"


/****************************** POP 4 *************************/

/* Folder commands:

   lfld [/ or foldername]   List folders child to current folder or foldername if specified, can specify /
   sfld <foldername>   select folder name to be current/loaded
   cfld <foldername>   create folder
   dfld <foldername>   delete folder

   foldername can be offset from current position or / delimited
   much like unix filesystem namespace. /inbox is reserved.
   names are case preserving but not case sensitive.

   In storage the / is stored, when entered via the protocol commands
   it is escaped with \ and \ is represented as \\
   I could have made // escape a real slash, but I thought it would be
   more consistent to be like unix.
   
   In other words, the hierarchy delimeter is / when the client
   wants to put / in the name you escape it with \/
   
   Can't delete /inbox, you can always SFLD /INBOX
   Can't create a folder unless all parents exist.
   (there must be a  /F1/F2 in order for there to be a /F1/F2/F3)
   Can't delete a folder unless it is empty of subfolders and messages, yes I know
   it's bad, but it keeps people from doing something like rm -rf /
   / is not a folder, but can be used in LFLD command.
   spaces are allowed inline unescaped for foldernames
   leading or trailing spaces are not allowed, they are trimmed off.
   can't delete current folder.
   client is going to have to go to another folder anyway, so might as well
   sfld /inbox first and then delete the folder.
   
   O3 stores foldernames as full paths with escaped characters and everything.
   It does not store trailing hierarchy delimeters and will not create
   a folder with a trailing / or \    (those would be illegal names)
*/   

void pop::processlfld(Memory &l)
  {
    /* If no param is passed, use currentfoldername
       otherwise, unescape \/ to TXT_FOLDERDELIM
    */
    Memory leader;
    Memory in(l);

    // can't use gettoken here, because foldernames can have spaces.
    trim(in);
    char *p = strchr (in, ' '); // find param
    if (p == NULL)
      leader = currentfoldername; // /inbox
    else
      leader = p+1;
    trim (leader);

    // now see if it's partial or full.
    Memory prefix; // the prefix string to look for in folder list
    if (*leader == '/') 
      prefix = leader;
    else
      { // add to current path
        prefix = currentfoldername;
        prefix.addstr(leader);
      }
      
    /* load up the folder table for this user and see what matches the prefix */
    zlist all;
    int lp;
    UINT userid = user->getdata().userid;
    int ret = folder->getfolderlist(userid, all);
    if (ret != RET_OK)
      {
        response ("-ERR unable to retrieve folder list.");
        return;
      }
      
    /* Let's make sure there's at least SOMETHING that matches what
       they've entered, unless it's "/" */
    // checking root?
    if (strcmp(leader, "/") == 0) // special case where trailing "/" is allowed
      prefix = "";
    else
      {
        tfolder f(*mailstore, ar, ls);
        ret = f.lookup(userid, prefix);
        if (ret != RET_OK)
          {
            Memory resp ("-ERR ");
            resp.addstr(prefix);
            resp.addstr(" Folder not found.");
            response (resp);
            return;
          }
      }

    int max = all.ListCount();
    response("+OK folder list follows.");

    prefix.addchar('/'); // don't find the parent, just the children, since folders
    int plen = strlen(prefix);
    // aren't stored with trailing hierarchy delims, this will never find the parent.

    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)all.GetItem(lp);
        folderrec *f = (folderrec *)(char *)s->getstr();
        // only get subset of what was asked for
        Memory name;
        fieldtomemory (name,  f->foldername, sizeof(f->foldername));
        Memory chk(name);
        if (strlen(chk) >= (UINT)plen) // matches with trailing / so that it only gets children
          *(chk+plen) = '\0';
        if (strcasecmp(prefix, chk) == 0)
          response (name);
      }
    response (".");
    return;
  }

int pop::getfolderparam(Memory &name, Memory in)
  {
    Memory leader;
    
    trim(in);
    char *p = strchr (in, ' '); // find param
    if (p == NULL)
      return FALSE;

    leader = p+1;
    trim (leader);

    // now see if it's partial or full.
    if (*leader == '/') 
      name = leader;
    else
      { // add to current path
        name = currentfoldername;
        addslash(name);
        name.addstr(leader);
      }
    return TRUE;
  }

void pop::processsfld(Memory &l)
  {
    /* This will cause a flush of the current state
       of the current folder. If the client doesn't want to save
       the changes, they should RSET */
       
    /* A parameter is required. If it starts with / it is a fully qualified
       if not it is appended to the current folder. */
    /* If there is an error, the flush happens anyway. */
    
    statusflush(); // we do the best we can, if there's an error it won't matter too much, client has to reload folder anyway

    Memory name;
    if (getfolderparam(name, l) == FALSE)
      {
        response ("-ERR folder name required for select folder.");
        return;
      }
      
    Memory savecurrent = currentfoldername; // in case something goes wrong

    // first see if it exists, so we can avoid a bad failure
    // this is double work, but it saves uglyness if it doesn't exist and we tried to load it
    UINT userid = user->getdata().userid;
    tfolder f(*mailstore, ar, ls);
    int ret = f.lookup(userid, name);
    if (ret == RET_NOTFOUND)
      {
        Memory resp ("-ERR ");
        resp.addstr(name);
        resp.addstr(" Folder not found.");
        response (resp);
        return;
      }
    loadmailbox (name);
    if (folder == NULL) // it didn't work for some reason
      {
        // this will cause a renumbering if there was an error loading the new folder
        // and the flush caused deletes.
        loadmailbox(savecurrent); // go back where we were
        Memory resp ("-ERR ");
        if (folder == NULL) // then we have REAL problems, but this is more likely since we can't get a folder we know is there.
          resp.addstr("Critical failure, unable to reload current folder on error with: ");
        else
          resp.addstr("Unable to select folder: ");
        resp.addstr(name);
        response (resp);
        return;
      }
    Memory resp ("+OK ");
    resp.addstr(name);
    resp.addstr(" selected.");
    response (resp);

    return;
  }

int pop::getfolderpiece(Memory &leader, char *&p)
  {
    // given /inbox/blah in p return "/inbox" in leader, false if there's a problem
    // process escaping accordingly
    // a return of TRUE with an empty leader means end of foldernames
    leader = Memory (50);
    if (*p == '\0')
      return TRUE; // end of string
    if (*p != '/')
      return FALSE;
    int first = TRUE;
    while (*p != '\0')
      {
        if (*p == '\\')
          {
            leader.addchar(*p);
            p++;
            if (*p == '\0')
              return FALSE; // can't end with backslash
            leader.addchar(*p); // don't process possible '\/' below
            p++;
            continue;
          }
        else
          if (first == FALSE)
            if (*p == '/')
              return TRUE;
        leader.addchar(*p);
        p++;
        first = FALSE;
      }
    if (strlen(leader) == 1)
      leader = ""; // we found a trailing "/"
    return TRUE;
  }

int pop::validatefolder(zlist &pieces, Memory &fname)
  {
    /* we are passed a trimmed fully qualified foldername
       Make sure there are no trailing characters at the end of the
       folder name.
       make sure all of the folder pieces leading up to the foldername
       exist (verify existance of all parents)
       and make sure the folder doesn't already exist. 
       return TRUE if it's okay, false if not for any reason.
       pieces will be a zlist of the parent folder pieces, each has a leading /
    */
    char *p = fname;
    Memory leader;
    Memory checker;
    while (1)
      {
        int ret = getfolderpiece(leader, p);
        if (ret == FALSE)
          return FALSE;
        trim(leader+1); // don't allow spaces at start or end of pathname piece. (the +1 skips the leading /)
        if (strlen(leader) == 1) // just the /
          return FALSE; // they're doing some stupid shit with "/   /"
        if (strlen(leader) == 0)
          break; // got to end
        stringentry *s = new stringentry(leader);
        pieces.Add(s);
      }
      
    // now run through the list taking all the pieces but the last one
    // make sure they exist 
    if (pieces.ListCount() == 0) // they put in / which isn't a valid folder
      return FALSE;
      
    UINT userid = user->getdata().userid;
    int lp;
    for (lp = 0; lp < pieces.ListCount()-1; lp++)  // check all levels except the one we're gonna create
      { 
        stringentry *s = (stringentry *)pieces.GetItem(lp);
        checker.addstr(s->getstr());
        tfolder f(*mailstore, ar, ls); // this looks ugly, but we already know the file's open, no hit on fserver will happen
        int ret = f.lookup(userid, checker);
        if (ret == RET_NOTFOUND)
          return FALSE;
      }
    // if we're here then all the parent folders exist.
    return TRUE;
  }

void pop::processcfld(Memory &l)
  {
    Memory param;
    if (getfolderparam(param, l) == FALSE)
      {
        response ("-ERR folder name required for create folder.");
        return;
      }
    
    /* See if folder exists already. */
    zlist pieces;
    int ret = validatefolder(pieces, param);
    if (ret == FALSE)
      {
        response ("-ERR Invalid folder name for create folder.");
        return;
      }
    // now try  and create the whole full folderpathname. we might 
    // get a duplicate but other than that it should work.
    UINT userid = user->getdata().userid;
    Memory checker;
    int lp;
    for (lp = 0; lp < pieces.ListCount(); lp++)
      { 
        stringentry *s = (stringentry *)pieces.GetItem(lp);
        checker.addstr(s->getstr());
      }
    tfolder f(*mailstore, ar, ls);
    ret = f.add(userid, checker);
    if (ret != RET_OK)
      {
        response ("-ERR Unable to create folder.");
        return;
      }
    response ("+OK Folder created.");
  }

void pop::processdfld(Memory &l)
  {
    /* make sure there are no messages and no subfolders then delete it.
       I realize there will be a small window of opportunity here.
       But if the user is that adamant about losing mail, then let them. 
       The io delete folder call will cause all messages to be wiped anyway,
       as well as any child folders and all of their messages. */
    
    Memory param;
    if (getfolderparam(param, l) == FALSE)
      {
        response ("-ERR folder name required for delete folder.");
        return;
      }
    
    UINT userid = user->getdata().userid;
    tfolder f(*mailstore, ar, ls); // a worker bee folder object so we don't screw up current

    // see if folder exists to delete
    int ret;
    ret = f.lookup(userid, param); // get this folder to delete, exact
    if (ret == RET_NOTFOUND)
      {
        response ("-ERR Folder not found.");
        return;
      }
    if (ret != RET_OK)
      {
        response ("-ERR Unable to find folder to delete.");
        return;
      }
    // can't delete the current folder.
    if (strcasecmp(param, currentfoldername) == 0)
      {
        response ("-ERR Can't delete currently selected folder.");
        return;
      }

    if (strcasecmp(param, TXT_INBOX) == 0)
      {
        response ("-ERR Can't delete " TXT_INBOX ".");
        return;
      }
    UINT folderid = f.getdata().folderid; // save this for msg check

    /* See if there are any child folders, easiest way is to ask for one... */
    Memory pchild(param);
    pchild.addchar('/'); // any child folder has this on the end
    ret = f.lookup(userid, pchild, FALSE); // look for ANY/fuzzy child
    if (ret == RET_OK)
      {
        response ("-ERR Folder to delete has other folders in it.");
        return;
      }
    // notfound is the only good way to get in here.
    if (ret != RET_NOTFOUND)
      {
        response ("-ERR Unable to retrieve folder list to delete.");
        return;
      }

    // now check to make sure there are no messages in this folder.
    tmsgptr mp(*mailstore, ar, ls);
    // we only need to look to see if there are ANY messages, like we did for folder above
    ret = mp.lookup(userid, folderid, "", FALSE); // do fuzzy search, msgid is ignored
    if (ret == RET_OK)
      {
        response ("-ERR Folder to delete has messages in it.");
        return;
      }
    // notfound is the only good way to get in here.
    if (ret != RET_NOTFOUND)
      {
        response ("-ERR Unable to retrieve message list to delete folder.");
        return;
      }
    /* Well, if you survived all that... */
    ret = f.del(userid, param);
    if (ret != RET_OK)
      {
        response ("-ERR Unable to delete folder.");
        return;
      }
    response ("+OK Folder deleted.");
  }


/* Move and Copy commands.
   They follow the same premise, you pass the destination folder as the
   only parameter then a list of uidl's in a typical pop3 list
   terminated by a . by itself.
   If it all works, you say OK, if ANYTHING goes wrong, either
   message not found or mailbox full or whatever, then you return -ERR
   and the destination folder is in a questionable state. 
   
   Since these commands do more client interaction (reading) they
   return FALSE if there's a socket error and the parent should
   bail on the session. TRUE otherwise.
   
   The reason copy and move have a batch mode and aren't just one-offs
   like the other commands (like getf/setf/dele) is because
   folder finding may be an expensive process, so while we lose out
   on the error details (about a particular message) at least the thing
   can run in a reasonable amount of time.
*/



int pop::processmove(Memory &fld)
  {
    return processmovecopy(fld, FALSE); // do move function
  }
  
int pop::processcopy(Memory &l)
  {
    // this requires a recalc of the quota
    return processmovecopy(l, TRUE);
  }

int pop::processmovecopy(Memory &fld, int iscopy)
  {
    /* Move and copy are so much the same just if out the differences */
    Memory param;
    int badparam = FALSE;
    if (getfolderparam(param, fld) == FALSE)
      badparam = TRUE;
    
    /* Now get the message list. */
    Memory l;
    zlist msgs;
    int counter = 0;
    int ret = sco->bufferedgetline(l);
    if (ret == FALSE)
      {
        int r = sco->errorstate();
        ls(NOTESEV) << "Error reading from socket: " << r << " " << strerror(r) << endl;
        return FALSE; // exit session
      }
    while (strcmp (l, ".") != 0)
      {
        trim(l);
        ls(NOTEDEB) << (char *) l << endl;

        // as with msgs, a msgid starting with a . should be escaped.
        // O3 will never generate a UIDL that starts with a . but
        // it's good excercise to keep in spec :-)
        char *p;
        if (*l == '.')
          p = l+1;
        else
          p = l;
          
        int mpos = getmsgpos(p); // do the number or uidl check
        // this may be a bit backwards because we convert to a number just
        // to convert to a uidl, but, well, that's nice.
        if ((mpos < 1) || (mpos > msglist.ListCount()))
          mpos = 0; // avoid out of bounds

        if (mpos != 0) // found? add it to list
          {
            stringentry *s = (stringentry *)msglist.GetItem(mpos-1); // zlist counts from 0, mpos from 1
            msgptrrec *m = (msgptrrec *)(char *)s->getstr();
            if ((m->status & MSGSTATUS_TMPDEL) == 0)
              {
                Memory msgid;
                fieldtomemory(msgid, m->msgid, sizeof(m->msgid));
                s = new stringentry(msgid);
                msgs.Add(s);
                counter++;
                if (counter > 100000)
                  {
                    /* arbitrary limit of number of messages you can move at once,
                       keep client from causing a DOS attack by sending too huge a list.
                       This is an O3 specific thing, better that the server stay up. */
                    response ("-ERR Message list too long.");
                    return TRUE; // the rest will be ignored by the command parser
                  }
              } // only add if not deleted
          }
        ret = sco->bufferedgetline(l);
        if (ret == FALSE)
          {
            int r = sco->errorstate();
            ls(NOTESEV) << "Error reading from socket: " << r << " " << strerror(r) << endl;
            return FALSE; // exit session
          }
      }
    // got the message list, check for errors

    /* As per the spec: Since moving messages can cause messages to move out
       of this folder and copy can cause another folder to get messages, update
       the status on disk so that copies get the new/read/whatever status.
       This sorta creates an odd situation, where we have commited tmpdels to disk
       but they're still in memory, but they can't be RSET because we're going to
       reload from disk before we exit. */
    
    statusflush(); // xxxz reload

    if (badparam)
      {
        response (iscopy ? (char *)"-ERR Destination folder name required for copy messages." :
                           (char *)"-ERR Destination folder name required for move messages.");
        return TRUE;
      }
    int max = msgs.ListCount();
    if (max == 0)
      {
        response (iscopy ? (char *)"-ERR No messages listed to copy." :
                           (char *)"-ERR No messages listed to move.");
        return TRUE;
      }

    int perfect = TRUE; // keep track of any errors

    UINT userid = user->getdata().userid;
    tfolder f(*mailstore, ar, ls); 

    // see if destination folder exists 

    // this is an O3 ism. Moving to the current folder is pointless,
    // and copying to the current folder is pointless for O3 because
    // we don't store multiple copies of a message in the same folder.
    if (strcasecmp(param, currentfoldername) == 0)
      {
        response ("-ERR Moving or copying to the current folder is pointless.");
        return TRUE;
      }
    
    ret = f.lookup(userid, param); // get this folder to move/copy to
    if (ret == RET_NOTFOUND)
      {
        response ("-ERR Destination folder not found.");
        return TRUE;
      }
    if (ret != RET_OK)
      {
        response ("-ERR Unable to read destination folder.");
        return TRUE;
      }
    UINT destfolderid = f.getdata().folderid; // save this for msg check


    /* now go through the message list in THIS folder (currentfolder->)
       and start changing the folderid and updating / inserting. */
       
    tmsgptr mp(*mailstore, ar, ls);
    int lp;
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)msgs.GetItem(lp); // the list that the user typed in
        // look for these messages in the current folder
        ret = mp.lookup(userid, folder->getdata().folderid, s->getstr());
        if (ret == RET_OK)
          {
            // for move, we leave alone the folderid in the data record,
            // because update has to know the source record to update,
            // and it uses the data field to update with after
            // it makes the source key
            UINT srcfolderid = mp.getdata().folderid; // save this so we can pass it to update
            mp.getdata().folderid = destfolderid; // move to this folder
            
            if (iscopy)
              {
                /* this is sort of a cop out. We don't check quota here
                   we only fix it at the end, so they can go way over max,
                   but what good does it do them? they're only screwing themselves
                   and not really using space on the server, it's reference counted. */
                ret = mp.add(userid, destfolderid, s->getstr());
                if (ret == RET_OK)
                  { // inc msg ref count
                    /* Get the messagedefstore from the old copy of this message
                       so we know we're going to update the refcount of an
                       existing message. */
                    Memory msgdefstore;
                    fieldtomemory(msgdefstore, mp.getdata().defstore, sizeof(mp.getdata().defstore)); // this is the MESSAGE's defstore not the user's mailstore defstore
                    msgstore ms(ar, ls, msgdefstore, s->getstr(), FALSE); /* go store it again */
                    ret = ms.store(); // this BETTER come back with RET_OK and not RET_NEEDDATA
                    /* msgstore takes care of deleeting the file if we go out of scope without
                       doing a write if we got back RET_NEEDDATA. that probably means
                       the msgdefstore configuration got messed up. */
                    if (ret != RET_OK)
                      { // remove the record we just added
                        if (mp.del(userid, destfolderid, s->getstr(), FALSE) != RET_OK) // don't try and remove/rename file, just record
                          {
                            ls(NOTEERR) << "Unable to delete copied record after msgstore fail. There is an orphaned message record." << endl;
                          }
                        perfect = FALSE;
                      }
                  } // if insert went okay
                else
                  perfect = FALSE;
              } // if iscopy
            else // moving not copying
              {
                ret = mp.update(userid, srcfolderid, s->getstr());
                if (ret != RET_OK)
                  {
                    perfect = FALSE;
                    ls(NOTESEV) << "Unable to update foldermsg record for move." << endl;
                  }
              } // moving
          }
        else
          {
            perfect = FALSE; // do as much as we can
            ls(NOTESEV) << "Request msg for move/copy that was not found." << endl;
          }
      } // for loop

    if (iscopy)
      {
        /* it would be quicker to calculate exactly how much the quota
           changed by, but this is simpler and alreay here. If you wanna
           speed up copy, then do that. */
        ret = recalculatequota(mailstore, userid, ar, ls);
        if (ret == FALSE)
          {
            ls(NOTEERR) << "Unable to recalculate quota on message copy." << endl;
          }
        else
          { // refresh the memory copy we have of the user...
            ret = user->lookup(userid);
            if (ret != RET_OK)
              { 
                // this is REALLY NOT a good sign, we read had it before
                // but can't reread it now. The copy in memory remains
                // hopefully we can survive the session.
                ls(NOTEERR) << "On recalc, unable to re-load user record for: " << userid << endl;
                ls(NOTEERR) << "Probably corrupt index." << endl;
              }
          }
        if (perfect == FALSE)
          response ("-ERR Unable to copy all messages.");
        else
          response ("+OK Messages copied.");
      }
    else // move command
      {
        if (perfect == FALSE)
          response("-ERR Unable to move all messages.");
        else
          response ("+OK Messages moved.");
      }

    /* good or bad, things have changed, if nothing else,
       the original copy of status flags are out of sync with the database 
       reload the folder from disk. */
    loadmailbox(currentfoldername); // something happened, get in sync with reality
    return TRUE;
  }
  
/* Flag commands.

SETF <uidl or msgnum> <flagname or number> <on or off>

GETF <uidl or msgnum> 
+OK 
<first set flagnumber> 
<second set flagnumber> 
.

*/

void pop::processsetf(Memory &l)
  {
    /* Take whatever message id they send, and the flag, and set it or clear it.
       This works like the DELE command, you can RSET it to it's form upon login
       or SFLD, or you can SFLD or QUIT to store it to disk. */
    /* Valid flag names are DELETED READ REPLIED FORWARDED
       You can also specify flags by number. There are 32 flags 1 to 32.
       For historical reasons that I will not go into the position of
       the named flags are defined in in global.h
       VISIBLE = 1  // not part of POP4 spec, internal server use only
       READ    = 3
       FLAGGED = 4
       DELETED = 5 
       REPLIED = 6
       FORWARDED = 7
       All others are user/client defined. */
       
    char *p = strchr (l, ' '); // find msg number
    int badparm = FALSE;
    Memory mid;
    Memory flagname;
    Memory state;
    
    if (gettoken(p, mid) == FALSE)
      badparm = TRUE;
    if (gettoken(p, flagname) == FALSE)
      badparm = TRUE;
    if (gettoken(p, state) == FALSE)
      badparm = TRUE;
      
    if (badparm)
      {
        response ("-ERR SETF command requires three parameters.");
        return;
      }
    
    int on = 2;
    if (strcasecmp(state, "on") == 0) 
      on = TRUE; // we all know TRUE != 2 ;-)
    if (strcasecmp(state, "off") == 0) 
      on = FALSE;
      
    if (on == 2)
      {
        response ("-ERR SETF flag state must be on or off.");
        return;
      } 

    int max = msglist.ListCount();
    int v = getmsgpos(mid); // get the number or uidl converted to a number
    if ((v < 1) || (v > max))
      {
        response ("-ERR Invalid msg id argument to SETF command.");
        return;
      }

    // find flag name/number field
    UINT flagnum = 0;
    if (strcasecmp(flagname, TXT_MSGFLAG_VISIBLE) == 0) 
      flagnum = MSGFLAG_VISIBLE;   
    else if (strcasecmp(flagname, TXT_MSGFLAG_READ) == 0) 
      flagnum = MSGFLAG_READ;
    else if (strcasecmp(flagname, TXT_MSGFLAG_FLAGGED) == 0)
      flagnum = MSGFLAG_FLAGGED;
    else if (strcasecmp(flagname, TXT_MSGFLAG_TMPDEL) == 0)
      flagnum = MSGFLAG_TMPDEL;
    else if (strcasecmp(flagname, TXT_MSGFLAG_REPLIED) == 0)
      flagnum = MSGFLAG_REPLIED;
    else if (strcasecmp(flagname, TXT_MSGFLAG_FORWARDED) == 0)
      flagnum = MSGFLAG_FORWARDED; 
    else
      flagnum = val(flagname);
      
    if ((flagnum < 1) || (flagnum > MAX_FLAGS))
      {
        response ("-ERR Invalid flag name argument to SETF command.");
        return;
      }
    
    UINT flagval = (1 << (flagnum - 1)); // get bit value
    
    stringentry *s = (stringentry *)msglist.GetItem(v-1); // zlist counts from 0
    msgptrrec *m = (msgptrrec *)(char *)s->getstr();

    /* You can set or clear any field regardless of other flags statuses. */
    if (on)
      m->status |= flagval;
    else
      m->status &= ~flagval;
    
    Memory resp(50);
    if (on == 1)
      sprintf (resp, "+OK Flag %d set.", flagnum);
    else
      sprintf (resp, "+OK Flag %d cleared.", flagnum);
    response (resp);
    return;
  }

void pop::processgetf(Memory &l)
  {
    /* Get the list of flags for this message in a pop3 style list */
    char *p = strchr (l, ' '); // find msg number
    int badparm = FALSE;
    Memory mid;
    
    if (gettoken(p, mid) == FALSE)
      badparm = TRUE;
      
    if (badparm)
      {
        response ("-ERR GETF command requires one parameter.");
        return;
      }

    int max = msglist.ListCount();
    int v = getmsgpos(mid); // get the number or uidl converted to a number
    if ((v < 1) || (v > max))
      {
        response ("-ERR Invalid msg id argument to GETF command.");
        return;
      }
    response ("+OK Flag list follows.");
    int lp;
    stringentry *s = (stringentry *)msglist.GetItem(v-1); // zlist counts from 0
    msgptrrec *m = (msgptrrec *)(char *)s->getstr();
    UINT status = m->status;
    for (lp = 0; lp < MAX_FLAGS; lp++)
      {
        UINT flagval = (1 << lp);
        char *name;
        if (status & flagval)
          {
            switch (lp+1) // figure out the name of this flag
              { 
                case MSGFLAG_VISIBLE:
                  continue; // visible is not POP4
                case MSGFLAG_READ:
                  name = TXT_MSGFLAG_READ;
                  break;
                case MSGFLAG_FLAGGED:
                  name = TXT_MSGFLAG_FLAGGED;
                  break;
                case MSGFLAG_TMPDEL:
                  name = TXT_MSGFLAG_TMPDEL;
                  break;
                case MSGFLAG_REPLIED:
                  name = TXT_MSGFLAG_REPLIED;
                  break;
                case MSGFLAG_FORWARDED:
                  name = TXT_MSGFLAG_FORWARDED;
                  break;
                default:
                  name = "";
                  break;
              }
            char g[80];
            
            // displaying the name is optional, we do it because we're trying to set a precedent :-)
            sprintf (g, "%d %s", lp+1, name);
            response(g);
          }
      }
    response (".");
  }

/* Quota command.
   The first two lines are bytes used and bytes max 0 = unlimited
   All the following lines are optional and server specific. */

void pop::processquot(Memory &l)
  {
    if (user == NULL)
      {
        response ("-ERR User quota information unavailable.");
        return;
      }
    int qmax = user->getdata().quotamax;
    int qused = user->getdata().quotaused;

    response("+OK Quota information follows.");
    Memory b;
    b.addnum(qused);
    response (b);
    b = "";
    b.addnum(qmax);
    response(b);
    response("."); // end of list
  }



/* Message part functions are in parts.cpp. */



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// And That Ladies and Gentlemen, Is The Other Half Of That.


