/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>

#include <Q/memm.h>
#include <Q/zlist.h>
#include <Q/acceptor.h>
#include <Q/socketio.h>
#include <Q/notecraft.h>
#include <Q/configcraft.h>
#include <Q/filelist.h>
#include <flat/fc.h>

#include "release.h"
#include "global.h"
#include "db.h"
#include "io.h"
#include "dbfinder.h"
#include "utils.h"
#include "msgstore.h"
#include "smtpsend.h"
#include "smtp.h"

#define NAME "smtp"

#define BUFSZ 300 // local buf sizes (gethostname)

/* This is the SMTP server for the O3 mail system. By Stu Mark.


Inbound mail is handled by looking up the domain list in the resource
table. We'll cache the list for a configureable amount of time. 
If we don't handle the mail for the domain, we see if
the incoming connection is from a place (source address or
interface) where we allow relaying.
If not, we delay on the rcpt to, and make them sit on it.

If the mail is going locally, we deliver it, if the mail is 
going outside, we queue it, and somebody's going to have to 
scan the directory every once in a while to retry outgoing mail.
The accept will timeout every minute and smtp's will vie for
a queue file to process. 
The queue control file (Af) will keep track of the number of attempts
and make a return to sender mail if it fails after a while.

*/


   
int main(int num, char *opts[])
  {
    ConfigCraft ar(O3NAME, num, opts, TRUE); // look for /etc/o3.cf

    addconfigdefaults(ar); /* we have to do this after refresh as well. */

    NoteCraft ls(NAME, ar[TXT_LOGFILE], val(ar[TXT_LOGLEVEL]));
 
    ioinit(ar, ls); // ugly, but it works, and there's only one.
    
    Memory chk;
    getdbrootdir(chk);
    if (strcmp(chk, "/") == 0)
      {
        printf ("Need config file to work with. Pass -M<cfgfilename> on command line\n");
        exit (1);
      }

    /* Now we check for what to do, maybe we want to be a command 
       line mailer and not run a server. */
    Memory param;
    if (ar.getcmdlineparam('h', param) != FALSE) 
      {
        help();
        return 0;
      }
    if (ar.getcmdlineparam('d', param) != FALSE) // look for -d
      { // param is junk, we just need to know if they're doing this or not
        Acceptor junkac(ls);
        smtp go(ar, junkac, ls, 0); // a non-database connecting smtp worker
        go.delivercmdlinemail();
        return 0;
      }

    if (ar.getcmdlineparam('q', param) != FALSE) // look for -q
      { // param is junk, we just run the queu
        Acceptor junkac(ls);
        // this might run a local delivery and connect to the database
        // on demand, but it will never accept connections.
        ls (NOTEINF) << "Running queue." << endl;
        smtp go(ar, junkac, ls, 0);
        go.runqueue();
        return 0;
      }

    ls(NOTEINF) << "Starting " << NAME << endl;
    ls(NOTEINF) << "Listening on port " << ar[NAME "-" TXT_PORT] << endl;
    Memory bindaddr = ar[NAME "-" TXT_BINDADDR];
    if (strlen(bindaddr) > 0)
      ls(NOTEINF) << "Binding to address " << ar[NAME "-" TXT_BINDADDR] << endl;
    ls(NOTEINF) << "Kids wanted " << ar[NAME "-" TXT_KIDS] << endl;

    fixsigpipe(ls);    
    
    Acceptor ac(NAME, val(ar[NAME "-" TXT_PORT]), opts, ls, bindaddr); // make a socket and listen in on this port
    if (ac.errorstate())
      {
        ls(NOTESEV) << "Unable to make listen socket: " << ac.errstr() << endl;
        return 1;
      }
    ac.sethostsallowlist(ar[NAME "-" TXT_HOSTS_ALLOW]);

    resetinprocess(ar, ls);  // if we died before reset working.* files. If we got past acceptor
                             // then we bound okay, so we're pretty sure somebody else isn't running.
        
    KidHandler kd(val(ar[NAME "-" TXT_KIDS]), ar, ac, runner);
    kd.run(); // parent never comes back kids need cleanup

    return 0;
  }

void addconfigdefaults(ConfigCraft &ar)
  {
    ar.adddefault(NAME "-" TXT_KIDS, "0");
    ar.adddefault(NAME "-" TXT_PORT, "25");
    ar.adddefault(NAME "-" TXT_QUEUEDIRECTORY, "/var/spool/mailqueue");  // where the af (queue info) files go 
    ar.adddefault(NAME "-" TXT_DEADQUEUEDIRECTORY, "/var/spool/deadmailqueue");  // where the bad af (queue info) files go 
    ar.adddefault(NAME "-" TXT_QUEUEDATADIRECTORY, "/var/spool/maildata"); // where the bf (message files) files go
    ar.adddefault(NAME "-" TXT_QUEUE_RETRY_SECONDS, "900"); // 15 minutes between tries
    ar.adddefault(NAME "-" TXT_MAX_DELIVERY_ATTEMPTS, "288"); // 3 days worth of 15 minutes
    ar.adddefault(TXT_LOGFILE, NAME ".log");
    ar.adddefault(TXT_LOGLEVEL, "0");
    ar.adddefault(TXT_SOCKETTIMEOUT, "30"); // wait 30 seconds for client to respond
  }

void help(void)
  {
    printf ("O3 SMTP server.\n");
    printf ("by stu mark. stu@pop4.org\n");
    printf ("\n");
    printf ("Usage: smtp -h                  - help screen\n");
    printf ("Usage: smtp -d -f from -t recip - deliver mail from stdin to recip\n");
    printf ("Usage: smtp -q                  - run the queue once\n");
    printf ("Usage: smtp                     - run in server mode\n");
    printf ("Usage: smtp -M configfile       - specify config file\n");
    printf ("Usage: smtp -S opt:value        - override config file option\n");
    
  }

void resetinprocess(ConfigCraft &ar, NoteCraft &ls)
  {
    // look for LOCKPREFIX files, and change them back to normal
    Memory qdir = ar[NAME "-" TXT_QUEUEDIRECTORY];
    filelist fl(qdir, "", TXT_LOCKPREFIX);
    int mx = fl.numitems();
    int lp;
    for (lp = 0; lp < mx; lp++)
      {
        Memory f(qdir);
        addslash(f);
        f.addstr(fl[lp]);
        ls(NOTEDIA) << "Resetting: " << (char *)f << endl;
        gungrablock(ar, ls, f);
      }
  }

int gungrablock(ConfigCraft &ar, NoteCraft &ls, char *fn)
  {
    Memory tname(fn);
    char *r = strrchr(tname, '/');
    if (r == NULL)
      r = tname;
    else
      r++; // skip the /

    // check to see if it's got the prefix
    if (memcmp(TXT_LOCKPREFIX, r, strlen(TXT_LOCKPREFIX)) != 0)
      {
        ls(NOTEERR) << "Sanity failure. Trying to requeue a non proc file: " << (char *)fn << endl;
        return FALSE;
      }
    char *savename = r + strlen(TXT_LOCKPREFIX);

    // we requeue to the same place, and the queue runner checks to make
    // sure we don't attempt to process this file too often.
    Memory slowname = ar[NAME "-" TXT_QUEUEDIRECTORY]; // base directory name 
    addslash(slowname);
    slowname.addstr(savename);
    ls(NOTEDIA) << "Unlocking " << fn << " to " << (char *)slowname << endl;
    int ret = rename(fn, slowname); // nothing should be called this, except a failed unlink from before
    if (ret != 0)
      {
        // we shouldn't get any error unlocking, we own it, shouldn't get contention, log all errors.
        // in all other cases, we have problems.
        ls(NOTEWRN) << "Error trying to rename file: " << (char *)fn << " to " << (char *)slowname << " Error " << errno << " " << strerror(errno) << endl;
        return FALSE; // not that they check or anything.
      }
    return TRUE;
  }


/* The application */

int runner(int num, ConfigCraft &ar, Acceptor &ac)
  {
    /* One of these gets run per process. */
    NoteCraft &ls = ac.getNC();
    
    smtp go(ar, ac, ls, num);
    return go.run();
  }

smtp::smtp (ConfigCraft &Ar, Acceptor &Ac, NoteCraft &Ls, int num) : ar(Ar), ac(Ac), ls(Ls)
  {
    /* Took out the qdir setting from here since we have to read the indirection each time anyway */
    ls(NOTEINF) << "Kid starting." << endl;
    char buf[BUFSZ];
    gethostname(buf, sizeof(buf)-1);
    xhostname = buf;

    cachedomainreadtime = 0;
    cachemsgdefstorereadtime = 0;
    finder = NULL;
    dirholder = NULL;

    /* Load up the relay settings */
    Memory interface;
    interface = ar[NAME "-" TXT_ALLOW_RELAY_INTERFACE];
    if (strlen(interface) > 0)
      {
        setiplist(interface, allowrelayinterfacelist);
        ls(NOTEDIA) << "allow relay interfaces: " << (char *)interface << endl;
      }
    Memory relayhosts;
    relayhosts = ar[NAME "-" TXT_ALLOW_RELAY_HOSTS];
    if (strlen(relayhosts) > 0)
      {
        setiplist (relayhosts, allowrelayhostlist);
        ls(NOTEDIA) << "allow relay hosts: " << (char *)relayhosts << endl;
      }
  }

smtp::~smtp(void)
  { 
    if (dirholder != NULL)
      delete dirholder;
    if (finder != NULL)
      delete finder;
  }

int smtp::run()
  {
    int failures = 0;
    while (1) // loop for each session
      {
        /* since the accept will loop internally on connections from 
           unacceptable places, we don't come back until
           we get a connection from a good place. */
      
        SOCKET inout;
        inout = ac.accept(60); /* Check queue every 60 seconds */
        if (inout == -1)
          {
            if (errno == ETIMEDOUT)
              {
                runqueue();
                continue;
              }
            ls(NOTESEV) << "Error on accept: " << ac.errstr() << endl;
            sleep(5);
            failures++;
            if (failures >= 5)
              return 1; // restart
            continue; // got any other ideas?
          }

        // start an smtp session.
        socketio Sco (inout, inout, val(ar[TXT_SOCKETTIMEOUT])); // takes ownership of the socket/FD
        sco = &Sco;
        ls (NOTEINF) << "Session starting." << endl;
        startsession();
        ls (NOTEINF) << "Session ending." << endl;
      } // closes socket, start a new session.
    
    return 0; // exits program
  }
                                  
void smtp::startsession(void)
  {
    state = needhelo;
    peerclaim = "unknown";
    heading(); // spit out heading.
    sleeptime = 5; // make they pay for bad rcpt's

    /* Here we check if we allow relaying from this interface or this
       client ip. If we do not, then we assume this is from the internet
       and we're doing local delivery at which point we check against
       our valid domain list. */
    /* We do this once at the beginning of each session so we can check the host list
       against this new incoming IP */
    /* This is an OR kinda thing. We will relay if it comes from this interface OR if
       it comes from an IP we allow. This is a judgement call. There are arguments for both
       ways, maybe it should be configurable. But I chose the liberal view. If it's either,
       on a good network or somebody special, then you can do it. */
    allowrelay = FALSE; // don't relay unless we have a reason to.
    // so by default this will be an internal (domain to itself) email system
    // the easiest way to turn it on is by interface. But you HAVE
    // to specifiy something or you can't mail out.

    Memory inaddr(50);
    if (getpeeripstring(inaddr, ac) == FALSE)
      {
        // hmmm.. can't tell. don't allow relaying
        ls(NOTEERR) << "Unable to determine local address. " << errno << " " << strerror(errno) << endl;
        response ("501 Unable to determine address path.");
        return;
      }

    int ret = isinlist(allowrelayinterfacelist, inaddr);
    if (ret)
      {
        allowrelay = TRUE;
        ls(NOTEDIA) << "allowing relay based on interface. " << (char *)inaddr << endl;
      }
    else
      ls(NOTEDIA) << "not allowing relay: " << (char *)inaddr << endl;
    
    struct sockaddr_in *sin;
    int sinlen;
    ac.getpeeraddress(sin, sinlen); // check the other end of socket
    stringfromip(ntohl(sin->sin_addr.s_addr), inaddr);
    ret = isinlist(allowrelayhostlist, inaddr);
    if (ret)
      {
        allowrelay = TRUE;
        ls(NOTEDIA) << "allowing relay based on host. " << (char *)inaddr << endl;
      }
    else
      ls(NOTEDIA) << "not allowing relay: " << (char *)inaddr << endl;
  
    // so now we know if we will allow relaying for this guy.

    while (1) // main command accept loop
      {
        Memory l;
        ls(NOTEDEB) << "Memleak: " << memleak << endl;
        int ret = sco->bufferedgetline(l);
        if (ret == FALSE)
          {
            int r = sco->errorstate();
            ls(NOTESEV) << "Error reading from socket: " << r << " " << strerror(r) << endl;
            break; // from while so we can do session cleanup.
          }
        trim(l);
        ls(NOTEDEB) << (char *) l << endl;

        /* do the primestore connection check here, keep trying 
           to reconnect if it's not working. */
        if (checkfinder(ar, ls, finder, dirholder) == FALSE)
          {
            response("400 Temporary datastore connection failure. Try later.");
            ls(NOTESEV) << "Unable to connect to database. " << endl;
            continue;
          }

        /* big ugly case statement. There aren't enough commands to justify an array scan. */
        // do rcpt first since it happens most frequently.
        if (strncasecmp(l, TXT_RCPT, sizeof(TXT_RCPT)-1) == 0)
          {
            if ((state != needrcpt) && (state != needdata))
              response(expect());
            else
              processrcpt(l);
          }
        else if (strncasecmp(l, TXT_HELO, sizeof(TXT_HELO)-1) == 0) // sizeof returns with \0
          {
            if (state != needhelo)
              response(expect());
            else
              {
                /* Get domain and respond. */
                Memory dom(l+sizeof(TXT_HELO)-1);
                trim (dom);
                peerclaim = dom;
                Memory r("250 ");
                r.addstr(xhostname); r.addstr(" Hello "); r.addstr(dom); 
                r.addstr(" SMTP server ready.");
                response(r);
                reset(); // clean lists and go to needmail from state;
              }
          }
        else if (strncasecmp(l, TXT_EHLO, sizeof(TXT_EHLO)-1) == 0)
          {
            if (state != needhelo)
              response(expect());
            else
              {
                /* Get domain and respond extended. */
                Memory dom(l+sizeof(TXT_EHLO)-1);
                trim (dom);
                peerclaim = dom;
                Memory r("250-"); // continuation
                r.addstr(xhostname); r.addstr(" Hello "); r.addstr(dom); r.addstr(" Go fax yourself.");
                response(r);
                r = "250 SIZE"; // for now, we do this.
                response(r);
                reset(); // clean lists and go to need mail from state;
              }
          }
        else if (strncasecmp(l, TXT_MAIL, sizeof(TXT_MAIL)-1) == 0)
          {
            if (state != needmailfrom)
              response(expect());
            else
              processmailfrom(l);
          }
        else if (strncasecmp(l, TXT_DATA, sizeof(TXT_DATA)-1) == 0)
          {
            if (state != needdata) // must get at least one recipient
              response(expect());
            else
              processdata();
          }
        else if (strncasecmp(l, TXT_RSET, sizeof(TXT_RSET)-1) == 0)
          {
            reset();
            response("250 Reset state.");
          }
        else if (strncasecmp(l, TXT_QUIT, sizeof(TXT_QUIT)-1) == 0)
          {
            ls(NOTEDIA) << "Quit command." << endl;
            reset();
            Memory r;
            r = "221 ";
            r.addstr(xhostname);
            r.addstr(" closing connection.");
            response(r);
            break; // end session.
          }
        else if (strncasecmp(l, TXT_NOOP, sizeof(TXT_NOOP)-1) == 0)
          {
            response ("250 OK. Nooping.");
          }
        else
          {
            response("500 Unrecognized command.");
          }
      } // end command loop
    // do any session cleanup here
  }
  
void smtp::response(char *s)
  {
    /* write this to the client, we make a function so we can do any
       other logging we want to. */
    ls(NOTEDEB) << s << endl;
    sco->sendline(s);
  }

char *smtp::expect(void)
  {
    /* make a string about what mode we're in to expect a command based on state. */
    switch (state)
      {
        case needhelo:
          xmsg = "503 Expecting HELO or EHLO.";
          break;
        case needmailfrom:
          xmsg = "503 Expecting MAIL.";
          break;
        case needrcpt:
          xmsg = "503 Expecting RCPT.";
          break;
        case needdata:
          xmsg = "503 Expecting RCPT or DATA.";
          break;
        default:
          xmsg = "503 Unexpected state. Expecting QUIT.";
          break;
      }
    return xmsg;
  }
  
void smtp::heading(void)
  {
    /* our logo... */
    
    char buf[BUFSZ];
    Memory str;
    str = "220 ";
    
    str.addstr(xhostname);
    str.addstr(" O3 SMTP Server. Build ");
    Memory relstr;
    getrelease(relstr);
    str.addstr(relstr);
    str.addstr("; ");
    /* line too long    str.addstr(" " __DATE__ " " __TIME__ "; "); */
    time_t now = time(NULL);
    strcpy(buf, ctime(&now));
    char *cr = strchr (buf, '\n');
    if (cr != NULL) *cr = '\0'; // ctime adds CRLF we don't want
    str.addstr(buf);

    sco->sendline(str);
  }


void smtp::reset(void) 
  { 
    // don't reset peerclaim or allowrelay, that's a per-connection item
    ls(NOTEDIA) << "resetting state." << endl;
    
    state = needmailfrom;
    envelopemailfrom = "";
    reciplist.WipeList();
    msgstoreid = "";
  }

void smtp::makequeuefilenames(Memory &aftemp, Memory &af, Memory &bf, Memory &idnoise)
  {
    af = ar[NAME "-" TXT_QUEUEDIRECTORY]; // base directory name 
    af.addstr("/A"); // the addslash(), the queue file filename prefix

    aftemp = ar[NAME "-" TXT_QUEUEDIRECTORY];
    aftemp.addstr("/temp.A"); // the temp filename
    
    bf = ar[NAME "-" TXT_QUEUEDATADIRECTORY]; // pair the queue and data directories
    bf.addstr ("/B"); // queue to data directory
    
    char idbuf[BUFSZ];
    struct timezone tz;
    struct timeval tv;
    gettimeofday(&tv, &tz);
    sprintf (idbuf, "f%06d%d%d", getpid(), (int)tv.tv_sec, (int)tv.tv_usec);
    af.addstr(idbuf); // now we have a filename
    bf.addstr(idbuf);
    /* We need to make a queuing filename that other smtp servers will NOT 
       find if it scans for until we're alldone. So we write to aftemp 
       and rename it to qn when we're completely done. */
    aftemp.addstr(idbuf);
    idnoise = idbuf;
    ls(NOTEDEB) << "Make queue files: " << (char *)af << " " << (char *)bf << endl;
  }

void smtp::maketempnamefromaf(Memory &newfn, char *origfn)
  {
    /* Give  /var/spool/mail/Afblah make /var/spool/mail/temp.Afblah */
    char *ps = strrchr(origfn, '/');
    if (ps == NULL)
      ps = origfn;
    ps++; // skip /
    newfn = ar[NAME "-" TXT_QUEUEDIRECTORY]; // base directory name 
    addslash(newfn);
    newfn.addstr("temp.");
    newfn.addstr(ps);
  }

void smtp::processmailfrom(Memory &l)
  {
    // validate the mail from command and get the address they give me.
    // check for the size option.
    
    // let's be anal.
    if (*(l+sizeof(TXT_MAIL)-1) != ':')
      {
        response ("500 Syntax error. Mail From: <address>");
        return;
      }
    /* Get param. */
    Memory mf(l+sizeof(TXT_MAIL));
    trim (mf);
    if (strlen(mf) == 0)
      {
        response ("501 Syntax error. Address expected. Mail From: <address>");
        return;
      }

    /* If they put in "fname lname"@domain.com this won't work, but I don't 
       think we have to eat that. */
    char *p = strchr(mf, ' ');
    if (p != NULL)
      *p = '\0';
    
    char *p2 = NULL;
    if (p != NULL)
      stristr(p, TXT_SIZE); // look for 'size' option
    if (p2 != NULL)
      {
        // do size check
        int maxsize = val(ar[NAME "-" TXT_ESMTP_SIZE]);
        if (maxsize > 0)
          {
            int sendsize = val(p2+strlen(TXT_SIZE)+1); // first thing after "SIZE="
            if (sendsize >= maxsize)
              {
                response("552 Message exceeds maximum acceptable message size.");
                return;
              }
          }
      }
      
    // if you want to check on BS names for spam control,
    // here's where you do it.
    envelopemailfrom = mf;
    ls(NOTEINF) << "Mail from: " << (char *)envelopemailfrom << endl;
    Memory r;
    r = "250 ";
    r.addstr(mf);
    r.addstr(" sender ok.");
    response(r);
    state = needrcpt;
  }

void smtp::processrcpt(Memory &l)
  {
    // validate the rcpt command and get the address they give me, add it to the list.

    // let's be anal.
    if (*(l+sizeof(TXT_RCPT)-1) != ':')
      {
        response ("500 Syntax error. Rcpt To: <address>");
        return;
      }
    /* Get param. */
    Memory mt(l+sizeof(TXT_RCPT));
    trim (mt);
    if (strlen(mt) == 0)
      {
        response ("501 Syntax error. Address expected. Rcpt To: <address>");
        return;
      }

    Memory defstore; // if local, where are they
    int userid = 0;
    Memory relayhost; // if this is a relay domain, put the destination here
    
    emailaddr rcpt(mt);
    int localdelivery = checkdomain (rcpt.getdomain());
    if (localdelivery == FALSE)
      {
        if (allowrelay == FALSE)
          {
            dobadrcpt();
            response ("554 Relaying not allowed.");
            return;
          }
      }
    else
      {
        // if it's a local delivery, look them up.
        
        /* If you wanna use LDAP or something like that, 
           here's where the hook goes. */

        // see if this is a relay domain. If it is, just accept it. 
        tresource res(*finder->getprimeconn(), ar, ls);
        Memory k = TXT_RELAYDOMAINPREFIX;
        k.addstr(rcpt.getdomain());
        // if this is found, we just add it to the destpack blindly.
        // it's technically a local delivery but if we say it's outbound
        // we're done :-)
        int ret = res.lookup(k);
        if (ret == RET_OK) 
          {  
            localdelivery = FALSE; // this effectivly makes it a relay, which is what we want
            fieldtomemory (relayhost, res.getdata().value, sizeof(res.getdata().value));
          }
        else
          { // if not found, then do normal process, check locally
            tdirectory dir(*finder, ar, ls);
            int ret = dir.lookupalias(rcpt.getname(), rcpt.getdomain());
            if (ret != RET_OK)
              {
                if (ret == RET_NOTFOUND)
                  {
                    dobadrcpt();
                    response("550 User unknown.");
                    return;
                  }
                response("451 Temporary failure. Unable to look up user.");
                ls(NOTESEV) << "Error from database looking up user: " << rcpt.getname() << "@" << rcpt.getdomain() << endl;
                return;
              }
            /* What we get back from lookupalias might be a disabled
               alias, but it is a fully functional account
               so we just have to make sure we don't send mail there. */
            /* Get his storage location */
            fieldtomemory (defstore, dir.getdata().defstore, sizeof(dir.getdata().defstore));
            userid = dir.getdata().userid; // userid is 0 if not local
            
            /* heh, just realized, we have to load his user record to see if his
               status is okay. */
            FlatConnection *dbc = finder->getconn(defstore);
            if (dbc == NULL)
              {
                ls(NOTESEV) << "Unable to connect to mailstore: " << (char *)defstore << endl;
                response("451 Temporary database failure.");
                return;
              }
            
            // a future, smarter version of dir.lookupalias will return this
            // so we don't have to do it again...
            tuser user(*dbc, ar, ls);
            ret = user.lookup(userid);
            if (ret == RET_NOTFOUND)
              { // this isn't a good sign
                ls(NOTEERR) << "Sanity failure. User " << userid << " not found in maildefstore: " << (char *)defstore << endl;
                response("451 Temporary database failure.");
                return;
              }
            int status = user.getdata().status;
            if (status != USERSTATUS_ENABLED)
              {
                response("550 Account disabled.");
                return;
              }
          } // if not relay domain
      }  
    
    destpack *dp = new destpack();
    dp->setaddr(rcpt, localdelivery, defstore, userid, relayhost);

    reciplist.Add(dp); // takes ownership
    ls(NOTEINF) << "Rcpt to: " << (char *)mt << endl;

    Memory r;
    r = "250 ";
    r.addstr(mt);
    r.addstr(" recipient is valid.");
    response(r);
    state = needdata;
  }

void smtp::dobadrcpt(void)
  {
    /* Make bad addresses more and more agonizing. */
    /* perhaps when we hit 60 seconds we should drop them? */
    ls(NOTEDIA) << "Bad rcpt. Sleeping: " << sleeptime << endl;
    sleep(sleeptime);
    sleeptime += 5;
    if (sleeptime > 59)
      sleeptime = 59;
  }

void smtp::processdata(void)
  {
    /* This is the interesting bit.
       We take in the file, store it in a temp file.
       We have to in case it's going to any other msgstores or if it's
       all outgoing.
       
       For all the people in our list, we attempt to deliver. We have to 
       check quota, and shove in message store (looking up all the defstores
       for all of our local recips) or write out a queue file for those
       that are going outside. We've already denyed relaying to those
       we're not going to relay for, so if it's going out, we deal with it
       here. Only when we've queued the file and successfully
       delivered locally, do we say okay.
       
       Of course this creates problems if we can store some and not others.
       I'd rather deliver it twice than not deliver it until EVERYBODY can
       get it. Worse case, we mail it to to ourselves, or put it in the 
       queue file that it's a local delivery. Let's do that.
      
       Write the recieved line and add required missing headers yada yada. 
    */

    Memory idnoise;    
    // create Af and Bf filenames
    Memory aftemp, af, bf;
    makequeuefilenames (aftemp, af, bf, idnoise);

    /* start sucking down the message content. Adding missing fields as we go. */

    FILE *fbf = fopen(bf, TXT_FOPENWRITE);
    if (fbf == NULL)
      {
        ls(NOTESEV) << "Error creating: " << (char *)bf << " Error: " << errno << " " << strerror(errno) << endl;
        response ("452 Temporary storage failure.");
        return;
      }

    filecloser x(fbf);
    Memory datestr; // used for received line and Date header if missing
    generaterfc822date(datestr);

    // write recieved line
    Memory l;
    Memory peerip;

    l = "Received: from ";
    l.addstr(peerclaim); 
    if (getpeeripstring(peerip, ac))
      {
        l.addstr("(");
        l.addstr(peerip);
        l.addstr(")");
      }
    l.addstr(" by ");
    l.addstr(xhostname);
    l.addstr(" with SMTP; ");
    l.addstr(datestr);
    fprintf (fbf, "%s\r\n", (char *)l); // errors will be caught when we print the first line

    response ("354 Send message. End message with . on a line by itself.");
    ls(NOTEINF) << "Queueing data info to: " << (char *)bf << endl;
    
    Memory headerforhash;
    int readingheaders = TRUE;
    int datemissing = TRUE;
    int messageidmissing = TRUE;
    int frommissing = TRUE;
    while (1)
      {
        int ret = sco->bufferedgetline(l); // zooma zooma
        if (ret == FALSE)
          {
            int r = sco->errorstate();
            ls(NOTESEV) << "Error reading from socket: " << r << " " << strerror(r) << endl;
            ls(NOTESEV) << "Cancelling message: " << (char *)aftemp << endl;
            x.close(); unlink(bf);
            response ("452 Temporary storage failure.");
            return;
          }

        ls(NOTEDEB) << (char *) l << endl;

        /* If we're reading headers, see if we at least have the required headers.
           If we don't, add them. */
        if (readingheaders)
          {
            if (strncasecmp(l, TXT_DATE, sizeof(TXT_DATE)-1) == 0) // -1 is for NT
              datemissing = FALSE;
            if (strncasecmp(l, TXT_MESSAGEID, sizeof(TXT_MESSAGEID)-1) == 0)
              messageidmissing = FALSE;
            if (strncasecmp(l, TXT_FROM, sizeof(TXT_FROM)-1) == 0)
              frommissing = FALSE;
            headerforhash.addstr(l);
            if ((l[0] == '\0') || ((l[0] == '.') && (l[1] =='\0'))) // we're at the blank line delimeter, or at end of message, check missing headers
              {
                // check for headers to see if they're going to be missing
                readingheaders = FALSE;
                if (datemissing)
                  {
                    fprintf (fbf, TXT_DATE " %s\r\n", (char *)datestr); // errors will be caught when we print the blank line
                    headerforhash.addstr(datestr);
                  }
                if (messageidmissing)
                  {
                    Memory midh(4000); // a bit excessive, eh?
                    sprintf(midh, TXT_MESSAGEID " <%s@%s>\r\n", (char *)idnoise, (char *)xhostname);
                    fprintf(fbf, "%s", (char *)midh);
                    headerforhash.addstr(midh);
                  }
                if (frommissing)
                  {
                    fprintf(fbf, TXT_FROM " %s\r\n", (char *)envelopemailfrom);
                    headerforhash.addstr(envelopemailfrom);
                  }
              }
          }

        // check the .. delim 
        if (l[0] == '.')
          if (l[1] != '\0')
            ret = fprintf (fbf, "%s\r\n", (char *)l+1); // this is the exact message going in the database, add the \r\n, but skip the first .
          else
            break; // end of message
        else
          ret = fprintf (fbf, "%s\r\n", (char *)l); // normal line of message

        if (ret < 0)
          {
            ls(NOTESEV) << "Error writing to: " << (char *)bf << " Error: " << errno << " " << strerror(errno) << endl;
            ls(NOTESEV) << "Cancelling message: " << (char *)bf << endl;
            x.close(); unlink(bf);
            response ("452 Temporary storage failure.");
            return;
          }
      } // end loop while

    if (x.close() == EOF)
      {
        ls(NOTESEV) << "Error closing: " << (char *)bf << " Error: " << errno << " " << strerror(errno) << endl;
        ls(NOTESEV) << "Cancelling message: " << (char *)aftemp << endl;
        unlink(bf);
        response ("452 Temporary storage failure.");
        return;
      }

                            /* * * * */

    /* Now we've got the exact deliverable message file in the Bf file.
       Now we start writing out the queue file, including failed
       local deliverys. */
   
    md5hash (msgstoreid, headerforhash); // make the message store ID

    int keepqueuefile = FALSE; /* If all are local deliverys we can get rid
                                  of this when we're done, if successful */
    ls(NOTEINF) << "Queueing control info to: " << (char *)af << endl;
    FILE *faf = fopen(aftemp, TXT_FOPENWRITE);
    if (faf == NULL)
      {
        ls(NOTESEV) << "Error creating: " << (char *)aftemp << " Error: " << errno << " " << strerror(errno) << endl;
        response ("452 Temporary storage failure.");
        unlink(bf);
        return;
      }
    filecloser xa(faf);

    /* Here is basically the definition of the af queue control file */
    // write out the location of the data file in the queue file.
    // and then all other queue related info.

    // keep everybody that does this in sync
    // this msgstore id comes from the hash of the headers above
    writeaffile(faf, bf, envelopemailfrom, msgstoreid, 0, 0, TRUE); // not a bounce msg, deliver now

    // caller has to queue their own destpacks
    /* Here's the control loop where we attempt delivery.
       Delivery will be to the queuefile (for outgoing) local, or to the 
       queuefile if a local delivery fails. */
    
    Memory lasterror;
    destpack *dp;
    int lp, max = reciplist.ListCount();
    for (lp = 0; lp < max; lp++)
      {
        int ret;
        dp = (destpack *)reciplist.GetItem(lp);
        int qthis = FALSE;
        if (dp->getislocal())
          {
            Memory errstr; // don't need to save this just now the requeue will catch it
            ret = deliverlocal(dp, bf, errstr); // this file to this package
            if (ret == FALSE)
              {
                ls(NOTEINF) << "Error on local delivery to: " << dp->getaddress()->getemail() << " queueing." << endl;
                qthis = TRUE;
                lasterror.addstr(errstr);
              }
          }
        else
          qthis = TRUE; // is relay to the outside, queue it.
        
        if (qthis)
          {
            keepqueuefile = TRUE; // keep af and bf, because we're using it.
            ret = dp->store(faf);
            if (ret == FALSE)
              {
                ls(NOTESEV) << "Error writing to: " << (char *)aftemp << " Error: " << errno << " " << strerror(errno) << endl;
                response ("452 Temporary storage failure.");
                xa.close();
                unlink(aftemp);
                unlink(bf); // don't leave this lying around
                return;
              }
          }
      }
    dp->markend(faf); // end of file mark
    // write out error msgs if any
    writeaferror(faf, lasterror);
    
    if (xa.close() == EOF)
      {
        ls(NOTESEV) << "Error closing: " << (char *)aftemp << " Error: " << errno << " " << strerror(errno) << endl;
        response ("452 Temporary storage failure.");
        unlink(bf);
        unlink(aftemp); // for whatever good it will do to try this.
        return;
      }

    // if we made it here, then we've stored the message.
    // see if we keep the files or are we all done?
    if (keepqueuefile)
      {
        // rename the file
        if (rename(aftemp, af) != 0)
          {
            ls(NOTESEV) << "Error renaming: " << (char *)aftemp << " to " << (char *)af << " Error: " << errno << " " << strerror(errno) << endl;
            ls(NOTESEV) << "Cancelling message: " << (char *)af << endl;
            unlink(aftemp); 
            unlink(bf);
            response ("452 Temporary storage failure.");
            return;
          }
      }
    else
      {
        unlink (aftemp); // never actually made an af file with any recips in it
        unlink (bf);
      }
    ls(NOTEINF) << "Message " << (char *)af << " accepted." << endl;
    response("250 Message accepted for delivery.");
    reset(); // this is implied
    return; /* done and done. */
  }

void smtp::writeaffile(FILE *ff, char *bfname, char *mailfrom, char *storeid,
                       int attempts, int isbounce, int delivernow)
  {    
    /* Make the queue header writing a function, so if we change it,
       the compiler will tell us where :-) */

    fprintf(ff, "%d\n", TXT_AFFILEVERSION); // the version number for this file format
    fprintf(ff, "%s\n", bfname);   // the name of the msg/bf file this queue item refers to
    fprintf(ff, "%s\n", mailfrom); // who the mail is from (envelope address)
    fprintf(ff, "%s\n", storeid);  // message hash if a local delivery message
    fprintf(ff, "%d\n", attempts); // how many times we've attempted to deliver this
    fprintf(ff, "%d\n", isbounce); // 0/1 is this is a bounce message? If so, don't reply if we can't deliver it.
    /* Write out the time now, as in the last time this file was dealt with. */
    time_t t = time(NULL);
    if (delivernow)
      t = 0; // make sure too-soon timer doesn't trip if they want to deliver now.
    fprintf(ff, "%d\n", (int)t);
    /* the last error message goes after the destination list because we
       don't know the error until after we try. */
  }    

int smtp::readaffile(FILE *ff, Memory &bfname, Memory &mailfrom, Memory &storeid,
                      int &attempts, int &isbounce, time_t &tm)
  {    
    /* Read the stuff in from the queue file ff */
    Memory l;
    int ret;

    ret = readline(l, ff); // the version number
    if (ret == 0) return RET_ERROR;

    ret = readline(bfname, ff);
    if (ret == 0) return RET_ERROR;

    ret = readline(mailfrom, ff);
    if (ret == 0) return RET_ERROR;

    ret = readline(storeid, ff);
    if (ret == 0) return RET_ERROR;

    ret = readline(l, ff);
    if (ret == 0) return RET_ERROR;
    attempts = val(l);
    
    ret = readline(l, ff);
    if (ret == 0) return RET_ERROR;
    isbounce = val(l);

    ret = readline(l, ff);
    if (ret == 0) return RET_ERROR;
    tm = val(l);

    return RET_OK;
  }    

/* It would be nice if this could be part of writeaf file, but the logistics
   don't work out. And it would be nice if there were a af file class, but
   once again, it's such a specific case, it's not worth breaking out just
   so the object oriented purists can feel warm and fuzzy. */
   
void smtp::writeaferror(FILE *f, char *errstr)
  {
    /* Write out the multi line error to the queue file. */
    // the message text will have CRLF on the end, the marker won't
    fprintf (f, "%s\n.\n", errstr);  // the reader strips off the CRLF
    // errors won't end in . by themselves.
  }

void smtp::readaferror(FILE *f, Memory &errstr)
  {
    /* read and rebuild the errstr line. */
    errstr = "";
    Memory l;
    while (1)
      {
        int ret = readline (l, f);
        if (ret == 0)
          break;
        if ((*l == '.') && (*(l+1) == '\0')) // check delim
          break;
        if (strlen(l) > 0)
          {
            errstr.addstr(l);
            errstr.addstr(CRLF); // keep it in memory with the CRLF's
          }
      }
  }

int smtp::checkdomain(char *domain)
  {
    /* See if this domain is in our list.
       Refresh the list every cache-wipe-interval seconds */
    time_t now = time(NULL);
    int interval = val(ar[NAME "-" TXT_CACHE_WIPE_INTERVAL]);
    if (interval < 60) // minimum one minute
      interval = 60;
    if (now - cachedomainreadtime > interval)
      {
        /* reread domain list */
        tresource res(*finder->getprimeconn(), ar, ls);
        if (res.getdomainlist(cachedomainlist) != RET_OK)
          { 
            ls(NOTESEV) << "Error reading domain list from resource file." << endl;
          }
        cachedomainreadtime = now;
      }
    int lp;
    int max = cachedomainlist.ListCount();
    for (lp = 0; lp < max; lp++)
      {
        stringentry *s = (stringentry *)cachedomainlist.GetItem(lp);
        if (strcmp(s->getstr(), domain) == 0)
          return TRUE;
      }
    return FALSE;
  }

int smtp::getcurrentmessagedefstore(Memory &defstore)
  {
    /* Return the current setting from the resource table
       Refresh the list every cache-wipe-interval seconds */
    /* Since the resource table is in one place with the directory
       all mailstores' defstores are going to change at the same
       time, perhaps that's not ideal. */
    time_t now = time(NULL);
    defstore = "";
    int interval = val(ar[NAME "-" TXT_CACHE_WIPE_INTERVAL]);
    if (interval < 60) // minimum one minute
      interval = 60;
    if (now - cachemsgdefstorereadtime > interval)
      {
        tresource res(*finder->getprimeconn(), ar, ls);
        int ret = res.lookup(TXT_MESSAGE_DEFSTORE);
        if (ret == RET_NOTFOUND)
          {
            ls(NOTESEV) << "Message defstore not set in resource file. " << endl;
            ls(NOTESEV) << "MAIL WILL NOT BE DELIVERED until this is set with admin program." << endl;
          }
        if (ret != RET_OK)
          return FALSE;
        fieldtomemory(cachemsgdefstore, res.getdata().value, sizeof(res.getdata().value));
        cachemsgdefstorereadtime = now;
      }
    defstore = cachemsgdefstore;
    return TRUE;
  }

int smtp::deliverlocal(destpack *dp, char *bf, Memory &errstr)
  {
    /* try to shove this in the database right now.
       return false if you can't and need to requeue, or
       TRUE if it delivered okay, or you've found an
       error and are generating an email to reply why you couldn't 
       accept delivery (overquota, whatever) */
    /* Any filters/forwarding/spam control you wanna do happens here.
       If you dealt with the message (drop, forward, spam/whatever)
       then return TRUE. */
    /* It's a fair bet that any errors are permanent failure errors.
       Can't write to a disk? no amout of time is going to fix that
       might as well send a bounce right away. */
    /* First check quota, then put the messagebody in, then 
       put the message pointer in his folder */

    /*  It is possible that we were given a command line
        mail to send and the queue running picks up somebody
        else's queue file that's being locally delivered.
        Check the database connection here. */
        
    if (checkfinder(ar, ls, finder, dirholder) == FALSE)
      {
        ls(NOTESEV) << "Unable to connect to mailstore " << endl;
        errstr = "Temporary database failure.";
        return FALSE;
      }

    FILE *f = fopen(bf, TXT_FOPENREAD);
    if (f == NULL)
      {
        ls(NOTEERR) << "Unable to open message file: " << (char *)bf << " Error " << errno << " " << strerror(errno) << endl;
        errstr = "Message file error.";
        return FALSE;
      }
    filecloser x(f); // in case I forget
    long msgsize = sizereset(f); // to check quota
    
    /* Get the user record to see where he is. */
    /* We have the defstore, look up the connection. */
    Memory userdefstore = dp->getdefstore();
    FlatConnection *dbc = finder->getconn(userdefstore);
    if (dbc == NULL)
      {
        ls(NOTEERR) << "Unable to connect to mailstore: " << (char *)userdefstore << endl;
        errstr = "Temporary database failure.";
        return FALSE;
      }
    
    tuser user(*dbc, ar, ls);
    UINT userid = dp->getuserid();
    int ret = user.lookup(userid);
    if (ret == RET_NOTFOUND)
      { // this isn't a good sign, probably out of sync database, or
        // mailstore connect info got crossed
        errstr = "User ";
        errstr.addnum(userid);
        errstr.addstr(" not found.");
        ls(NOTEERR) << "Unable to find user " << userid << " on mailstore " << (char *)userdefstore << endl;
        ls(NOTEERR) << "Maybe mailstore connections are wrong or need to reindex database." << endl;
        return FALSE;
      }
    if (ret != RET_OK)
      {
        ls(NOTEERR) << "Database Error. Unable to load user record for: " << userid << " on mailstore " << (char *)userdefstore << endl;
        errstr = "Unable to load user record for user ";
        errstr.addnum(userid);
        return FALSE;
      }
      
    /* Check the account type. Mail Account will fall through and deliver.
       alias has already been taken care of. We have to do program and
       file and smtp */
    if (user.getdata().type == USERTYPE_PROGRAM)
      {
        // run the program in alt, pipe the bf file to it, if it works, 
        // return OKAY otherwise return FALSE
        
        // we're going to cheap out and system it. Future versions will
        // be more correct. This makes piping the message in real easy :-)
        
        x.close(); // don't need to read it ourselves
        Memory cmd;
        fieldtomemory(cmd, user.getdata().alt, sizeof(user.getdata().alt));
        cmd.addstr(" < ");
        cmd.addstr(bf);
        ls(NOTEINF) << "Running: " << (char *)cmd << endl;
        int ret = system((char *)cmd);
        if (ret == -1)
          {
            ls(NOTEERR) << "Unable to fork for run of program. Error: " << errno << " " << strerror(errno) << endl;
            return FALSE;
          }
          
        ret = WEXITSTATUS(ret); // get program's return code
        if (ret != 0)
          {
            errstr = "Program process for ";
            errstr.addstr(dp->getaddress()->getemail());
            errstr.addstr(" returned ");
            errstr.addnum(ret);
            ls(NOTEERR) << "Program process for " << dp->getaddress()->getemail()  << " returned " << ret << endl;
            ls(NOTEERR) << "Program process: " << (char *) cmd << endl;
            return FALSE;
          }
        return TRUE;
      } /* if it was a program type */
      
    /* There is a small window where the quota can get munged. That's
       why we recalculate when we think we're going over. */
    int qmax = user.getdata().quotamax;
    int qused = user.getdata().quotaused;
    if ((qmax > 0) && (qused + msgsize > qmax)) // 0 is flag for no quota
      {
        recalculatequota(dbc, userid, ar, ls);
        // look it up again.
        ret = user.lookup(dp->getuserid());
        if (ret != RET_OK)
          { 
            // this is REALLY NOT a good sign, we read it before but can't now?
            ls(NOTEERR) << "Unable to re-load user record for: " << userid << " on mailstore " << (char *)userdefstore << endl;
            ls(NOTEERR) << "Probably corrupt index." << endl;
            errstr = "Unable to load user record for user ";
            errstr.addnum(userid);
            return FALSE;
          }
        qmax = user.getdata().quotamax;
        qused = user.getdata().quotaused;
        if (qused + msgsize > qmax)
          {
            /* They are REALLY overquota. Send the sender a message. */
            ls(NOTEINF) << "User " << userid << " is overquota, bouncing message." << endl;
            int ret = generateoverquotamessage(dp, bf); // use the from or reply-to
            if (ret == FALSE)
              {
                errstr = "Unable to generate overquota response.";
                return FALSE; // we couldn't even generate a reply message, don't accept
              }
            /* While we're here, we should put an overquota message
               in the recips mailbox, free of quota charge. xxxz */
            insertoverquotamessage(user);
            return TRUE; // we delivered this as much as we're ever going to.
          }
        // if we get here, we're quota okay.
      }

    /* We store the message. Get the current message store */
    Memory msgdefstore;
    ret = getcurrentmessagedefstore(msgdefstore); // this is for the message
    if (ret == FALSE)
      {
        ls(NOTEERR) << "Unable to get current message storage location from resource table." << endl;
        errstr = "Mail server configuration error. (Message store)";
        return FALSE;
      }
    
    int flushoften = val(ar[NAME "-" TXT_FLUSH_OFTEN]);
    // This msgstore id comes from the af file
    msgstore ms(ar, ls, msgdefstore, msgstoreid, flushoften);
    
    ret = ms.store();
    if (ret == RET_NEEDDATA) // this is only going to happen once per session
      {
        Memory buf(TRANSFER_BUF);
        while (ret > 0)
          {
            ret = fread (buf, 1, TRANSFER_BUF, f);
            if (ret == 0)
              {
                if (feof(f))
                  break; // this is okay
                // we can assume error
                ls(NOTEERR) << "Error reading message file: " << (char *)bf << " Error " << errno << " " << strerror(errno) << endl;
                ms.remove(); // don't let the half crap file linger
                errstr = "Error reading message file.";
                return FALSE;
              }
            int wret = ms.write(buf, ret);
            if (wret != RET_OK)
              {
                ls(NOTEERR) << "Error writing to message store for: " << (char *)msgstoreid << " Error " << errno << " " << strerror(errno) << endl;
                ms.remove(); // don't let the half crap file linger
                errstr = "Error writing to message store.";
                return FALSE;
              }
          }
        ret = RET_OK; // when the ms goes out of scope it will close
      }

    if (ret != RET_OK) // either from store or from needdata
      {
        ls(NOTEERR) << "Unable to store initial message content." << endl;
        /* If we got an error here, it's because we couldn't fopen the file,
           or we couldn't rename it. Either way, the state didn't change,
           so there's nothing to derefcount or unlink */
        errstr = "Error initally writing to message store.";
        return FALSE;
      }

    ret = ms.close(); // benign if never open
    if (ret != RET_OK)
      {
        ls(NOTEERR) << "Unable to store message content." << endl;
        ms.remove(); // we make no pointers, so drop the file.
        errstr = "Error writing to message store.";
        return FALSE;
      }

    x.close(); // the bf file

    /* If we're here, the message body has been stored in the message 
       store, now we insert the message in to the user's msgptr file
       and update his quota. */
    // find the inbox
    tfolder folder (*dbc, ar, ls);
    ret = folder.lookup(userid, TXT_INBOX);
    if (ret == RET_NOTFOUND)
      {
        ls(NOTEERR) << "Unable to find inbox for user " << userid << endl;
        ms.remove(); // we make no pointers, so drop the file.
        errstr = "Unable to find Inbox for user: ";
        errstr.addnum(userid);
        return FALSE;
      }
    if (ret != RET_OK)
      {
        ls(NOTEERR) << "Database Error. Looking for inbox for user " << userid << endl;
        ms.remove(); // we make no pointers, so drop the file.
        errstr = "Unable to load Inbox for user: ";
        errstr.addnum(userid);
        return FALSE;
      }
    
    /* Make and insert the message pointer record. */
    tmsgptr msg(*dbc, ar, ls);
    msg.getdata().userid = userid;
    msg.getdata().folderid = folder.getdata().folderid;
    memorytofield (msg.getdata().msgid, msgstoreid, sizeof(msg.getdata().msgid));
    msg.getdata().status = MSGSTATUS_VISIBLE;
    msg.getdata().insertdate = time(NULL); // now
    msg.getdata().msgsize = msgsize;
    memorytofield(msg.getdata().defstore, msgdefstore, sizeof(msg.getdata().defstore)); // defstore of message not of user
    ret = msg.add(userid, folder.getdata().folderid, msgstoreid);
    if (ret == RET_DUPLICATE) // same recip twice?
      { /* we don't store more than one copy of this message. we don't count it
           against their quota, and we don't inc the ref count for it either. */
        ms.remove(); // drop the ref count to where it was.
        return TRUE; // but say we delivered it, because, well... at some point we did.
      }
      
    if (ret != RET_OK)
      {
        ls(NOTEERR) << "Database Error. Inserting mail for user " << userid << endl;
        ms.remove(); // we make no pointers, so drop the file.
        errstr = "Unable to insert mail for user: ";
        errstr.addnum(userid);
        return FALSE;
      }
    
    /* Well that's that, just the quota and we're done. */
    // the userrecord is still intact.
    user.getdata().quotaused += msgsize;
    ret = user.update(userid); // if this fails, it's not a problem, the recalc will catch it. 
    // or they'll get some free space. ooohhh
    if (ret != RET_OK)
      {
        ls(NOTEINF) << "Unable to update quota for " << userid << ". Mail will be delivered." << endl;
      }
    ls(NOTEINF) << "Message: " << bf << " delivered to " << userid << endl;
    return TRUE; // Yaaaaa!
  }

void smtp::insertoverquotamessage(tuser &user)
  {
    /* This will do a one time try local delivery of an overquota
       message to the recipient.
       The message will bypass the quota check, because obviously
       it would fail.
       The message Id for the message willb e calculated to be
       unique per day. So we will only deliver one per day, because
       duplicates won't get inserted. */
       /* xxxz */
  }

int smtp::generateoverquotamessage(destpack *dp, char *bf)
  {
    /* Generate a message saying this user is overquota and
       add it to the outgoing queue. If we can't return FALSE,
       if it queues okay return TRUE;
       
       Get the from or reply to from the file so we know who to send
       it to. 
    */

    Memory subject("Undeliverable message, recipient over quota.");
    Memory msg;
    msg = "The user ";
    msg.addstr(dp->getaddress()->getemail());
    msg.addstr(" has too much mail in their mailbox. No more mail"
               " will be accepted until room is made available by"
               " this user.");

    return generatemessage(bf, subject, msg);
  }

int smtp::generateundeliverablemessage(char *failedreciplist, char *bf, char *lasterror)
  {
    /* Generate a message saying that this message is undeliverable.
       Send it back to the sender. It would be handy to pass
       along the last error message we got.... */

    Memory subject("Undeliverable message, errors trying to send.");
    Memory msg;
    msg = "The attached message was undeliverable to ";
    msg.addstr(failedreciplist);
    if (strlen(lasterror) > 0)
      {
        msg.addstr(CRLF "Error: ");
        msg.addstr(lasterror);
      }

    return generatemessage(bf, subject, msg);
  }

int smtp::generatemessage(char *bf, Memory &subject, Memory &msg)
  {
    /* Generate a bouce message to the sender in bf (the incoming
       attempted to deliver email) with this subject and content.
       Adds the message as an MIME attachment */
    Memory sendto;

    ls(NOTEINF) << "Generating auto-reply message." << endl;
    int ret = getfromorreplyto(sendto, envelopemailfrom, bf);
    if (ret == RET_ERROR)
      {
        /* error dealing with file. */
        ls(NOTEERR) << "Unable to read message file: " << bf << " Error: " << errno << " " << strerror(errno) << endl;
        return FALSE;
      }
    if (ret == RET_NOTFOUND)
      {
        ls(NOTEINF) << "No sender in email to reply to. No message generated." << endl;
        return TRUE; /* we tried, but there's nowhere to reply to. */
      }
    /* Make a queue filename bf file. */
    Memory raftemp, raf, rbf, idnoise;
    makequeuefilenames(raftemp, raf, rbf, idnoise);
    ret = makemessage(rbf, sendto, subject, msg, bf); /* third param as optional attachment */
    if (ret != RET_OK)
      {
        ls(NOTEERR) << "Unable to create reply message. Error " << errno << " " << strerror(errno) << endl;
        unlink(rbf); // remove the message file
        return FALSE; // try again later.
      }

    /* Write out the queue file */
    FILE *faf = fopen(raftemp, TXT_FOPENWRITE);
    if (faf == NULL)
      {
        ls(NOTEERR) << "Unable to queue message (" << (char *)raftemp << "). Error " << errno << " " << strerror(errno) << endl;
        unlink(rbf); // remove the message file
        return FALSE; // try again later.
      }
    filecloser x(faf);
    
    // this IS a bounce message, note as such
    msgstoreid = ""; // there is no msgstoreid for this generated message, it's outgoing
    Memory mailfrom = ar[NAME "-" TXT_SYSTEMMAILFROM];
    writeaffile(faf, rbf, mailfrom, msgstoreid, 0, 1, TRUE); // new message attempts = 0, bounce = TRUE, delivernow
    /* make a destpack for the recip and add it tot he queue file */
    destpack dp;
    emailaddr adr(sendto);
    /* The destination will always be external. We will mail it out even if it
       is us. If it is, it will still be quick because the delivery will
       be to localhost. */
    dp.setaddr(adr, FALSE, "", 0, ""); 
    ret = dp.store(faf);
    if (ret == FALSE)
      {
        ls(NOTESEV) << "Error writing to: " << (char *)raftemp << " Error: " << errno << " " << strerror(errno) << endl;
        x.close();
        unlink(rbf); // clean this
        unlink(raftemp); // remove junk af file too.
        return FALSE; // try again later
      }
    dp.markend(faf); // end of file mark
    writeaferror(faf, ""); // being the first queueing of this item, there's no errmsg
    x.close();

    /* The mail file is written, the queue file is written, rename it, 
       and it will get delivered. */
    ls (NOTEINF) << "Queueing auto-reply message: " << (char *)raf << endl;
    if (rename(raftemp, raf) != 0)
      {
        ls(NOTESEV) << "Error renaming: " << (char *)raftemp << " to " << (char *)raf << " Error: " << errno << " " << strerror(errno) << endl;
        ls(NOTESEV) << "Cancelling message: " << (char *)raf << endl;
        unlink(rbf);
        unlink(raftemp); 
        return FALSE; // try again later.
      }
    return TRUE; // done and done
  }


int smtp::makemessage(char *outfile, char *sendto, char *subject, 
                      char *msgtext, char *attachment)
  { /* third param as optional attachment */
    /* return RET_OK if we were able to write the whole message file. */
   
    Memory buf; 
    FILE *f = fopen(outfile, TXT_FOPENWRITE);
    if (f == NULL)
      {
        ls(NOTEERR) << "Unable to create auto-reply message file. Error: " << errno << " " << strerror(errno) << endl;
        return RET_ERROR;
      }
    filecloser x(f);
    
    int domime = FALSE;
    
    if (attachment != NULL) 
      domime = TRUE;

    // write the headers.
    Memory l;
    l = TXT_TO " ";
    l.addstr(sendto);
    foldline(f, l); 
    
    // messages generated by us, always come from this guy.
    l = TXT_FROM " ";
    l.addstr(ar[NAME "-" TXT_SYSTEMMAILFROM]);
    foldline(f, l);
    
    l = TXT_DATE " ";
    generaterfc822date(buf);
    l.addstr(buf);
    foldline(f, l);
    
    Memory dary;
    generateboundary(dary, sendto); // insert randomness

    l = TXT_MESSAGEID " ";
    l.addstr("<");
    l.addstr(dary);
    l.addstr("@");
    l.addstr(xhostname);
    l.addstr(">");
    foldline(f, l);
    
    l = TXT_SUBJECT " ";
    l.addstr(subject);
    foldline(f, l);
    
    if (domime)
      {
        l = TXT_MIME_VERSION " ";
        l.addstr("1.0");
        foldline(f, l);
        
        l = TXT_CONTENT_TYPE " ";
        l.addstr(TXT_MULTIPART_MIXED);
        l.addstr("; " TXT_BOUNDARY "=\"");
        l.addstr(dary);
        l.addstr("\"");
        foldline(f, l);

        Memory prefixb(dary);
        dary = "--"; // from now on, we use the -- prefixed version
        dary.addstr(prefixb);
      }

    foldline(f, ""); // blank line body separator
    if (domime)
      {
        foldline (f, TXT_MIMETEXTMSG);
        foldline (f, "");

        // put boundary and type header for text/plain
        foldline (f, dary);

        l = TXT_CONTENT_TYPE " ";
        l.addstr(TXT_TEXT_PLAIN "; " TXT_CHARSET);
        foldline (f, l);
        
        // this is our message text viewable part, denote that it is inline
        foldline(f, TXT_CONTENT_DISPOSITION " " TXT_INLINE);
        foldline(f, ""); // body separator for this part
      }    
    
    // output message      
    formatline(f, msgtext);
    
    if (domime == FALSE)
      return RET_OK; // if it's juts a simple message, we're done.
      
    /* write the end boundary for the message part, then add the attachment */
    foldline(f, ""); /* techically the CRLF is part of the boundary */
    foldline(f, dary);
    foldline(f, TXT_CONTENT_TYPE " " TXT_MESSAGE_RFC822);
    foldline(f, TXT_CONTENT_TRANSFER_ENCODING " " TXT_7BIT);
    foldline(f, TXT_CONTENT_DISPOSITION " " TXT_ATTACHMENT);
    foldline(f, ""); // body separator

    // spit out the message
    FILE *inf = fopen(attachment, TXT_FOPENREAD);
    if (inf == NULL)
      {
        ls(NOTEERR) << "Unable to read auto-reply message attachment (original message). Error: " << errno << " " << strerror(errno) << endl;
        x.close();
        unlink(outfile);
        return RET_ERROR; // try again later
      }
    filecloser x2(inf);
    while (1)
      {
        int ret = readline (l, inf);
        if (ret == 0)
          break;
        formatline(f, l);
      }
    x2.close();
    foldline(f, "");
    
    l = dary;
    l.addstr("--"); // end of parts boundary
    foldline(f, l);
    
    if (x.close() == EOF)
      {
        unlink (outfile);
        return RET_ERROR;
      }

    // well then, I think we're done
    return RET_OK;
  }



/*                           The other half.                    */

/* process the queue. Pick up a file and try to deliver it. */




void smtp::runqueue(void)
  {
    /* we are given a small shot at running the message queue.
       If we do one at a time, they will only get delivered
       one message a minute, so we will run through the whole
       queue trying to deliver each once.
       I can see there being a problem where all the kids are
       tied up delivering messages and none can accept mail
       so perhaps we should make a separate queue runner mode
       and not have all the kids try and run the queue. 
    */
    
    // do a scandir, take an image, and romp through the list once. 
    Memory qdir = ar[NAME "-" TXT_QUEUEDIRECTORY];
    filelist fl(qdir, "", "Af"); // get all the Af files
    int mx = fl.numitems();
    int lp;
    for (lp = 0; lp < mx; lp++)
      {
        Memory fn(qdir);
        addslash(fn);
        fn.addstr(fl[lp]);
        runitem(fn);
      }
  }

void smtp::deadqueue(char *fn)
  {
    // move this file to the deadqueue 
    Memory dest(ar[NAME "-" TXT_DEADQUEUEDIRECTORY]);
    addslash(dest);
    
    ls(NOTEINF) << "Moving file to dead queue: " << fn << endl;
    char *s = strrchr(fn, '/');
    if (s == NULL)
      s = fn+1; // skip the /
    dest.addstr(s);
    // if this fails there's not much we can do about it.
    if (rename (fn, dest) == -1)
      {
        ls(NOTEERR) << "Unable to move " << fn << " to dead queue " << (char *)dest << endl;
        ls(NOTEERR) << "Error: " << errno << " " << strerror(errno) << ". Removing file." << endl;
        unlink(fn);
      }
  }

int smtp::grablock(Memory &newname, char *oldname)
  {
    /* atomically rename this file thereby locking others out
       from trying to process it. */
    // try and lock the file (proc.*) if we get an error, 
    // log if it's not FNF, return RET_ERROR

    ls(NOTEDIA) << "Grabbing lock on queue file: " << oldname << endl;
    newname = oldname;
    char *r = strrchr(newname, '/');
    if (r == NULL)
      r = newname;
    else
      r++; // skip the /
    Memory savename(r);
    strcpy (r, TXT_LOCKPREFIX);
    newname.addstr(savename); // this will resize to be big enough to hold the name with the working.prefix

    int ret = link(oldname, newname);
    if (ret != 0)
      {
        if (errno == EEXIST)
          return RET_ERROR; // this is okay, somebody else already has it. they're processing it
        if (errno == ENOENT)
          return RET_ERROR; // this is also okay, the source file doesn't exist because somebody beat me to it
        // in all other cases, we have problems.
        ls(NOTEWRN) << "Error trying to lock file: " << oldname << " to " << (char *)newname << " Error " << errno << " " << strerror(errno) << endl;
        return RET_ERROR; // either way, caller can't process this file.
      }
    // we have the lock file, remove the original so nobody else gets it.
    ret = unlink(oldname);
    if (ret != 0)
      {
        ls(NOTEWRN) << "Error trying to remove link: " << (char *)oldname << " Error " << errno << " " << strerror(errno) << endl;
        // foreboding of issues later, but best we process the file now and deal with duplicates than never deliver it at all
      }
    return RET_OK;
  }

int smtp::ungrablock(char *fn)
  {
    return gungrablock(ar, ls, fn);
  }

void smtp::runitem(char *origfn)
  {
    /* attempt to deliver this item, be it local or outgoing */

    Memory fn;
    if (grablock(fn, origfn) != RET_OK)
      return; // couldn't get that file for some reason
      
    FILE *faf = fopen(fn, TXT_FOPENREAD);
    if (faf == NULL)
      {
        if (errno != 2) // if it's not fnf, we should let somebody know
          ls(NOTEERR) << "Error trying to open queue file: " << fn << " Error: " << errno << " " << strerror(errno) << endl;
        return; // not now I guess, may have been picked up by somebody else.
      }
    filecloser x(faf);
    
    Memory bfname;
    Memory mailfrom;
    Memory storeid;
    int attempts;
    int isbounce;
    time_t lastcheck;
    Memory lasterror;
    // we REALLY need to make affile a class */
    int ret = readaffile(faf, bfname, mailfrom, storeid, attempts, isbounce, lastcheck);
    if (ret != RET_OK)
      {
        ls(NOTEERR) << "Invalid queue file: " << fn << endl;
        x.close();
        deadqueue(fn); // the working. file
        return;
      }
    
    // check the time, make sure we've waited 15 minutes or whatever
    time_t now = time(NULL);
    int secs = val(ar[NAME "-" TXT_QUEUE_RETRY_SECONDS]);
    if (secs < MIN_QUEUE_RETRY_SECONDS)
      secs = MIN_QUEUE_RETRY_SECONDS;
    if (now - lastcheck < secs)
      {
        ungrablock(fn); // not for us to do now.
        return;
      }
  
    ls(NOTEINF) << "Processing queue file: " << fn << endl;
        
    // we have to read in the whole file before we check attempts so we
    // can have the last error message should we need to mail it.
    /* make believe we are the current session. Actually, we ARE the 
       current session. */
    peerclaim = ""; // only used in data command
    envelopemailfrom = mailfrom;
    msgstoreid = storeid;

    /* Read in all the destination packages. */
    reciplist.WipeList(); // might as well put it here.
    while (1)
      {
        destpack *dp = new destpack();
        ret = dp->restore(faf);
        if (ret == RET_ENDOFDATA)
          {
            delete dp;
            break;
          }
        if (ret != RET_OK)
          {
            ls(NOTEERR) << "Invalid queue file: " << fn << endl;
            x.close();
            deadqueue(fn); // the working. file
            delete dp;
            return; 
          }
        reciplist.Add(dp);
      }

    readaferror(faf, lasterror); // load in the lasterror message from last time
    x.close(); // we don't delete so in case of powerfailure, we don't lose anything

    /* Okay, now we can process the file whacking the working file
       if we need to requeue something. */

    /* check the attempt count */
    
    /* If we wanna be nice and do that 4 hour useless notification thing, 
       do it here, personally, I find it annoying. */
    if (attempts > val(ar[NAME "-" TXT_MAX_DELIVERY_ATTEMPTS]))
      {
        if (isbounce)
          {
            // then we don't bounce it, we drop it quietly
            ls(NOTEINF) << "Bounce message queue file: " << origfn << endl;
            ls(NOTEINF) << "has attempted delivery " <<
                           val(ar[NAME "-" TXT_MAX_DELIVERY_ATTEMPTS]) << 
                           " times. Discarding." << endl;
            ls(NOTEINF) << "Last error: " << (char *)lasterror << endl;
            unlink(bfname);
            unlink(fn);
            return; // we're done here.
          }
        /* We should attempt to bounce this message to the sender with an excuse. */
        /* We send one message to the sender with the failed list. */
        Memory failedlist;
        int lp;
        for (lp = 0; lp < reciplist.ListCount(); lp++)
          {
            destpack *dp = (destpack *)reciplist.GetItem(lp);
            Memory th = dp->getaddress()->getemail();
            failedlist.addstr(th);
            failedlist.addstr(" ");
          }

        ret = generateundeliverablemessage(failedlist, bfname, lasterror);
        if (ret == FALSE)
          {
            /* If we can't generate a bounce message, trash it. */
            ls(NOTEINF) << "Queue file: " << origfn << " has attempted delivery "
                        << val(ar[NAME "-" TXT_MAX_DELIVERY_ATTEMPTS]) 
                        << ". Bounce message create failed. Discarding." << endl;
            unlink(bfname);
            unlink(fn);
            return; // we're done here.
          }
        /* We tried everything in this queue file max number of times,
           it continued to fail, and we sent a bounce message back.
           Kill off the message queue item, we're done with it. */
        unlink(bfname);
        unlink(fn);
        return; // we're done.
      }
      
    /* We will gather up all the errors that happen during this process
       of all these recips. That will be the last error. If people can
       accept it over multiple retries, the final bounce error message 
       will be only the problems left after all retries are exhausted. */
    lasterror = ""; // reset the error message pile.

    // we requeue to the temp. file if we have to.
    /* in most cases we leave the bf file alone because there's still
       stuff to do, only if all deliverys succeed do we remove it. */
    faf = NULL; // flag for writing to new queue file
    attempts++; // so when we write out next queue file we note we tried again

    Memory newfn; // file name of new queue file if we need one

    // hmm this looks a lot like the data command.     
    // I guess it's kinda lousy that we have to do everything twice.
    // For account types, we can put it in deliverlocal
    // because that's sorta what they are.
    
    int lp;
    destpack *dp;
    int max = reciplist.ListCount();
    for (lp = 0; lp < max; lp++)
      {
        int ret;
        dp = (destpack *)reciplist.GetItem(lp);
        int qthis = FALSE;
        if (dp->getislocal())
          {
            Memory errorstr;
            ret = deliverlocal(dp, bfname, errorstr); // this file to this package
            /* If a message got here then it must be to a valid person
               so any errors must be temporary. Any overquotas come back
               as TRUE, so FALSE is a requeuable event, and the reader
               will take care of too many retries. */
            if (ret == FALSE)
              {
                // I think we can't deliver a local bounce, they are always outgoing
                if (isbounce) // if we got a failure trying to deliver a bounce message, just drop it.
                  {
                    ls(NOTEINF) << "Bounce message queue file: " << origfn << " has permanently failed delivery. Discarding." << endl;
                    // we will fall through the loop and it will be deleted.
                  }
                else
                  {
                    lasterror.addstr(errorstr);
                    lasterror.addstr(CRLF);
                    qthis = TRUE;
                  } // else not a bounce message that failed
                // fall back into for loop without queueing, we're done with this guy.
              }
          }
        else // it's an outbound delivery
          {
            // a nice optomization would be to group these...
            // it's an easy hack, sort the list by domain and 
            // spin the lp in here until the domain changes.
            Memory errorstr;
            emailaddr *toaddr = dp->getaddress();
            Memory toemail = toaddr->getproperemail();

            /* get the global host. relay domains only work if
               you can send directly, if you have a global domain,
               chances are, that's because everybody has your IP blacklisted.
               But still if you have a machine on your local network
               you want to foward a domain to, you can still do this.
               But we will never send out a domain relayed mail via
               the global server. */
            
            Memory globalhost;
            int globalport;

            Memory relayhost = dp->getrelayhost();
            if (strlen(relayhost) > 0)
              {
                Memory junk;
                splithostportpath(relayhost, globalhost, globalport, junk);
              }
            else
              { // read it from config file if it's there.
                globalhost = ar[NAME "-" TXT_GLOBAL_OUTBOUND_SMTP_SERVER];
                globalport = val(ar[NAME "-" TXT_GLOBAL_OUTBOUND_SMTP_PORT]);
              }

            smtpsend s(ar, ls, toemail, envelopemailfrom, bfname);
            s.addrecip(toemail); // send one at a time.
            ret = s.go(errorstr, globalhost, globalport);
            if (ret != RET_OK)
              {
                if (ret == RET_PERMANENT_ERROR)
                  {
                    if (isbounce) // if we got a failure trying to deliver a bounce message, just drop it.
                      {
                        ls(NOTEINF) << "Bounce message queue file: " << origfn << " has permanently failed delivery. Discarding." << endl;
                        // we will fall through the loop and it will be deleted.
                      }
                    else
                      {
                        // don't queue it, send back a bounce and continue on
                        ret = generateundeliverablemessage(toemail, bfname, errorstr);
                        if (ret == FALSE)
                          { /* If we can't generate a bounce message, trash it. */
                            ls(NOTEERR) << "Unable to send bounce message for remote delivery permanent failure delivery to " << (char *)toemail << " Queue file: " << fn << ". Bounce message create failed. Discarding." << endl;
                          }
                      } // failed on a non bounce message
                    // fall back into for loop without queueing, we're done with this guy.
                  }
                else
                  { // temporary requeueable error
                    lasterror.addstr(errorstr);
                    lasterror.addstr(CRLF);
                    qthis = TRUE;
                  }
              }
          }

        if (qthis)
          {
            // faf != NULL denotes keep bf, because we're going to use it again
            // make a new temp.af file
            if (faf == NULL) // first time we have to queue something new
              {
                // make a tempfilename from origfn
                maketempnamefromaf(newfn, origfn);
                faf = fopen(newfn, TXT_FOPENWRITE);
                if (faf == NULL)
                  {
                    ls(NOTEERR) << "Unable to create new requeue file: " << (char *)newfn << " Error: " << errno << " " << strerror(errno) << endl;
                    /* can't write new temp queue file, put old one back.
                       this may cause redelivery if #1 works,  but #2 fails
                       but better that then losing mail. */
                    ungrablock(fn); // we did the best we could
                    return;
                  }
                // write out the top of the file
                // this messageid and envmailfrom come from the queue file we read
                writeaffile(faf, bfname, envelopemailfrom, msgstoreid, attempts, isbounce, FALSE); // this is requeue, don't deliver now
              }
            // queue this undeliverable destination
            ret = dp->store(faf);
            if (ret == FALSE)
              {
                ls(NOTESEV) << "Error writing to: " << (char *)newfn << " Error: " << errno << " " << strerror(errno) << endl;
                ungrablock(fn);
                fclose (faf); 
                return; // we did all we can for this q item, try again later.
                // a problem here is that this is the same queue file, not a new one
                // a new queue file would have a newer time in it and wait 15 minutes
                // to try again, this one's going to spin hard, but then again
                // we're having a filesystem writing problem, what else you gonna do?
              }
          } // if we should requeue this item because it failed.
      } // while delivering to recips in this queue file.
    
    /* If we are here, we have attempted to deliver to everybody in this
       queue file. If faf is not NULL then finish writing the af temp 
       queue file, delete the old af queue file, and rename the new
       one to the old. If nothing else they won't be identical because of
       the retry time, and attempt #. */
    if (faf)
      {
        dp->markend(faf); // end of file mark
        writeaferror(faf, lasterror);
        fclose(faf); // no filecloser here.
      }
    else  // there were no errors, we can remove the queue and msg file
      {
        unlink(bfname);
      }
    
    // either way delete old proc.af file, we're done with it.
    // we either finished it, or have written a new q file.
    unlink(fn); // the proclock queue file.

    // and the last piece, if we have a new queue file, make it active.
    if (faf)
      {
        ret = rename(newfn, origfn);
        if (ret != 0)
          ls(NOTESEV) << "Error renaming: " << (char *)newfn << " to " << origfn << " Error: " << errno << " " << strerror(errno) << endl;
      }
  }

void smtp::delivercmdlinemail(void)
  {
    /* They specified -d on the command line. look for the
       From and To addresses, and try and do a delivery.
       If it fails it will requeue and the server will deal with it. */
    /* hmmm, how about we put it in the queue and call run queue 
       that sounds easy. */
       
    Memory from;
    Memory to;
    
    if (ar.getcmdlineparam('f', from) == FALSE)
      {
        printf ("To send mail, you must specify -f<from address>\n");
        return;
      }
    if (ar.getcmdlineparam('t', to) == FALSE)
      {
        printf ("To send mail, you must specify -t<to address>\n");
        return;
      }

    ls(NOTEINF) << "Sending mail from command line from " << (char *)from << " to " << (char *)to << endl;
    /* Make a queue filename bf file. */
    Memory aftemp, af, bf, idnoise;
    makequeuefilenames(aftemp, af, bf, idnoise);
    md5hash (msgstoreid, idnoise); // make a consistent size message store ID
    
    /* Read from stdin the mail file and write it to the
       bf file. It must be a decently valid mail file, we 
       do no checking. */
    
    Memory buf; 
    char last1 = 0;
    char last2 = 0;
    FILE *f = fopen(bf, TXT_FOPENWRITE);
    if (f == NULL)
      {
        ls(NOTEERR) << "Unable to create message file: " << (char *)bf << " Error: " << errno << " " << strerror(errno) << endl;
        printf ("Unable to write to: %s. Error %d %s\n", (char *)bf, errno, strerror(errno));
        return;
      }
    filecloser xf(f);
    int c = fgetc(stdin);
    while (c != EOF)
      {
        if (c == '\n')
          if (last1 != '\r')
            fputc('\r', f); // make sure file is CRLF delimited
        last2 = last1;
        last1 = c;
        fputc (c, f);
        c = fgetc(stdin);
      }
    if ((last2 != '\r') || (last1 != '\n'))
      { // make sure last line is blank
        fputc('\r', f);
        fputc('\n', f);
      }
    if (ferror(stdin))
      {
        printf ("Unable to read message from stdin. Error %d %s\n", errno, strerror(errno));
        xf.close();
        unlink(bf);
        return;
      }
    if (ferror(f))
      {
        printf ("Unable to write message file %s. Error %d %s\n", (char *)bf, errno, strerror(errno));
        xf.close();
        unlink(bf);
        return;
      }
    xf.close();
    
    /* Write out the queue file to the tempname */
    FILE *faf = fopen(aftemp, TXT_FOPENWRITE);
    if (faf == NULL)
      {
        ls(NOTEERR) << "Unable to create queue file: " << (char *)aftemp << " Error: " << errno << " " << strerror(errno) << endl;
        printf ("Unable to queue message: %s. Error %d %s\n", (char *)aftemp, errno, strerror(errno));
        unlink(bf); // remove the message file
        return;
      }
    filecloser xa(faf);
    // this msgstore id comes from a hash of idnoise
    writeaffile(faf, bf, from, msgstoreid, 0, 0, TRUE); // new message attempts = 0, bounce = FALSE, deliver now
    /* make a destpack for the recip and add it to the queue file */
    destpack dp;
    emailaddr adr(to);
    dp.setaddr(adr, FALSE, "", 0, ""); 
    int ret = dp.store(faf);
    if (ret == FALSE)
      {
        printf ("Error writing to queue: %s. Error %d %s\n", (char *)aftemp, errno, strerror(errno));
        xa.close();
        unlink(bf); // clean this
        unlink(aftemp); // clean bad af file
        return;
      }
    dp.markend(faf); // end of file mark
    writeaferror(faf, ""); // being the first queueing of this item, there's no errmsg
    xa.close();

    /* The mail file is written, the queue file is written, rename it, 
       and it will get delivered. */
    if (rename(aftemp, af) != 0)
      {
        printf ("Error renaming %s to %s. Error %d %s\n", (char *)aftemp, (char *)af, errno, strerror(errno));
        printf ("Cancelling message.\n");
        unlink(aftemp); 
        unlink(bf);
        return;
      }
    // we made a valid queue file, not run the queue to deliver it.
    runqueue(); // we tried or will try later.
  }

// And That Ladies and Gentlemen, Is That.


