/*
    This file is part of O3.
    Copyright 2002, 2003 Stu Mark
    This program is distributed under the terms of the GNU 
    General Public License 

    O3 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    O3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with O3; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <resolv.h>
#include <arpa/nameser.h>
#include <errno.h>
#include <netdb.h>
#include <sys/stat.h>

#include <Q/memm.h>
#include <Q/socketio.h>
#include <Q/connector.h>
#include <Q/configcraft.h>
#include <Q/notecraft.h>
#include <Q/zlist.h>

#include "global.h"
#include "smtpsend.h"


/* This guy does the smtp client conversation to send a mail.
   Does the MX record lookup and will follow cnames 5 levels. */

#ifndef MAXPACKET        // make sure a maximum packet size is declared by BIND
  #define MAXPACKET 8192 // BIND maximum packet size
#endif
#ifndef HFIXEDSZ         // make sure header size is declared
  #define HFIXEDSZ 12
#endif
 
typedef union
  {
    HEADER hdr;                                     // define a header structure
    u_char qbuf[MAXPACKET];          // define a query buffer
  } mxquery;
       
/* Sending a piece of mail yields either success (RET_OK),
   temporary error (RET_ERROR)or permenant error (RET_PERMANENT_ERROR) */
       
smtpsend::smtpsend (ConfigCraft &Ar, NoteCraft &Ls, 
                    char *toaddr, char *from, 
                    char *Msgfile) : ar(Ar), ls(Ls)
  {
    msgfile = Msgfile;
    emailaddr ize(from);
    mailfrom = ize.getproperemail();
    
    /* Get the mx list of machines to try from this toaddr */
    emailaddr dom(toaddr);
    mxhost = dom.getdomain();
  }

smtpsend::~smtpsend(void)
  {
  }

void smtpsend::addrecip(char *s)
  {
    // add this name to the recip list
    // these MUST have <> brackets around them
    stringentry *str = new stringentry(s);
    reciplist.Add(str);
  }
  
int smtpsend::badback (char *s1, char *s2, Memory &ret, socketio &sco, NoteCraft &ls, int retval)
  { 
    // common error return function
    Memory in;
    ret = s1;
    ret.addstr(s2);
    sco.sendline ("quit");
    sco.getline (in); // get response
    ls(NOTEDIA) << in << endl;
    ls(NOTEDIA) << ret << endl;
    return retval;
  }

int smtpsend::gethostlistfordomain(Memory &rmsg, zlist &hostlist, char *mxdomain)
  {
    /* do dns lookup for the MX records for this host.
       for each one, if it's a cname, follow it one level.
       get the A record hostname and strentry add it to the
       hosts list.
       If there are no mx records, look up the A record and try that.
       If nothing, return an empty list. */

    // deal with mx record pointing to cname, keep resolving up to
    // some max number.

    int ret = res_init();
    if (ret != 0)
      {
        rmsg = "Unable to initialize resolver.";
        return RET_ERROR;
      }

    hostlist.WipeList();
    ret = gethostlistfordomainsub(rmsg, hostlist, mxdomain, 5); // follow cname 5 levels deep max
    return ret;
  }

int smtpsend::go(Memory &rmsg, Memory &globalhost, int globalport)
  {
    /* A feature for those of us running via an ISP that blocks outgoing
       smtp to anybody but them. 
       If it's there's a global outbound smtp server, use that. 
       Otherwise go where we were told to. */

    // something fundamentally wrong about hardcoding this.
    // But I think it's funny.
    int port = 25; // this ain't ever gonna change unless overridden
    zlist hosts;

    Memory host = globalhost;
    if (strlen(host) > 0)
      { // just make the list out of this one global host
        port = globalport;
        stringentry *s = new stringentry(host);
        hosts.Add(s);
      }
    else
      {
        /* here we get the list of mxes for this mxhost 
           and we loop try until we get one that connects.
           
           Perhaps we should keep trying mxes until we get 
           something that not only connects, but answers?
           Let's compromise, we'll wait for the greeting. */
           
        if (gethostlistfordomain(rmsg, hosts, mxhost) != RET_OK)
          return RET_ERROR;
      }

    /* Now we have a list of smtp servers to try and send this mail to. */

    int mx = hosts.ListCount();
    if (mx == 0)
      {
        rmsg = "No mail handlers for this domain: ";
        rmsg.addstr(mxhost);
        return RET_PERMANENT_ERROR;
      }
    int lp;
    int ret = RET_ERROR; // we have a host to try, so if we can't connect it's a temporary failure
    for (lp = 0; lp < mx; lp++)
      {
        stringentry *s = (stringentry *)hosts.GetItem(lp);
        Memory h = s->getstr();
        ret = gohost(rmsg, h, port);
        if (ret == RET_OK)
          return ret; // this one worked
        if (ret == RET_CONNECTFAILED)
          continue; // try the next one
        break; // some other bad error
      }
    return ret; // either we ran out of hosts or got some other kind of error
  }

int smtpsend::gohost(Memory &rmsg, char *server, int port)
  {
    /* Return RET_OK if the mail was sent, RET_ERROR or whatever if any error.
       RET_CONNECTFAILED if we can't get there, causing caller
       to retry other mx hosts.
       
       If there's even one bad address, don't send it.
       Fill rmsg with text of all the problems found. 
       and parts of the smtp coversation that were interesting
       about the failure. */

    rmsg = "";
    if (reciplist.ListCount() == 0)
      {
        rmsg = "No recipients in list.";
        return RET_PERMANENT_ERROR;
      }
      
    int timeout = val(ar[TXT_SOCKETTIMEOUT]);

    Memory hostname(257); // who to say hello from 
    int ret = gethostname(hostname, 256);
    if (ret != 0) // how could this happen
      hostname = TXT_DEFAULTHOSTNAME;

    ls(NOTEINF) << "Delivering mail to: " << server << ":" << port << endl;
    Connector con(server, port, timeout); // does not own socket.
    if (con.errorstate() != 0)
      {
        rmsg = "Error on connection to SMTP server: ";
        rmsg.addstr(strerror(con.errorstate()));
        ls(NOTEERR) << (char *)rmsg << endl;
        return RET_CONNECTFAILED;
      }
    socketio sco(con.victim(), con.victim(), timeout); // this scope owns the connection
    
    /* Let's chat */
    Memory in;
    ret = sco.getline(in); // get the header
    if (ret == FALSE)
      { 
        rmsg = "Error communicating with SMTP server on greeting.";
        return RET_CONNECTFAILED; // retry other host
      }
   
    // do ehlo and see if message size is here
    Memory resp("EHLO ");
    resp.addstr(hostname);
    ret = sco.sendline(resp);
    
    int dosize = FALSE;

    while (1)
      {
        ret &= sco.getline(in);
        if (ret == FALSE)
          return badback("During HELO, error communicating with SMTP server.", "", rmsg, sco, ls, RET_ERROR);
        if (memcmp(in, "250", 3) != 0)
          return badback("During HELO, error from SMTP server: ", in, rmsg, sco, ls, RET_ERROR); // temp error

        // check options.
        if (strncasecmp(in+4, TXT_SIZE, strlen(TXT_SIZE)) == 0)
          dosize = TRUE;
        if (*(in+3) == ' ') // the last option
          break;
      }

    // do the mail from
    ls(NOTEDEB) << "Sending envelope information." << endl;
    resp = TXT_MAIL_FROM " ";
    resp.addstr(mailfrom); 
    if (dosize)
      { // add the file size
        struct stat fsize;
        if (stat(msgfile, &fsize) == 0) // if we can get it.
          {
            resp.addstr(" " TXT_SIZE "=");
            resp.addnum(fsize.st_size);
          }
      }
    ret &= sco.sendline(resp);
    ls(NOTEDEB) << (char *)resp << endl;

    ret &= sco.getline(in);
    if (ret == FALSE)
      return badback("During MAIL FROM, error communicating with SMTP server.", "", rmsg, sco, ls, RET_ERROR); // all communications errors are temp errors

    if (memcmp(in, "4", 1) == 0) // a 4 series error is temporary
      return badback("During MAIL FROM, temporary error from SMTP server: ", in, rmsg, sco, ls, RET_ERROR);
    if (memcmp(in, "250 ", 4) != 0) // a 250 is good, everything else is permanent
      return badback("During MAIL FROM, permanent error from SMTP server: ", in, rmsg, sco, ls, RET_PERMANENT_ERROR);

    /* Now do all the rcpt to's */
    int lp;
    Memory rcpterr;
    int retval = RET_ERROR;
    for (lp = 0; lp < reciplist.ListCount(); lp++)
      {  
        resp = "RCPT to: ";
        stringentry *na = (stringentry *)reciplist.GetItem(lp);
        resp.addstr(na->getstr());
        ret = sco.sendline(resp);
        ls(NOTEDEB) << (char *)resp << endl;

        ret &= sco.getline(in);
        if (ret == FALSE)
          return badback("During RCPT, error communicating with SMTP server.", "", rmsg, sco, ls, RET_ERROR);

        /* When we do fix this to do multiple rcpts to the same domain, we have
           to watch when we return temporary and permanent errors for individual
           recips. */
        if (memcmp(in, "250 ", 4) != 0)
          {
            rcpterr.addstr("During RCPT TO: ");
            stringentry *se = (stringentry *)reciplist.GetItem(lp);
            rcpterr.addstr(se->getstr());
            rcpterr.addstr(" - ");
            rcpterr.addstr (in);
            if (*in == '5') // 500 level error?
              retval = RET_PERMANENT_ERROR;
          }
      }
    /* See if we got any errors on rcpt: */
    if (strlen(rcpterr) > 0)
      {
        /* uh oh. no go. */
        return badback("", rcpterr, rmsg, sco, ls, retval);
      }
    
    /* If we're here, okay to go. Just send data and dump the message. */
    resp = "data";
    ret = sco.sendline(resp);
    ls(NOTEDEB) << "Sending message data" << endl;

    ret &= sco.getline(in);
    if (ret == FALSE)
      return badback("During DATA, error communicating with SMTP server.", "", rmsg, sco, ls, RET_ERROR);
    if (memcmp(in, "354 ", 4) != 0)
      return badback("During DATA, error from SMTP server: ", in, rmsg, sco, ls, RET_ERROR);

    ret = sendmsg (sco, msgfile, ls); // returns true or false, not RET_*
    ret &= sco.senddata((void *)CRLF, strlen(CRLF));
    ret &= sco.sendline("."); // end the message text
    if (ret == FALSE)
      return badback("During DATA, error sending message to SMTP server.", "", rmsg, sco, ls, RET_ERROR);

    ret &= sco.getline(in);
    if (ret == FALSE)
      return badback("During DATA, error communicating with SMTP server.", "", rmsg, sco, ls, RET_ERROR);
    retval = RET_ERROR;
    if (memcmp(in, "250 ", 4) != 0)
      {
        if (*in == '5')
          retval = RET_PERMANENT_ERROR;
        return badback("Error from SMTP server: ", in, rmsg, sco, ls, retval);
      }

    // we're done.
    sco.sendline ("quit");
    sco.getline (in); // get response
    ls(NOTEINF) << "Sending message complete" << endl;
    return TRUE; // good things come to those who complete correctly closes the connection
  }

int sendmsg (socketio &sco, char *msgfile, NoteCraft &ls)
  {
    /* this used to be a nice and simple one liner that
       would just dump out the message because the message
       was in perfect condition, it was even dot escaped
       so we could SMTP it. The problem is we're going
       to need to store it in good form in sent-mail
       which means we can't build it with the dots, we
       have to build it as displayable so we can
       store it correctly and do the dot escape
       as we smtp it, which is here. */
    // returns true or false
    Memory line(1000);
    FILE *f = fopen(msgfile, "r");
    if (f == NULL)
      {
        ls(NOTEERR) << "Unable to open " << msgfile << " to send message. " << endl;
        ls(NOTEERR) << "Error: " << errno << " " << strerror(errno) << endl;
        return FALSE; // denote that something bad happened
      }
    filecloser xkill(f);
      
    int ret = readline(line, f);
    int sret = TRUE;
    
    while (ret > 0)
      {
        if (line[0] == '.')
          sret &= sco.senddata((void *)".", 1);
        sret &= sco.senddata(line, strlen(line));
        sret &= sco.senddata((void *)CRLF, strlen(CRLF));
        ret = readline (line, f);
      }
    return sret;
  }

int prefixsort (const void *i1, const void *i2)
  {
    /* these are strentries with a 5 digit prefix to hostname to sort by.
       It's space padded so just do alphasort */
    stringentry **ss1 = (stringentry **)i1;
    stringentry **ss2 = (stringentry **)i2;
    stringentry *s1 = *ss1;
    stringentry *s2 = *ss2;
    char *m1 = (char *)s1->getstr();
    char *m2 = (char *)s2->getstr();
    return strcmp(m1, m2);
  }

int gethostlistfordomainsub(Memory &rmsg, zlist &hostlist, char *domain, int follows)
  {    
    /* Fill up hostlist with mx hosts for domain
       return RET_OK or ret_error */
    UCHAR *end;
    UCHAR *pos; // pointers to end of message, generic u_char pointer.
    int ret;
    HEADER *hptr;
    mxquery dnsresponse; 
    int numanswers;
    int numquestions;
    
    USHORT pref, type;
    struct hostent *h; // for the A record, if needed
   
    // make query
    ret = res_query(domain, ns_c_in, ns_t_mx, (UCHAR *)&dnsresponse, sizeof(dnsresponse));
    if (ret < 0)
      {
        switch(h_errno)
          {
            case NO_DATA:
            case NO_RECOVERY: // no MX records, try A
              h = gethostbyname(domain);
              if (h != NULL)
                { // returned a resolved result, store
                  if (h->h_name != NULL)
                    {
                      if ( strlen( h->h_name) != 0)
                        {
                          Memory resp(MAXDNAME+10);
                          sprintf (resp, "%5d%s", 10, h->h_name); // my silly format, force preference at 10
                          stringentry *s = new stringentry(resp);
                          hostlist.Add(s);
                        }
                    }
                  else
                    {
                      Memory resp(MAXDNAME+10);
                      sprintf (resp, "%5d%s", 10, domain); // It's me!
                      stringentry *s = new stringentry(resp);
                      hostlist.Add(s);
                    }
                  return RET_OK; 
                }
              rmsg = "Domain ";
              rmsg.addstr(domain);
              rmsg.addstr(" not found.");
              return RET_ERROR;

            default:
              break;
          }

        rmsg = "Domain search query for ";
        rmsg.addstr(domain);
        rmsg.addstr(" failed. Domain not found.");
        return RET_ERROR;
      }  // if query failed

    // end of response position, don't exceed our workspace
    UINT n = ret;
    if (n > sizeof(dnsresponse))
      n = sizeof(dnsresponse);
   
    // skip the question portion of the DNS packet
    hptr = (HEADER *)&dnsresponse;
    pos = (UCHAR *)&dnsresponse + HFIXEDSZ; // point after the header
    end = (UCHAR *)&dnsresponse + n; // point right after the entire answer

    for (numquestions = ntohs(hptr->qdcount); numquestions > 0; numquestions--)
      {
        // loop on question count (taken from header), and skip them one by one
        if ((n = dn_skipname(pos, end)) < 0)
          {
            rmsg = "Invalid DNS packet, search query for ";
            rmsg.addstr(domain);
            rmsg.addstr(" failed.");
            return RET_ERROR;
          }
        pos += n + QFIXEDSZ;
      }

    numanswers = ntohs(hptr->ancount); // number of answers
   
    while ((--numanswers >= 0) && (pos < end))
      {
        // pos constantly changes (moves forward in the answer buffer)
   
        // decompress the domain's default host name into the buffer so 
        // we can skip it and check if it's an MXhost
        Memory MXdomain(MAXDNAME);
        if ((n = dn_expand((UCHAR *)&dnsresponse, end, pos, MXdomain, MAXDNAME)) < 0)
          break; // error in answer
        pos += n;   // skip the name, go to its attributes

        type = ntohs(*((USHORT *)pos));
        pos += sizeof(short int);
        pos += INT16SZ + INT32SZ; // skip the class and TTL portion of the answer RR

        n = ntohs(*((USHORT *)pos));
        pos += sizeof(short int);  
        if (type == ns_t_cname)
          {
            // follow the cname up to 5 times, because I'm a nice guy
            int ret;
            Memory srmsg;
            if (follows == 0)
              { // we've followed this cname too far already. Bail
                pos += n;
                continue;
              }
            /* get the cname and follow it. Adding to the list as it goes.
               If we come back okay, then we're done with this entry,
               move on to next one. */
            Memory MXhost(MAXDNAME);
            if ((n = dn_expand((UCHAR *)&dnsresponse, end, pos, MXhost, MAXDNAME)) < 0) // expand the cname
              break; // bad format dns packet, get out
 
            ret = gethostlistfordomainsub(srmsg, hostlist, MXhost, follows-1);
            if (ret != RET_OK)
              { // we're being nice here, if we have problems doing this, drop it
                pos += n;
                continue;
              }
            rmsg.addstr(srmsg);
            rmsg.addstr(CRLF);
            pos += n; // we dealt with this entry, go on to next one
            continue;
          }
        if (type != ns_t_mx)
          { // not an mx record
            pos += n;
            continue;
          }
        pref = ntohs(*((USHORT *)pos));
        pos += sizeof (short int);
        Memory MXhost(MAXDNAME);
        if ((n = dn_expand((UCHAR *)&dnsresponse, end, pos, MXhost, MAXDNAME)) < 0) // expand the MXRR
          break; // error in decompression
        pos += n; // skip the resource info
   
        Memory resp(MAXDNAME+10);
        sprintf (resp, "%5d%s", pref, (char *)MXhost); // my silly format so we can zlist sort on preference
        stringentry *str = new stringentry(resp);
        hostlist.Add(str);
      }

    hostlist.Sort(prefixsort);
    // now remove the order prefix
    int mx = hostlist.ListCount();
    int lp;
    for (lp = 0; lp < mx; lp++)
      {
        stringentry *s = (stringentry *)hostlist.GetItem(lp);
        s->getstr() = s->getstr()+5; // remove the first 5 chars
      }
    return RET_OK;
  }



